#include	"X.h"
#include	"Xmd.h"
#include	"Xproto.h"
#include	"gcstruct.h"
#include	"windowstr.h"
#include	"scrnintstr.h"
#include	"pixmapstr.h"
#include	"regionstr.h"

#define PSZ 8
#include	"../../cfb/cfb.h"
#include	"cfb16.h"
#if defined(CONFIG_CFB24)
#include	"cfb24.h"
#endif
#include	"cfb32.h"

#include <asm/vc_ioctl.h>
#include <string.h>
#include <asm/io.h>

unsigned char *IMSMMIOBase = 0;
int enforceTT128 = FALSE;
extern int inXpmac;
extern int vgaBitsPerPixel;
#define xf86VTSema inXpmac

extern GCOps cfb32TEOps1Rect, cfb32TEOps, cfb32NonTEOps1Rect, cfb32NonTEOps,
	     cfb24TEOps1Rect, cfb24TEOps, cfb24NonTEOps1Rect, cfb24NonTEOps,
	     cfb16TEOps1Rect, cfb16TEOps, cfb16NonTEOps1Rect, cfb16NonTEOps,
	     cfbTEOps1Rect, cfbTEOps, cfbNonTEOps1Rect, cfbNonTEOps;
extern unsigned char *DacIndex;
extern unsigned char *DacData;
extern unsigned char savedHWColormap[];

static int screen_width;
static int tt128Revision;

/* IMS TwinTurbo blitter registers and macros */
#define S1SA     0x0000
#define S2SA     0x0004
#define SP       0x0008
#define DSA      0x000c
#define CNT      0x0010
#define DP_OCTL  0x0014
#define CLR      0x0018
#define BI       0x0020
#define MBC      0x0024
#define BLTCTL   0x0028
#define SSTATUS  0x0090

static inline void regw(unsigned long regindex, unsigned long regdata)
{
  register unsigned long base_addr = (unsigned long)IMSMMIOBase;
  asm volatile ("stwbrx %1,%2,%3; eieio"
          : "=m" (*(volatile unsigned *)(base_addr+regindex))
          : "r" (regdata), "b" (regindex), "r" (base_addr));
}


static inline unsigned long regr(unsigned long regindex)
{
  register unsigned long base_addr = (unsigned long)IMSMMIOBase, val;
  asm volatile ("lwbrx %0,%1,%2; eieio"
           : "=r"(val)
           : "b"(regindex), "r"(base_addr),
             "m" (*(volatile unsigned *)(base_addr+regindex)));
  return(val);
}

static inline void tt128BitBlt(int Bpp, int line_pitch, int xsrc, int ysrc,
			       int xdst, int ydst, int w, int h)
{
	unsigned long srcStart, dstStart;
	unsigned int sp, dp_octl, cnt, bltctl;

	xsrc *= Bpp;
	xdst *= Bpp;
	w *= Bpp;
	w--;
	h--;

	bltctl = 0x05;
	sp = line_pitch << 16;
	cnt = h << 16;

	if (ysrc < ydst) {
		ysrc += h;
		ydst += h;
		sp |= -(line_pitch) & 0xffff;
		dp_octl = -(line_pitch) & 0xffff;
	} else {
		sp |= line_pitch;
		dp_octl = line_pitch;
	}
	if (xsrc < xdst) {
		xsrc += w;
		xdst += w;
		bltctl |= 0x80;
		cnt |= -(w) & 0xffff;
	} else {
		cnt |= w;
	}
	srcStart = ysrc * line_pitch + xsrc;
	dstStart = ydst * line_pitch + xdst;

	while(regr(SSTATUS) & 0x80);
	regw(S1SA, srcStart);
	regw(SP, sp);
	regw(DSA, dstStart);
	regw(CNT, cnt);
	regw(DP_OCTL, dp_octl);
	regw(BLTCTL, bltctl);
	while(regr(SSTATUS) & 0x80);
	while(regr(SSTATUS) & 0x40);
}


/*
 * DoBitbltCopy
 *
 * Author: Keith Packard
 * Modifications: Radoslaw Kapitan
 */ 
static void tt128DoBitbltCopy(DrawablePtr pSrc, DrawablePtr pDst, int alu, 
			   RegionPtr prgnDst, DDXPointPtr pptSrc, unsigned long planemask)
{
	BoxPtr pbox;
	int nbox;
	BoxPtr pboxTmp, pboxNext, pboxBase, pboxNew1, pboxNew2;
	/* temporaries for shuffling rectangles */
	DDXPointPtr pptTmp, pptNew1, pptNew2;
	/* shuffling boxes entails shuffling the
	   source points too */
	int careful;
	register int Bpp, line_pitch;

	if (!(pSrc->type == DRAWABLE_WINDOW && pDst->type == DRAWABLE_WINDOW) || !xf86VTSema) {
		switch (pSrc->bitsPerPixel) {
			case 8:
				cfbDoBitbltCopy(pSrc, pDst, alu, prgnDst, pptSrc, planemask);
				break;
			case 16:
				cfb16DoBitbltCopy(pSrc, pDst, alu, prgnDst, pptSrc, planemask);
				break;
#if defined(CONFIG_CFB24)
			case 24:
				cfb24DoBitbltCopy(pSrc, pDst, alu, prgnDst, pptSrc, planemask);
				break;
#endif
			case 32:
				cfb32DoBitbltCopy(pSrc, pDst, alu, prgnDst, pptSrc, planemask);
				break;
		}
		return;
	}

	pbox = REGION_RECTS(prgnDst);
	nbox = REGION_NUM_RECTS(prgnDst);

	pboxNew1 = NULL;
	pptNew1 = NULL;
	pboxNew2 = NULL;
	pptNew2 = NULL;

	/* XXX we have to err on the side of safety when both are windows,
	 * because we don't know if IncludeInferiors is being used.
	 */
	careful = ((pSrc == pDst) ||
		  ((pSrc->type == DRAWABLE_WINDOW) &&
		  (pDst->type == DRAWABLE_WINDOW)));

	if (careful && (pptSrc->y < pbox->y1)) {
		if (nbox > 1) {
			/* keep ordering in each band, reverse order of bands */
			pboxNew1 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox);
			if (!pboxNew1)
				return;
			pptNew1 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
			if (!pptNew1) {
				DEALLOCATE_LOCAL(pboxNew1);
				return;
			}
			pboxBase = pboxNext = pbox+nbox-1;
			while (pboxBase >= pbox) {
				while ((pboxNext >= pbox)
				    && (pboxBase->y1 == pboxNext->y1))
					pboxNext--;
				pboxTmp = pboxNext+1;
				pptTmp = pptSrc + (pboxTmp - pbox);
				while (pboxTmp <= pboxBase) {
					*pboxNew1++ = *pboxTmp++;
					*pptNew1++ = *pptTmp++;
				}
				pboxBase = pboxNext;
			}
			pboxNew1 -= nbox;
			pbox = pboxNew1;
			pptNew1 -= nbox;
			pptSrc = pptNew1;
		}
	}
 
	if (careful && (pptSrc->x < pbox->x1)) {
		if (nbox > 1) {
			/* reverse order of rects in each band */
			pboxNew2 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox);
			pptNew2 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
			if (!pboxNew2 || !pptNew2) {
				if (pptNew2)
					DEALLOCATE_LOCAL(pptNew2);
				if (pboxNew2)
					DEALLOCATE_LOCAL(pboxNew2);
				if (pboxNew1) {
					DEALLOCATE_LOCAL(pptNew1);
					DEALLOCATE_LOCAL(pboxNew1);
				}
				return;
			}
			pboxBase = pboxNext = pbox;
			while (pboxBase < pbox+nbox) {
				while ((pboxNext < pbox+nbox)
				    && (pboxNext->y1 == pboxBase->y1))
					pboxNext++;
				pboxTmp = pboxNext;
				pptTmp = pptSrc + (pboxTmp - pbox);
				while (pboxTmp != pboxBase) {
					*pboxNew2++ = *--pboxTmp;
					*pptNew2++ = *--pptTmp;
				}
				pboxBase = pboxNext;
			}
			pboxNew2 -= nbox;
			pbox = pboxNew2;
			pptNew2 -= nbox;
			pptSrc = pptNew2;
		}
	}

	Bpp = vgaBitsPerPixel >> 3;
	line_pitch = screen_width * Bpp;
	while (nbox--) {
		tt128BitBlt(Bpp, line_pitch,
			 pptSrc->x, pptSrc->y, pbox->x1, pbox->y1,
			 pbox->x2 - pbox->x1, pbox->y2 - pbox->y1);
		pbox++;
		pptSrc++;
	}

	/* free up stuff */
	if (pboxNew2) {
		DEALLOCATE_LOCAL(pptNew2);
		DEALLOCATE_LOCAL(pboxNew2);
	}
	if (pboxNew1) {
		DEALLOCATE_LOCAL(pptNew1);
		DEALLOCATE_LOCAL(pboxNew1);
	}
}

static RegionPtr tt128CopyArea(DrawablePtr pSrcDrawable, DrawablePtr pDstDrawable, GC *pGC,
	      int srcx, int srcy, int width, int height, int dstx, int dsty)
{
	void (*doBitBlt)() = tt128DoBitbltCopy;

	switch(pSrcDrawable->bitsPerPixel) {
	  case 8:
	    if (pGC->alu != GXcopy || (pGC->planemask & 0xFF) != 0xFF) {
		doBitBlt = cfbDoBitbltGeneral;
		if ((pGC->planemask & 0xFF) == 0xFF) {
			switch (pGC->alu) {
				case GXxor:
					doBitBlt = cfbDoBitbltXor;
					break;
				case GXor:
					doBitBlt = cfbDoBitbltOr;
					break;
			}
		}
	    }
	    return cfbBitBlt(pSrcDrawable, pDstDrawable, pGC, 
			     srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
	  case 16:
	    if (pGC->alu != GXcopy || (pGC->planemask & 0xFFFF) != 0xFFFF) {
		doBitBlt = cfb16DoBitbltGeneral;
		if ((pGC->planemask & 0xFFFF) == 0xFFFF) {
			switch (pGC->alu) {
				case GXxor:
					doBitBlt = cfb16DoBitbltXor;
					break;
				case GXor:
					doBitBlt = cfb16DoBitbltOr;
					break;
			}
		}
	    }
	    return cfb16BitBlt(pSrcDrawable, pDstDrawable, pGC, 
			       srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
#if defined(CONFIG_CFB24)
	  case 24:
	    if (pGC->alu != GXcopy || (pGC->planemask & 0xFFFFFF) != 0xFFFFFF) {
		doBitBlt = cfb24DoBitbltGeneral;
		if ((pGC->planemask & 0xFFFFFF) == 0xFFFFFF) {
			switch (pGC->alu) {
				case GXxor:
					doBitBlt = cfb24DoBitbltXor;
					break;
				case GXor:
					doBitBlt = cfb24DoBitbltOr;
					break;
			}
		}
	    }
	    return cfb24BitBlt(pSrcDrawable, pDstDrawable, pGC, 
			       srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
#endif
	  case 32:
	    if (pGC->alu != GXcopy || (pGC->planemask & 0xFFFFFFFF) != 0xFFFFFFFF) {
		doBitBlt = cfb32DoBitbltGeneral;
		if ((pGC->planemask & 0xFFFFFFFF) == 0xFFFFFFFF) {
			switch (pGC->alu) {
				case GXxor:
					doBitBlt = cfb32DoBitbltXor;
					break;
				case GXor:
					doBitBlt = cfb32DoBitbltOr;
					break;
			}
		}
	    }
	    return cfb32BitBlt(pSrcDrawable, pDstDrawable, pGC, 
			       srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
	}
}

static void tt128TileFillRect(DrawablePtr pDrawable, GCPtr pGC, int nrectFill, xRectangle *prectInit)
{
	switch (pDrawable->bitsPerPixel) {
		case 8:
			cfbPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			break;
		case 16:
			if (pGC->fillStyle == FillSolid || pGC->fillStyle == FillTiled)
				cfb16PolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			else
				miPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			break;
#if defined(CONFIG_CFB24)
		case 24:
			if (pGC->fillStyle == FillSolid || pGC->fillStyle == FillTiled)
				cfb24PolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			else
				miPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			break;
#endif
		case 32:
			if (pGC->fillStyle == FillSolid || pGC->fillStyle == FillTiled)
				cfb32PolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			else
				miPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
			break;
	}
}

static void tt128PolyFillRect(DrawablePtr pDrawable, GCPtr pGC, int nrectFill, xRectangle *prectInit)
{
	if (pDrawable->type != DRAWABLE_WINDOW || !xf86VTSema) {
		switch (pDrawable->bitsPerPixel) {
			case 8:
				cfbPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				break;
			case 16:
				if (pGC->fillStyle == FillSolid || pGC->fillStyle == FillTiled)
					cfb16PolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				else
					miPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				break;
#if defined(CONFIG_CFB24)
			case 24:
				if (pGC->fillStyle == FillSolid || pGC->fillStyle == FillTiled)
					cfb24PolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				else
					miPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				break;
#endif
			case 32:
				if (pGC->fillStyle == FillSolid || pGC->fillStyle == FillTiled)
					cfb32PolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				else
					miPolyFillRect(pDrawable, pGC, nrectFill, prectInit);
				break;
		}
		return;
	}

	/*
	 * Optimize for solid fills
	 */
	if ((pGC->fillStyle == FillSolid)
	    /* || ((pGC->fillStyle == FillOpaqueStippled)
		&& (pGC->fgPixel == pGC->bgPixel)) */ ) {
		int    nClipRects;     /* Number of clipping rectangles */
		BoxPtr pClipRects;     /* Current clipping box */
		int    clipXMin;       /* Upper left corner of clip rect */
		int    clipYMin;       /* Upper left corner of clip rect */
		int    clipXMax;       /* Lower right corner of clip rect */
		int    clipYMax;       /* Lower right corner of clip rect */
		int    drawableXOrg;   /* Drawables x origin */
		int    drawableYOrg;   /* Drawables y origin */
		cfbPrivGC *priv;
		int line_pitch, Bpp, clr;
		unsigned int *clrReg = (unsigned int *)(IMSMMIOBase + CLR);

		drawableXOrg = pDrawable->x;
		drawableYOrg = pDrawable->y;

		priv = (cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr);
		nClipRects = REGION_NUM_RECTS(priv->pCompositeClip);
		pClipRects = REGION_RECTS(priv->pCompositeClip);

		Bpp = vgaBitsPerPixel >> 3;
		line_pitch = screen_width * Bpp;
		clr = pGC->fgPixel;

		while(regr(SSTATUS) & 0x80);
		regw(DP_OCTL, line_pitch);
		regw(SP, line_pitch);
		regw(BI, 0xffffffff);
		regw(MBC, 0xffffffff);
		switch(Bpp) {
		  case 1:
		    *clrReg = clr | (clr << 8) | (clr << 16) | (clr << 24);
		    break;
		  case 2:
		    *clrReg = clr | (clr << 16);
		    break;
		  case 4:
		    *clrReg = clr;
		    break;
		  default: /* ??? */
		    *clrReg = clr;
		    break;
		}
		while(regr(SSTATUS) & 0x80);
		while(regr(SSTATUS) & 0x40);

		for (; nClipRects > 0; nClipRects--, pClipRects++) {
			xRectangle *pRect;   /* current rectangle to fill */
			int         nRects;  /* Number of rectangles to fill */

			clipYMin = pClipRects->y1;
			clipYMax = pClipRects->y2;
			clipXMin = pClipRects->x1;
			clipXMax = pClipRects->x2;
			for (nRects = nrectFill, pRect = prectInit;
			     nRects > 0;
			     nRects--, pRect++) {
				int rectX1;     /* points used to find the width */
				int rectX2;     /* points used to find the width */
				int rectY1;     /* points used to find the height */
				int rectY2;     /* points used to find the height */
				int rectWidth;  /* Width of the rect to be drawn */
				int rectHeight; /* Height of the rect to be drawn */

				/*
				 * Clip and translate each rect
				 */
				rectX1 = max((pRect->x + drawableXOrg), clipXMin);
				rectY1 = max((pRect->y + drawableYOrg), clipYMin);
				rectX2 = min((pRect->x + pRect->width + drawableXOrg), clipXMax);
				rectY2 = min((pRect->y + pRect->height + drawableYOrg), clipYMax);

				rectWidth = rectX2 - rectX1;
				rectHeight = rectY2 - rectY1;

				if ((rectWidth > 0) && (rectHeight > 0)) {
					while(regr(SSTATUS) & 0x80);
					regw(DSA, rectY1 * line_pitch + rectX1 * Bpp);
					regw(S1SA, rectY1 * line_pitch + rectX1 * Bpp);
					regw(CNT, ((rectHeight - 1) << 16) | (rectWidth * Bpp - 1));
					if(tt128Revision > 1)
					  regw(BLTCTL, 0x200000);
					else
					  regw(BLTCTL, 0x840);
					while(regr(SSTATUS) & 0x80);
					while(regr(SSTATUS) & 0x40);
				}
			} /* end for loop through each rectangle to draw */
		} /* end for loop through each clip rectangle */

	} else { /* end section to draw solid patterns */
		tt128TileFillRect(pDrawable, pGC, nrectFill, prectInit);
	}
}


extern WindowPtr *WindowTable;

void tt128CopyWindow(WindowPtr pWin, DDXPointRec ptOldOrg, RegionPtr prgnSrc)
{
	DDXPointPtr pptSrc;
	register DDXPointPtr ppt;
	RegionRec rgnDst;
	register BoxPtr pbox;
	register int dx, dy;
	register int i, nbox;
	WindowPtr pwinRoot;

	pwinRoot = WindowTable[pWin->drawable.pScreen->myNum];

	REGION_INIT(pWin->drawable.pScreen, &rgnDst, NullBox, 0);

	dx = ptOldOrg.x - pWin->drawable.x;
	dy = ptOldOrg.y - pWin->drawable.y;
	REGION_TRANSLATE(pWin->drawable.pScreen, prgnSrc, -dx, -dy);
	REGION_INTERSECT(pWin->drawable.pScreen, &rgnDst, &pWin->borderClip, prgnSrc);

	pbox = REGION_RECTS(&rgnDst);
	nbox = REGION_NUM_RECTS(&rgnDst);
	if (!nbox || !(pptSrc = (DDXPointPtr) ALLOCATE_LOCAL(nbox * sizeof(DDXPointRec)))) {
		REGION_UNINIT(pWin->drawable.pScreen, &rgnDst);
		return;
	}
	ppt = pptSrc;

	for (i = nbox; --i >= 0; ppt++, pbox++) {
		ppt->x = pbox->x1 + dx;
		ppt->y = pbox->y1 + dy;
	}

	tt128DoBitbltCopy((DrawablePtr)pwinRoot, (DrawablePtr)pwinRoot, GXcopy, &rgnDst, pptSrc, ~0L);
	DEALLOCATE_LOCAL(pptSrc);
	REGION_UNINIT(pWin->drawable.pScreen, &rgnDst);
}

void tt128InitGC(void)
{
  switch(vgaBitsPerPixel) {
    case 8:
      cfbTEOps.CopyArea = tt128CopyArea;
      cfbNonTEOps.CopyArea = tt128CopyArea;
      cfbTEOps1Rect.CopyArea = tt128CopyArea;
      cfbNonTEOps1Rect.CopyArea = tt128CopyArea;
      cfbTEOps.PolyFillRect = tt128PolyFillRect;
      cfbNonTEOps.PolyFillRect = tt128PolyFillRect;
      cfbTEOps1Rect.PolyFillRect = tt128PolyFillRect;
      cfbNonTEOps1Rect.PolyFillRect = tt128PolyFillRect;
      break;
    case 16:
      cfb16TEOps.CopyArea = tt128CopyArea;
      cfb16NonTEOps.CopyArea = tt128CopyArea;
      cfb16TEOps1Rect.CopyArea = tt128CopyArea;
      cfb16NonTEOps1Rect.CopyArea = tt128CopyArea;
      cfb16TEOps.PolyFillRect = tt128PolyFillRect;
      cfb16NonTEOps.PolyFillRect = tt128PolyFillRect;
      cfb16TEOps1Rect.PolyFillRect = tt128PolyFillRect;
      cfb16NonTEOps1Rect.PolyFillRect = tt128PolyFillRect;
      break;
#if defined(CONFIG_CFB24)
    case 24:
      cfb24TEOps.CopyArea = tt128CopyArea;
      cfb24NonTEOps.CopyArea = tt128CopyArea;
      cfb24TEOps1Rect.CopyArea = tt128CopyArea;
      cfb24NonTEOps1Rect.CopyArea = tt128CopyArea;
      cfb24TEOps.PolyFillRect = tt128PolyFillRect;
      cfb24NonTEOps.PolyFillRect = tt128PolyFillRect;
      cfb24TEOps1Rect.PolyFillRect = tt128PolyFillRect;
      cfb24NonTEOps1Rect.PolyFillRect = tt128PolyFillRect;
      break;
#endif
    case 32:
      cfb32TEOps.CopyArea = tt128CopyArea;
      cfb32NonTEOps.CopyArea = tt128CopyArea;
      cfb32TEOps1Rect.CopyArea = tt128CopyArea;
      cfb32NonTEOps1Rect.CopyArea = tt128CopyArea;
      cfb32TEOps.PolyFillRect = tt128PolyFillRect;
      cfb32NonTEOps.PolyFillRect = tt128PolyFillRect;
      cfb32TEOps1Rect.PolyFillRect = tt128PolyFillRect;
      cfb32NonTEOps1Rect.PolyFillRect = tt128PolyFillRect;
      break;
  }
}

void tt128BlitterInit(struct vc_mode mode, unsigned char **framebuffer)
{
  int bpp = mode.depth, width = mode.width, height = mode.height, pitch = mode.pitch;

  if(framebuffer) {
    int tt128found = TRUE;

    if(strncmp(mode.name, "IMS", 3))
      tt128found = FALSE;
    else if((mode.disp_reg_address - (mode.fb_address & 0xff000000)) != 0x00800000)
      tt128found = FALSE;
    if(!tt128found) {
      if(enforceTT128) {
	mode.disp_reg_address = (mode.fb_address & 0xff000000) + 0x00800000;
	ErrorF("Warning: TwinTurbo128 base assumed at 0x%08x.\n", mode.disp_reg_address);
      }
      else {
      IMSMMIOBase = 0;
      return;
      }
    }
    ErrorF("TwinTurbo128 detected. Mapping regsiter...\n");
    if (pmacMapDispRegs(&IMSMMIOBase, mode.disp_reg_address, 0x41000)) {
      ErrorF("Warnig: Couldn't map TwinTurbo128. Acceleration disabled.\n");
      IMSMMIOBase = 0;
      return;
    }
    tt128Revision = (regr(SSTATUS) & 0x0f00) >> 8;
    ErrorF("TwinTurbo128 Rev.%d successfully mapped. Starting acceleration...\n", tt128Revision);
    if(!tt128found && enforceTT128) {
      int i;
      unsigned char *rindex;

      DacIndex = IMSMMIOBase + 0x40000;
      DacData =  DacIndex + 4;
      rindex = DacIndex + 12;
      *rindex = 0; eieio();
      for(i = 0; i < 256*3; i++)
	savedHWColormap[i] = *DacData;
    }
  }
  if(!IMSMMIOBase)
    return;

  switch (bpp) {
    case 16:
      vgaBitsPerPixel = 16;
      screen_width = pitch >> 1;
      break;
    case 24:
    case 32:
      vgaBitsPerPixel = 32;
      screen_width = pitch >> 2;
      break;
    default:
      vgaBitsPerPixel = 8;
      screen_width = pitch;
      break;
  }
}
