#include "powermac.h"
#include "../../Xi/stubs.c"
#include <linux/config.h>
#include <linux/mman.h>
#include <asm/vc_ioctl.h>
#include <stdio.h>

int console_fd;

int
pmacConsoleIoctl(unsigned long op, unsigned long data)
{
    if (ioctl(console_fd, op, data) < 0)
	FatalError("ioctl 0x%x failed: %s\n", op, strerror(errno));
}

int
pmacIsMklinux()
{
    int fd;
    fd = open("/proc/osfmach3",O_RDONLY);
    if (fd > 0) {
	close(fd);
	return TRUE;
    }
    return FALSE;
}

void
pmacMonitorSleep( Bool powerdown )
{
    int mode;
#ifndef CONFIG_OSFMACH3
    mode = powerdown ? VESA_POWERDOWN : VESA_NO_BLANKING;
    ioctl( console_fd, VC_POWERMODE, mode );
#endif
}

void
pmacToggleMonitorSleep()
{
    int status;
    Bool gotosleep;
#ifndef CONFIG_OSFMACH3
    status = ioctl( console_fd, VC_POWERMODE, VC_POWERMODE_INQUIRY );
    gotosleep = (status == VESA_NO_BLANKING) ? TRUE : FALSE;
    pmacMonitorSleep( gotosleep );
#endif
}

void
pmacToggleLed( char ledmask )
{
    unsigned char leds = 0;

    ioctl( console_fd, KDGETLED, &leds );
    if (leds & ledmask)
	leds &= ~ledmask;
    else
	leds |= ledmask;
    ioctl( console_fd, KDSETLED, leds );
}

int
pmacMapFrameBuffer(unsigned char **addr, unsigned long phys, unsigned int bytes)
{
    unsigned long pgsize = getpagesize();
    int fd;
    off_t pos;

    fd = 1;
    if ( ioctl( console_fd, VC_FBMAP, &fd) ) {
	if ((fd = open("/dev/mem", O_RDWR)) < 0)
	    FatalError("failed to open /dev/mem: %s\n", strerror(errno));
	pos = (off_t) (phys & -pgsize);
	*addr = (char *) mmap(0, bytes + (phys & (pgsize-1)), PROT_READ|PROT_WRITE, MAP_SHARED, fd, pos);
	close(fd);

	if ((long) *addr == -1)
	    return -1;

	*addr += phys & (pgsize - 1);
	ErrorF("Mapped frame buffer (physaddr 0x%lx size 0x%lx) at %p.\n",
		phys, bytes, *addr);
    } else {
	*addr = phys;
	ErrorF("Mapped frame buffer via BAT at (physaddr 0x%lx size 0x%lx).\n",phys,bytes);
    }
    return 0;
}


int
pmacMapDispRegs(unsigned char **addr, unsigned long phys, unsigned int bytes)
{
    unsigned long pgsize = getpagesize();
    int fd;
    off_t pos;

    pos = (off_t) (phys & -pgsize);
    if ((fd = open("/dev/mem", O_RDWR)) < 0)
      FatalError("failed to open /dev/mem: %s\n", strerror(errno));
    *addr = (char *) mmap(0, bytes + (phys & (pgsize-1)), PROT_READ|PROT_WRITE, MAP_SHARED, fd, pos);
    close(fd);

    if ((long) *addr == -1)
      return -1;
    *addr += phys & (pgsize - 1);
    ErrorF("Mapped physaddr 0x%lx size 0x%lx at %p.\n", phys, bytes, *addr);
    return 0;
}

#if 0 /* perhaps for multi-head support */
int
pmacMapExtraFrameBuffer(unsigned long addr, unsigned int bytes)
{
	if (mmap(addr, bytes, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_ANONYMOUS|MAP_FIXED, 0, 0) != addr) {
		ErrorF("failed to map extra page after frame buffer: %s\n",
		       strerror(errno));
	}
	ErrorF("mapped %ld extra bytes after frame buffer at addr 0x%lx\n",
	       bytes, addr);
}
#endif

#if 0 /* it's now the kernel's job to handle cmaps, mklinux and linuxpmac */
int
pmacMapCmapRegs(unsigned char **a_addr, unsigned char **d_addr,
		unsigned long a_phys, unsigned long d_phys)
{
	int fd;
	unsigned long pgsize = getpagesize();
	off_t pos;
	unsigned char *cmap_regs;

	pos = (off_t) (a_phys & -pgsize);
	if ((fd = open("/dev/mem", O_RDWR)) < 0) {
		FatalError("failed to open /dev/mem: %s\n", strerror(errno));
	}
	cmap_regs = (unsigned char *) mmap(0, pgsize, PROT_READ|PROT_WRITE,
					   MAP_SHARED, fd, pos);
	if ((long) cmap_regs == -1) {
		FatalError("couldn't map colormap regs (%lx): %s\n",
			   a_phys, strerror(errno));
	}
	*a_addr = cmap_regs + (a_phys & (pgsize - 1));

	if ((off_t) (d_phys & -pgsize) != pos) {
		pos = (off_t) (d_phys & -pgsize);
		cmap_regs = (unsigned char *) mmap(0, pgsize, PROT_READ|PROT_WRITE,
						   MAP_SHARED, fd, pos);
		if ((long) cmap_regs == -1) {
			FatalError("couldn't map colormap regs (%lx): %s\n",
				   d_phys, strerror(errno));
		}
	}
	*d_addr = cmap_regs + (d_phys & (pgsize - 1));

	close(fd);
	return 0;
}
#endif
