/* $XFree86: xc/programs/Xserver/hw/xfree86/accel/mach64/mach64im.c,v 3.5 1996/12/23 06:39:18 dawes Exp $ */
/*
 * Copyright 1992,1993,1994,1995,1996 by Kevin E. Martin, Chapel Hill, North Carolina.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Kevin E. Martin not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Kevin E. Martin makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * KEVIN E. MARTIN, RICKARD E. FAITH, AND TIAGO GONS DISCLAIM ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Modified for the Mach-8 by Rickard E. Faith (faith@cs.unc.edu)
 * Modified for the Mach32 by Kevin E. Martin (martin@cs.unc.edu)
 * Modified for the Mach64 by Kevin E. Martin (martin@cs.unc.edu)
 */
/* $XConsortium: mach64im.c /main/4 1996/05/13 16:36:12 kaleb $ */

#include "X.h"
#include "misc.h"
/*#include "xf86.h"*/
/*#include "xf86_HWlib.h"*/
#include "mach64.h"
/*#include "mach64im.h"*/
#define XPMAC_CODE

#if (__BYTE_ORDER == __BIG_ENDIAN)
#define HOST_ENDIAN_CNTL HOST_BIG_ENDIAN_EN
#else
#define HOST_ENDIAN_CNTL 0
#endif

#if (BITMAP_BIT_ORDER == MSBFirst)
#define BYTE_PIX_ORDER  BYTE_ORDER_MSB_TO_LSB
#else
#define BYTE_PIX_ORDER  BYTE_ORDER_LSB_TO_MSB
#endif

#if defined(__powerpc__)
static inline void memcpy_double(char *dst, char *src, int n)
{
  while((unsigned long)dst & 0x07) {
    *dst++ = *src++;
    --n;
    if (n < 1) return;
  }
  n -= 8;
  while(n >= 0) {
    *(double *)dst = *(double *)src;
    n -= 8; src += 8; dst += 8;
  }
  n += 8;
  while (n > 0) {
    *dst++ = *src++;
    --n;
  }
}

/* Someone who knows the way to initiate burst writing */
/* other than 8 byte transfer, please improve me!!     */
static inline void WriteToHost(int *src, int count)
{
  double *host = (double *)(mach64MemReg + HOST_DATA0);
  int i;

  if((unsigned long)src & 0x07) {
    WaitQueue(1);
    *(int *)host = *src++;
    --count;
  }
  while (count & ~0x0f) {
    WaitQueue(16);
    for(i = 0; i < 8; i++) {
      *host = *(double *)src;
      ++((double *)src);
      count -= 2;
    }
  }
  if(count < 1)
    return;
  WaitQueue(count);
  while(count >= 2) {
    *host = *(double *)src;
    ++((double *)src);
    count -= 2;
  }
  if(count)
    *(int *)host = *src;
}

extern int mach64_is_GT;
static void mach64ImageReadVT();
static void mach64ImageWriteVT();
#endif

void	(*mach64ImageReadFunc)();
void	(*mach64ImageWriteFunc)();

static void mach64ImageRead();
static void mach64ImageWrite();
static void mach64ImageWriteHW();

static unsigned long PMask;
static int BytesPerPixel;
static int screenStride;


void
mach64ImageInit()
{
    int i;

#if !defined(XPMAC_CODE)
    PMask = (1UL << mach64InfoRec.depth) - 1;
    BytesPerPixel = mach64InfoRec.bitsPerPixel / 8;
    screenStride = mach64VirtX * BytesPerPixel;
#else
    PMask = (1UL << mach64BitDepth) - 1;
    BytesPerPixel = vgaBitsPerPixel / 8;
    screenStride = mach64VirtX * BytesPerPixel;
#endif
#if defined(__powerpc__)
    if(!mach64_is_GT || vgaBitsPerPixel < 16) {
      mach64ImageReadFunc = mach64ImageReadVT;
      mach64ImageWriteFunc = mach64ImageWriteVT;
    }
    else
#endif
      {
	mach64ImageReadFunc = mach64ImageRead;
	mach64ImageWriteFunc = mach64ImageWrite;
      }
}


static void
mach64ImageWriteHW(x, y, w, h, psrc, pwidth, px, py, alu, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    int			alu;
    unsigned long	planemask;
{
    int wordsPerLine;
    int old_DP_PIX_WIDTH;
#if !defined(__powerpc__)
    register int count;
    register int *pword;
#endif

    WaitQueue(9);
    old_DP_PIX_WIDTH = regr(DP_PIX_WIDTH);
    regw(HOST_CNTL, HOST_ENDIAN_CNTL);
    regw(SC_LEFT_RIGHT, (((x+w-1) << 16) | (x & 0x0000ffff)));
    regw(SC_TOP_BOTTOM, (((y+h-1) << 16) | (y & 0x0000ffff)));
    regw(DP_WRITE_MASK, planemask);
    regw(DP_MIX, (alu << 16) | alu);
    regw(DP_SRC, (MONO_SRC_ONE | FRGD_SRC_HOST | BKGD_SRC_HOST));
#if !defined(XPMAC_CODE)
    switch(mach64InfoRec.bitsPerPixel)
#else
    switch(vgaBitsPerPixel)
#endif
    {
	case 8:
            regw(DP_PIX_WIDTH, (SRC_8BPP | HOST_8BPP | DST_8BPP));
	    break;
	case 16:
            regw(DP_PIX_WIDTH, (SRC_16BPP | HOST_16BPP | DST_16BPP));
	    break;
	case 32:
            regw(DP_PIX_WIDTH, (SRC_32BPP | HOST_32BPP | DST_32BPP));
	    break;
    }
    w = (w + 3) / 4; /* round up to int boundry */
    psrc += pwidth * py + px * BytesPerPixel;
    wordsPerLine = w * BytesPerPixel;
    regw(DST_Y_X, ((x << 16) | (y & 0x0000ffff)));
    regw(DST_HEIGHT_WIDTH, (((w * 4) << 16) | (h & 0x0000ffff)));

    while (h--)
    {
#if defined(__powerpc__)
	WriteToHost((int *)psrc, wordsPerLine);
	psrc += pwidth;
#else
	count = wordsPerLine;
	pword = (int *)psrc;

	while (count >= 16)
	{
            WaitQueue(16);
            regw(HOST_DATAF, *pword++);
            regw(HOST_DATAE, *pword++);
            regw(HOST_DATAD, *pword++);
            regw(HOST_DATAC, *pword++);
            regw(HOST_DATAB, *pword++);
            regw(HOST_DATAA, *pword++);
            regw(HOST_DATA9, *pword++);
            regw(HOST_DATA8, *pword++);
            regw(HOST_DATA7, *pword++);
            regw(HOST_DATA6, *pword++);
            regw(HOST_DATA5, *pword++);
            regw(HOST_DATA4, *pword++);
            regw(HOST_DATA3, *pword++);
            regw(HOST_DATA2, *pword++);
            regw(HOST_DATA1, *pword++);
            regw(HOST_DATA0, *pword++);
	    count -= 16;
	}
	WaitQueue(count);
	switch(count)
	{
	    case 15: regw(HOST_DATAE, *pword++);
	    case 14: regw(HOST_DATAD, *pword++);
	    case 13: regw(HOST_DATAC, *pword++);
	    case 12: regw(HOST_DATAB, *pword++);
	    case 11: regw(HOST_DATAA, *pword++);
	    case 10: regw(HOST_DATA9, *pword++);
	    case  9: regw(HOST_DATA8, *pword++);
	    case  8: regw(HOST_DATA7, *pword++);
	    case  7: regw(HOST_DATA6, *pword++);
	    case  6: regw(HOST_DATA5, *pword++);
	    case  5: regw(HOST_DATA4, *pword++);
	    case  4: regw(HOST_DATA3, *pword++);
	    case  3: regw(HOST_DATA2, *pword++);
	    case  2: regw(HOST_DATA1, *pword++);
	    case  1: regw(HOST_DATA0, *pword);
	    default:
		break;
	}
	psrc += pwidth;
#endif
    }

    WaitQueue(3);
    regw(SC_LEFT_RIGHT, ((mach64MaxX << 16) | 0 ));
    regw(SC_TOP_BOTTOM, ((mach64MaxY << 16) | 0 ));
    regw(DP_PIX_WIDTH, old_DP_PIX_WIDTH);

    WaitIdleEmpty();
}


static void
mach64ImageWrite(x, y, w, h, psrc, pwidth, px, py, alu, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    int			alu;
    unsigned long	planemask;
{
    pointer curvm;
    int byteCount, j;

    if ((w <= 0) || (h <= 0) || alu == MIX_DST)
	return;

    if ((alu != MIX_SRC) || ((planemask & PMask) != PMask)) {
	mach64ImageWriteHW(x, y, w, h, psrc, pwidth, px, py, alu, planemask);
	return;
    }

    psrc += pwidth * py + px * BytesPerPixel;
    curvm = (pointer)((unsigned char *)mach64VideoMem +
	    (x + y * mach64VirtX) * BytesPerPixel);
    byteCount = w * BytesPerPixel;
    WaitIdleEmpty();

    while(h--) {
#if defined(__powerpc__)
      if (!(((unsigned long)curvm ^ (unsigned long)psrc) & 0x07))
	memcpy_double((char *)curvm, (char *)psrc, byteCount);
      else
	memcpy((void *)curvm, psrc, byteCount);
#else
      MemToBus((void *)curvm, psrc, byteCount);
#endif
      curvm = (pointer)((unsigned char *)curvm + screenStride); 
      psrc += pwidth;
    }
}


#if defined(__powerpc__)
static void
mach64ImageWriteVT(x, y, w, h, psrc, pwidth, px, py, alu, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    int			alu;
    unsigned long	planemask;
{
    if ((w <= 0) || (h <= 0) || alu == MIX_DST)
	return;
    mach64ImageWriteHW(x, y, w, h, psrc, pwidth, px, py, alu, planemask);
}
#endif


static void
mach64ImageRead(x, y, w, h, psrc, pwidth, px, py, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    unsigned long	planemask;
{
    int j;
    pointer curvm;

    if ((w <= 0) || (h <= 0))
	return;

    if ((planemask & PMask) != PMask) {
#ifdef NOT_DEFINED
        mach64ImageReadHW(x, y, w, h, psrc, pwidth, px, py, planemask);
	return;
#else
	ErrorF ("mach64ImageRead: unsupported planemask\n");
#endif
    }
    psrc += pwidth * py + px * BytesPerPixel;
    curvm = (pointer)((unsigned char *)mach64VideoMem +
	    (x + y * mach64VirtX) * BytesPerPixel);
    WaitIdleEmpty();

    while(h--) {
#if defined(__powerpc__)
      if (!(((unsigned long)curvm ^ (unsigned long)psrc) & 7))
	memcpy_double((char *)psrc, (char *)curvm, w * BytesPerPixel);
      else
        memcpy(psrc, (void *)curvm, w * BytesPerPixel);
#else
	BusToMem(psrc, (void *)curvm, w * BytesPerPixel);

#endif
	curvm = (pointer)((unsigned char *)curvm + screenStride); 
	psrc += pwidth;
    }
}


#if defined(__powerpc__)
static void
mach64ImageReadVT(x, y, w, h, psrc, pwidth, px, py, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    unsigned long	planemask;
{
    int j;
    pointer curvm;

    if ((w <= 0) || (h <= 0))
	return;

    if ((planemask & PMask) != PMask)
	ErrorF ("mach64ImageRead: unsupported planemask\n");
    psrc += pwidth * py + px * BytesPerPixel;
    curvm = (pointer)((unsigned char *)mach64VideoMem +
	    (x + y * mach64VirtX) * BytesPerPixel);
    WaitIdleEmpty();

    while(h--) {
      switch(BytesPerPixel) {
        case 2:
	  for(j = 0; j < w; j++)
	    ((short *)psrc)[j] = ((short *)curvm)[j];
	  break;
        case 4:
	  for(j = 0; j < w; j++)
	    ((long *)psrc)[j] = ((long *)curvm)[j];
	  break;
        default:
	  for(j = 0; j < w; j++)
	    ((char *)psrc)[j] = ((char *)curvm)[j];
	    break;
	}
	curvm = (pointer)((unsigned char *)curvm + screenStride); 
	psrc += pwidth;
    }
}
#endif


void
mach64ImageStippleFunc(x, y, w, h, psrc, pwidth, px, py, fgPixel, bgPixel, 
		       alu, planemask, opaque)
    int                 x;
    int                 y;
    int                 w;
    int                 h;
    char                *psrc;
    int                 pwidth;
    int                 px;
    int                 py;
    Pixel               fgPixel;
    Pixel               bgPixel;
    int                 alu;
    unsigned long       planemask;
    int 		opaque;
{
    int *pline;
    int wordsPerLine;
    int old_DP_PIX_WIDTH;
#if !defined(__powerpc__)
    int *pword;
    int count;
#endif

    if (alu == MIX_DST || w <= 0 || h <= 0)
	return;

    WaitQueue(10);
    old_DP_PIX_WIDTH = regr(DP_PIX_WIDTH);
    regw(SC_LEFT_RIGHT, (((x+w-1) << 16) | (x & 0x0000ffff)));
    regw(SC_TOP_BOTTOM, (((y+h-1) << 16) | (y & 0x0000ffff)));
    regw(DP_WRITE_MASK, planemask);
    regw(DP_FRGD_CLR, fgPixel);
    regw(DP_BKGD_CLR, bgPixel);
    if( opaque )
        regw(DP_MIX, (alu << 16) | alu);
    else
        regw(DP_MIX, (alu << 16) | MIX_DST);
    regw(DP_SRC, (MONO_SRC_HOST | FRGD_SRC_FRGD_CLR | BKGD_SRC_BKGD_CLR));

#if !defined(XPMAC_CODE)
    switch(mach64InfoRec.bitsPerPixel)
#else
    switch(vgaBitsPerPixel)
#endif
    {
        case 8:
    	    regw(DP_PIX_WIDTH, ( BYTE_PIX_ORDER | SRC_1BPP | HOST_1BPP | DST_8BPP));
            break;
        case 16:
    	    regw(DP_PIX_WIDTH, ( BYTE_PIX_ORDER | SRC_1BPP | HOST_1BPP | DST_16BPP));
            break;
        case 32:
    	    regw(DP_PIX_WIDTH, ( BYTE_PIX_ORDER | SRC_1BPP | HOST_1BPP | DST_32BPP));
            break;
    }
    w = (w + (px % 32) + 31) >> 5; /* round up to int boundry and take into */
				   /* account the pixels skipped over */
    wordsPerLine = pwidth >> 2;
    pline = (int*)psrc + (wordsPerLine * py) + (px >> 5);
    regw(DST_Y_X, (((x - (px % 32)) << 16) | (y & 0x0000ffff)));
    regw(DST_HEIGHT_WIDTH, (((w * 32) << 16) | h));

    while (h--)
    {
#if defined(__powerpc__)
	WriteToHost((int *)pline, w);
	pline += wordsPerLine;
#else
        count = w;
	pword = pline;

        while (count >= 16)
        {
            WaitQueue(16);
            regw(HOST_DATAF, *pword++);
            regw(HOST_DATAE, *pword++);
            regw(HOST_DATAD, *pword++);
            regw(HOST_DATAC, *pword++);
            regw(HOST_DATAB, *pword++);
            regw(HOST_DATAA, *pword++);
            regw(HOST_DATA9, *pword++);
            regw(HOST_DATA8, *pword++);
            regw(HOST_DATA7, *pword++);
            regw(HOST_DATA6, *pword++);
            regw(HOST_DATA5, *pword++);
            regw(HOST_DATA4, *pword++);
            regw(HOST_DATA3, *pword++);
            regw(HOST_DATA2, *pword++);
            regw(HOST_DATA1, *pword++);
            regw(HOST_DATA0, *pword++);
	    count -= 16;
	}

        WaitQueue(16);
        switch(count)
        {
            case 15: regw(HOST_DATAE, *pword++);
            case 14: regw(HOST_DATAD, *pword++);
            case 13: regw(HOST_DATAC, *pword++);
            case 12: regw(HOST_DATAB, *pword++);
            case 11: regw(HOST_DATAA, *pword++);
            case 10: regw(HOST_DATA9, *pword++);
            case  9: regw(HOST_DATA8, *pword++);
            case  8: regw(HOST_DATA7, *pword++);
            case  7: regw(HOST_DATA6, *pword++);
            case  6: regw(HOST_DATA5, *pword++);
            case  5: regw(HOST_DATA4, *pword++);
            case  4: regw(HOST_DATA3, *pword++);
            case  3: regw(HOST_DATA2, *pword++);
            case  2: regw(HOST_DATA1, *pword++);
            case  1: regw(HOST_DATA0, *pword  );
                break;
	}
	pline += wordsPerLine;
#endif
    }

    WaitQueue(3);
    regw(DP_PIX_WIDTH, old_DP_PIX_WIDTH);
    regw(SC_LEFT_RIGHT, ((mach64MaxX << 16) | 0 ));
    regw(SC_TOP_BOTTOM, ((mach64MaxY << 16) | 0 ));

    WaitIdleEmpty();
}
