#include "mach64.h"
#include <asm/vc_ioctl.h>
#include <string.h>
#include <asm/io.h>

int mach64alu[16] = {
    MIX_0,
    MIX_AND,
    MIX_SRC_AND_NOT_DST,
    MIX_SRC,
    MIX_NOT_SRC_AND_DST,
    MIX_DST,
    MIX_XOR,
    MIX_OR,
    MIX_NOR,
    MIX_XNOR,
    MIX_NOT_DST,
    MIX_SRC_OR_NOT_DST,
    MIX_NOT_SRC,
    MIX_NOT_SRC_OR_DST,
    MIX_NAND,
    MIX_1
};

int mach64MaxX, mach64MaxY;
int mach64VirtX, mach64VirtY;
short mach64WeightMask;
int mach64BitDepth;
pointer mach64MemReg = (pointer)0; /* virtual address */
pointer mach64VideoMem = (pointer)0; /* virtual address */
unsigned int mach64LinearOffset;
int enforceMach64 = FALSE;
static int boot_without_driver = FALSE;
extern int vgaBitsPerPixel;
extern int vgaBytesPerPixel;
extern unsigned char *DacIndex;
extern unsigned char *DacData;
extern unsigned char savedHWColormap[];

int mach64_is_GT;
static int mach64GTs[] = {
  MACH64_GT_ID,
  MACH64_GU_ID,
  MACH64_GB_ID,
  MACH64_GD_ID,
  MACH64_GI_ID,
  MACH64_GP_ID,
  MACH64_GQ_ID,
  MACH64_GV_ID,
  MACH64_GW_ID,
  MACH64_GY_ID,
  MACH64_GZ_ID,
  MACH64_LD_ID,
  MACH64_LG_ID,
  MACH64_LB_ID,
  MACH64_LI_ID,
  MACH64_LP_ID
};


static inline int is_mach64gt(void)
{
  int chipid, i;

  chipid = regr(CONFIG_CHIP_ID) & CFG_CHIP_TYPE;
  for(i = 0; i < sizeof(mach64GTs)/sizeof(int); i++)
    if(chipid == mach64GTs[i])
      return(1);
  return(0);
}


void Mach64BlitterInit(struct vc_mode mode, unsigned char **framebuffer)
{
  int bpp = mode.depth, width = mode.width, height = mode.height;
  int BytesPerPixel, pitch = mode.pitch;


  if(framebuffer) {
    int mach64found = TRUE;

    if(strncmp(mode.name, "ATY Mach64", 10))
      mach64found = FALSE;
    else if((mode.disp_reg_address - (mode.fb_address & 0xff000000)) != 0x007ffc00)
      mach64found = FALSE;
    if(!mach64found) {
      if(enforceMach64) {
	mode.disp_reg_address = (mode.fb_address & 0xff000000) + 0x007ffc00;
	ErrorF("Warning: Mach64 base assumed at 0x%08x.\n", mode.disp_reg_address);
	boot_without_driver = TRUE;
      }
      else {
	mach64MemReg = 0;
	return;
      }
    }
    ErrorF("Mach64 detected. Mapping regsiter...\n");
    if (pmacMapDispRegs(&mach64MemReg, mode.disp_reg_address, 0x1000)) {
      ErrorF("Warnig: Couldn't map Mach64. Acceleration disabled.\n");
      mach64MemReg = 0;
      return;
    }
    ErrorF("Mach64 successfully mapped. Starting acceleration...\n");
    mach64VideoMem = *framebuffer;
    if(boot_without_driver) {
      int i;
      unsigned char *rindex;

      DacIndex = mach64MemReg + DAC_W_INDEX;
      DacData = mach64MemReg + DAC_DATA;
      rindex = mach64MemReg + DAC_R_INDEX;
      *rindex = 0; eieio();
      for(i = 0; i < 256*3; i++)
	savedHWColormap[i] = *DacData;
    }
  }
  if(!mach64MemReg)
    return;

  switch (bpp) {
    case 16:
      mach64WeightMask = RGB16_555;
      mach64BitDepth = 15;
      vgaBitsPerPixel = 16;
      vgaBytesPerPixel = 2;
      mach64VirtX = pitch >> 1;
      break;
    case 24:
    case 32:
      mach64BitDepth = 24;
      vgaBitsPerPixel = 32;
      vgaBytesPerPixel = 4;
      mach64VirtX = pitch >> 2;
      break;
    default:
      mach64BitDepth = 8;
      vgaBitsPerPixel = 8;
      vgaBytesPerPixel = 1;
      mach64VirtX = pitch;
      break;
    }
    mach64MaxX = width - 1;
    mach64VirtY = height;
    mach64MaxY = height - 1;
    if(boot_without_driver)
      mach64LinearOffset = regr(CRTC_OFF_PITCH) & 0x000fffff; /* actually offset >> 3 */
    else
      mach64LinearOffset = 0;

    /* Reset drawing engine and bus FIFO */
    regw(GEN_TEST_CNTL, regr(GEN_TEST_CNTL) & ~GUI_ENGINE_ENABLE);
    regw(GEN_TEST_CNTL, regr(GEN_TEST_CNTL) | GUI_ENGINE_ENABLE);
    regw(BUS_CNTL, regr(BUS_CNTL) | BUS_HOST_ERR_ACK | BUS_FIFO_ERR_ACK);

    WaitIdleEmpty();
    regw(DST_OFF_PITCH, ((mach64VirtX >> 3) << 22) | mach64LinearOffset);
    regw(SRC_OFF_PITCH, ((mach64VirtX >> 3) << 22) | mach64LinearOffset);
    regw(CONTEXT_MASK, 0xffffffff);
    regw(DST_Y_X, 0);
    regw(DST_HEIGHT, 0);
    regw(DST_BRES_ERR, 0);
    regw(DST_BRES_INC, 0);
    regw(DST_BRES_DEC, 0);
    regw(DST_CNTL, (DST_X_LEFT_TO_RIGHT | DST_Y_TOP_TO_BOTTOM));
    regw(SRC_Y_X, 0);
    regw(SRC_HEIGHT1_WIDTH1, 0);
    regw(SRC_Y_X_START, 0);
    regw(SRC_HEIGHT2_WIDTH2, 0);
    regw(SRC_CNTL, 0);

    WaitQueue(15);
    regw(HOST_CNTL, 0);
    regw(PAT_REG0, 0);
    regw(PAT_REG1, 0);
    regw(PAT_CNTL, 0);
    regw(SC_LEFT_RIGHT, ((mach64MaxX << 16) | 0 ));
    regw(SC_TOP_BOTTOM, ((mach64MaxY << 16) | 0 ));
    regw(DP_BKGD_CLR, 0);
    regw(DP_FRGD_CLR, 1);
    regw(DP_WRITE_MASK, 0xffffffff);
    regw(DP_MIX, (MIX_SRC << 16) | MIX_DST);
    regw(DP_SRC, FRGD_SRC_FRGD_CLR);
    regw(CLR_CMP_CLR, 0);
    regw(CLR_CMP_MASK, 0xffffffff);
    regw(CLR_CMP_CNTL, 0);
    regw(GUI_TRAJ_CNTL, DST_X_LEFT_TO_RIGHT | DST_Y_TOP_TO_BOTTOM);

    WaitQueue(3);
    switch (vgaBitsPerPixel) {
    case 8:
	regw(DP_PIX_WIDTH, HOST_8BPP | SRC_8BPP | DST_8BPP | BYTE_ORDER_LSB_TO_MSB);
	regw(DP_CHAIN_MASK, DP_CHAIN_8BPP);
	break;
    case 16:
	if (mach64WeightMask == RGB16_555) {
 	    regw(DP_PIX_WIDTH, HOST_16BPP | SRC_15BPP | DST_15BPP | BYTE_ORDER_LSB_TO_MSB);
	    regw(DP_CHAIN_MASK, DP_CHAIN_15BPP);
	} else {
	    regw(DP_PIX_WIDTH, HOST_16BPP | SRC_16BPP | DST_16BPP | BYTE_ORDER_LSB_TO_MSB);
	    regw(DP_CHAIN_MASK, DP_CHAIN_16BPP);
	}
	break;
    case 32:
	regw(DP_PIX_WIDTH, HOST_32BPP | SRC_32BPP | DST_32BPP |BYTE_ORDER_LSB_TO_MSB);
	regw(DP_CHAIN_MASK, DP_CHAIN_32BPP);
	break;
    }
    mach64_is_GT = is_mach64gt();

    WaitIdleEmpty();
#if 0
    ErrorF("BUS_CNTL = 0x%08X\n", regr(BUS_CNTL));
    ErrorF("MEM_CNTL = 0x%08X\n", regr(MEM_CNTL));
    ErrorF("CRTC_GEN_CNTL = 0x%08X\n", regr(CRTC_GEN_CNTL));
#endif
    mach64ImageInit();
}
