/* $XConsortium: mgabitblt.c /main/7 1996/10/28 04:24:32 kaleb $ */
/*
 * MGA Millennium (MGA2064W) with Ti3026 RAMDAC driver v.1.1
 *
 * The driver is written without any chip documentation. All extended ports
 * and registers come from tracing the VESA-ROM functions.
 * The BitBlt Engine comes from tracing the windows BitBlt function.
 *
 * Author:	Radoslaw Kapitan, Tarnow, Poland
 *			kapitan@student.uci.agh.edu.pl
 *		original source
 *
 * Now that MATROX has released documentation to the public, enhancing
 * this driver has become much easier. Nevertheless, this work continues
 * to be based on Radoslaw's original source
 *
 * Contributors:
 *		Andrew Vanderstock, Melbourne, Australia
 *			vanderaj@mail2.svhm.org.au
 *		additions, corrections, cleanups
 *
 *		Dirk Hohndel
 *			hohndel@XFree86.Org
 *		integrated into XFree86-3.1.2Gg
 *		fixed some problems with PCI probing and mapping
 */
 
/* $XFree86: xc/programs/Xserver/hw/xfree86/vga256/drivers/mga/mgabitblt.c,v 3.5 1996/10/19 15:16:20 dawes Exp $ */

#define CONFIG_FB_MATROX
#include	"X.h"
#include	"Xmd.h"
#include	"Xproto.h"
#include	"gcstruct.h"
#include	"windowstr.h"
#include	"scrnintstr.h"
#include	"pixmapstr.h"
#include	"regionstr.h"

#define PSZ 8
#include	"../../cfb/cfb.h"
#include	"cfb16.h"
#if defined(CONFIG_CFB24)
#include	"cfb24.h"
#endif
#include	"cfb32.h"

#include <asm/vc_ioctl.h>
#include <string.h>
#include "mgareg.h"
#include "mga.h"

static void (*stdDoBitbltCopy)();
int MGAScrnWidth;
unsigned char *MGAMMIOBase = 0;
int vgaBitsPerPixel;
int vgaBytesPerPixel;
int MGAydstorg;
int MGAinterleave = 1; /* Millenium II Mac is always interleave*/
int MGAMaxFastBlitY;
int MGABppShift;

extern GCOps cfb32TEOps1Rect, cfb32TEOps, cfb32NonTEOps1Rect, cfb32NonTEOps,
	     cfb24TEOps1Rect, cfb24TEOps, cfb24NonTEOps1Rect, cfb24NonTEOps,
	     cfb16TEOps1Rect, cfb16TEOps, cfb16NonTEOps1Rect, cfb16NonTEOps,
	     cfbTEOps1Rect,   cfbTEOps,   cfbNonTEOps1Rect,   cfbNonTEOps;
int fix4mbBoundary = FALSE;


void MGABlitterInit(struct vc_mode mode, unsigned char **framebuffer)
{
    int bpp = mode.depth, width = mode.width, height = mode.height;
    int shift, pitch = mode.pitch;
    int BytesPerPixel, offset, offset_modulo, ydstorg_modulo;
    unsigned int opmode;
    int maxFastBlitMem =  (MGAinterleave ? 4096 : 2048) * 1024;

    if(framebuffer) {
      if(strncmp(mode.name, "MTRX", 4)) {
	MGAMMIOBase = 0;
	return;
      }
      else if(mode.disp_reg_address < 0x80000000) {
	MGAMMIOBase = 0;
	return;
      }
      ErrorF("Millennium detected. Mapping MGA2x64W regsiter...\n");
      if (pmacMapDispRegs(&MGAMMIOBase, mode.disp_reg_address, 0x4000)) {
	ErrorF("Warnig: Couldn't map MGA2164W register. Acceleration disabled.\n");
	MGAMMIOBase = 0;
	return;
      }
      ErrorF("MGA2164W register successfully mapped. Starting acceleration...\n");
    }
    if(!MGAMMIOBase)
      return;

    switch(bpp) {
      case 16:
        stdDoBitbltCopy = cfb16DoBitbltCopy;
	/* set 16 bpp, turn off dithering, turn on 5:5:5 pixels */
        shift = 1 + (1 << 30) + (1 << 31);
	MGABppShift = 1;
	opmode = 0x00010100;
	vgaBitsPerPixel = 16;
	break;
      case 24:
      case 32:
        stdDoBitbltCopy = cfb32DoBitbltCopy;
        shift = 2;
	MGABppShift = 2;
	opmode = 0x00020200;
	vgaBitsPerPixel = 32;
	break;
      default:
        stdDoBitbltCopy = cfbDoBitbltCopy;
        shift = 0;
	MGABppShift = 0;
	opmode = 0x00000000;
	vgaBitsPerPixel = 8;
	break;
    }
    MGAScrnWidth = pitch >> MGABppShift;
    BytesPerPixel = vgaBitsPerPixel / 8;

    MGAydstorg = (mode.fb_address & 0x001fffff)/BytesPerPixel;
    if(fix4mbBoundary) {
      if (MGAScrnWidth*height*BytesPerPixel > 4*1024*1024) {
	offset = (4*1024*1024) % (width * BytesPerPixel);
	offset_modulo = 4;
	ydstorg_modulo = 64;
	if (vgaBitsPerPixel == 24)
	  offset_modulo *= 3;
	offset_modulo <<= 1;
	ydstorg_modulo <<= 1;
	MGAydstorg = offset / BytesPerPixel;
	while ((offset % offset_modulo) != 0 || (MGAydstorg % ydstorg_modulo) != 0) {
	  offset++;
	  MGAydstorg = offset / BytesPerPixel;
	}
      }
    }
    MGAMaxFastBlitY = maxFastBlitMem/(MGAScrnWidth*BytesPerPixel);
    opmode |= MGAOPM_DMA_BLIT;

    MGAWAITFIFOSLOTS(10);
    MGAREG(MGAREG_FCOL) = 0x00000000;
    MGAREG(MGAREG_SHIFT) = 0x00000000;
    MGAREG(MGAREG_OPMODE) = opmode;
    MGAREG(MGAREG_MACCESS) = shift;
    MGAREG(MGAREG_YDSTORG) = MGAydstorg;
    MGAREG(MGAREG_PLNWT) = 0xFFFFFFFF;
    MGAREG(MGAREG_PITCH) = MGAScrnWidth;
    MGAREG(MGAREG_CXBNDRY) = 0xFFFF0000;  /* (maxX << 16) | minX */
    MGAREG(MGAREG_YTOP) = 0x00000000;  /* minPixelPointer */
    MGAREG(MGAREG_YBOT) = 0x007FFFFF;  /* maxPixelPointer */
}

static void MGABitBlt(widthSrc, xsrc, ysrc, xdst, ydst, w, h, xdir, ydir)
int widthSrc, xsrc, ysrc, xdst, ydst, w, h, xdir, ydir;
{
    long srcStart, srcStop, regSGN;
    
    if(ydir == 1)    /* top to bottom */
    {
        regSGN = 0;
    }
    else             /* bottom to top */
    {
        ysrc += h - 1;
        ydst += h - 1;
        regSGN = 4;
    }
    
    if(xdir == 1)    /* left to right */
    {
        srcStart = ysrc * MGAScrnWidth + xsrc + MGAydstorg;
        srcStop  = srcStart + w - 1;
    }
    else             /* right to left */
    { 
        srcStop  = ysrc * MGAScrnWidth + xsrc + MGAydstorg;
        srcStart = srcStop + w - 1;
        regSGN |= 1;
    }
#if 0
    if( !(((xsrc << MGABppShift) ^ (xdst << MGABppShift)) & 0x7f) &&
	((ysrc + h) <= MGAMaxFastBlitY) && ((ysrc + h) <= MGAMaxFastBlitY) ) {
      MGAWAITFIFOSLOTS(8);
      MGAREG(MGAREG_DWGCTL) = 0x040A600C; /* Do fast blit */
      MGAREG(MGAREG_SGN) = regSGN;
      MGAREG(MGAREG_AR5) = widthSrc;
      MGAREG(MGAREG_AR3) = srcStart;
      MGAREG(MGAREG_AR0) = srcStop;
      MGAREG(MGAREG_FXBNDRY) = ((xdst + w - 1) << 16) | xdst;
      MGAREG(MGAREG_YDSTLEN + MGAREG_EXEC) = (ydst << 16) | h;
      MGAREG(MGAREG_DWGCTL) = 0x040c6008;
    }
    else {
#endif
      MGAWAITFIFOSLOTS(7);
      MGAREG(MGAREG_DWGCTL) = 0x040C4008;
      MGAREG(MGAREG_SGN) = regSGN;
      MGAREG(MGAREG_AR5) = widthSrc;
      MGAREG(MGAREG_AR3) = srcStart;
      MGAREG(MGAREG_AR0) = srcStop;
      MGAREG(MGAREG_FXBNDRY) = ((xdst + w - 1) << 16) | xdst;
      MGAREG(MGAREG_YDSTLEN + MGAREG_EXEC) = (ydst << 16) | h;
#if 0
    }
#endif
#if 0
    if(!MGAWaitForBlitter())
                ErrorF("MGA: BitBlt Engine timeout\n");
#endif
    MGAWAITFREE();
    MGAREG8(MGAREG_CACHEFLUSH) = 0;
}

/*
 * DoBitbltCopy
 *
 * Author: Keith Packard
 * Modifications: Radoslaw Kapitan
 */ 
void MGADoBitbltCopy(pSrc, pDst, alu, prgnDst, pptSrc, planemask)
DrawablePtr	pSrc, pDst;
int		alu;
RegionPtr	prgnDst;
DDXPointPtr	pptSrc;
unsigned long   planemask;
{
  int widthSrc, widthDst;	/* add to get to same position in next line */

  BoxPtr pbox;
  int nbox;
  
  BoxPtr pboxTmp, pboxNext, pboxBase, pboxNew1, pboxNew2;
                                  /* temporaries for shuffling rectangles */
  DDXPointPtr pptTmp, pptNew1, pptNew2;
                                  /* shuffling boxes entails shuffling the
                                     source points too */

  int xdir;			/* 1 = left right, -1 = right left/ */
  int ydir;			/* 1 = top down, -1 = bottom up */
  int careful;

  if(!xf86VTSema ||
  	!(pSrc->type == DRAWABLE_WINDOW && pDst->type == DRAWABLE_WINDOW))
  {
      (*stdDoBitbltCopy)(pSrc, pDst, alu, prgnDst, pptSrc, planemask);
      return;
  }
  
  widthSrc = widthDst = MGAScrnWidth;

  /* XXX we have to err on the side of safety when both are windows,
   * because we don't know if IncludeInferiors is being used.
   */
  careful = ((pSrc == pDst) ||
	     ((pSrc->type == DRAWABLE_WINDOW) &&
	      (pDst->type == DRAWABLE_WINDOW)));
  
  pbox = REGION_RECTS(prgnDst);
  nbox = REGION_NUM_RECTS(prgnDst);
  
  pboxNew1 = NULL;
  pptNew1 = NULL;
  pboxNew2 = NULL;
  pptNew2 = NULL;
  if (careful && (pptSrc->y < pbox->y1))
    {
      /* walk source botttom to top */
      ydir = -1;
      widthSrc = -widthSrc;
      widthDst = -widthDst;
      
      if (nbox > 1)
	{
	  /* keep ordering in each band, reverse order of bands */
	  pboxNew1 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox);
	  if(!pboxNew1)
	    return;
	  pptNew1 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
	  if(!pptNew1)
	    {
	      DEALLOCATE_LOCAL(pboxNew1);
	      return;
	    }
	  pboxBase = pboxNext = pbox+nbox-1;
	  while (pboxBase >= pbox)
	    {
	      while ((pboxNext >= pbox) &&
		     (pboxBase->y1 == pboxNext->y1))
		pboxNext--;
	      pboxTmp = pboxNext+1;
	      pptTmp = pptSrc + (pboxTmp - pbox);
	      while (pboxTmp <= pboxBase)
	        {
		  *pboxNew1++ = *pboxTmp++;
		  *pptNew1++ = *pptTmp++;
	        }
	      pboxBase = pboxNext;
	    }
	  pboxNew1 -= nbox;
	  pbox = pboxNew1;
	  pptNew1 -= nbox;
	  pptSrc = pptNew1;
        }
    }
  else
    {
      /* walk source top to bottom */
      ydir = 1;
    }
  
  if (careful && (pptSrc->x < pbox->x1))
    {
      /* walk source right to left */
      xdir = -1;
      
      if (nbox > 1)
	{
	  /* reverse order of rects in each band */
	  pboxNew2 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox);
	  pptNew2 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
	  if(!pboxNew2 || !pptNew2)
	    {
	      if (pptNew2) DEALLOCATE_LOCAL(pptNew2);
	      if (pboxNew2) DEALLOCATE_LOCAL(pboxNew2);
	      if (pboxNew1)
		{
		  DEALLOCATE_LOCAL(pptNew1);
		  DEALLOCATE_LOCAL(pboxNew1);
		}
	      return;
	    }
	  pboxBase = pboxNext = pbox;
	  while (pboxBase < pbox+nbox)
	    {
	      while ((pboxNext < pbox+nbox) &&
		     (pboxNext->y1 == pboxBase->y1))
		pboxNext++;
	      pboxTmp = pboxNext;
	      pptTmp = pptSrc + (pboxTmp - pbox);
	      while (pboxTmp != pboxBase)
	        {
		  *pboxNew2++ = *--pboxTmp;
		  *pptNew2++ = *--pptTmp;
	        }
	      pboxBase = pboxNext;
	    }
	  pboxNew2 -= nbox;
	  pbox = pboxNew2;
	  pptNew2 -= nbox;
	  pptSrc = pptNew2;
	}
    }
  else
    {
      /* walk source left to right */
      xdir = 1;
    }
  
  while(nbox--) {
    MGABitBlt(widthSrc,
	         pptSrc->x, pptSrc->y, pbox->x1, pbox->y1,
	         pbox->x2 - pbox->x1, pbox->y2 - pbox->y1,
	         xdir, ydir);
    pbox++;
    pptSrc++;
  }
  
  /* free up stuff */
  if (pboxNew2)
    {
      DEALLOCATE_LOCAL(pptNew2);
      DEALLOCATE_LOCAL(pboxNew2);
    }
  if (pboxNew1)
    {
      DEALLOCATE_LOCAL(pptNew1);
      DEALLOCATE_LOCAL(pboxNew1);
    }
#if 0  
  if(!MGAWaitForBlitter())
      ErrorF("MGA: BitBlt Engine timeout\n");
#endif
}


RegionPtr MGACopyArea(pSrcDrawable, pDstDrawable, pGC,
		      srcx, srcy, width, height, dstx, dsty)
DrawablePtr pSrcDrawable, pDstDrawable;
GC *pGC;
int srcx, srcy, width, height, dstx, dsty;
{
  void (*doBitBlt)() = MGADoBitbltCopy;

  switch(pSrcDrawable->bitsPerPixel) {
    case 8:
      if (pGC->alu != GXcopy || (pGC->planemask & 0xFF) != 0xFF) {
	doBitBlt = cfbDoBitbltGeneral;
	if ((pGC->planemask & 0xFF) == 0xFF) {
	  switch (pGC->alu) {
	    case GXxor:
	      doBitBlt = cfbDoBitbltXor;
	      break;
	    case GXor:
	      doBitBlt = cfbDoBitbltOr;
	      break;
	  }
	}
      }
      return cfbBitBlt(pSrcDrawable, pDstDrawable, pGC, 
		       srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
    case 16:
      if (pGC->alu != GXcopy || (pGC->planemask & 0xFFFF) != 0xFFFF) {
	doBitBlt = cfb16DoBitbltGeneral;
	if ((pGC->planemask & 0xFFFF) == 0xFFFF) {
	  switch (pGC->alu) {
	    case GXxor:
	      doBitBlt = cfb16DoBitbltXor;
	      break;
	    case GXor:
	      doBitBlt = cfb16DoBitbltOr;
	      break;
	  }
	}
      }
      return cfb16BitBlt(pSrcDrawable, pDstDrawable, pGC, 
			 srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
    case 24:
    case 32:
      if (pGC->alu != GXcopy || (pGC->planemask & 0xFFFFFFFF) != 0xFFFFFFFF) {
	doBitBlt = cfb32DoBitbltGeneral;
	if ((pGC->planemask & 0xFFFFFFFF) == 0xFFFFFFFF) {
	  switch (pGC->alu) {
	    case GXxor:
	      doBitBlt = cfb32DoBitbltXor;
	      break;
	    case GXor:
	      doBitBlt = cfb32DoBitbltOr;
	      break;
	  }
	}
      }
      return cfb32BitBlt(pSrcDrawable, pDstDrawable, pGC, 
			 srcx, srcy, width, height, dstx, dsty, doBitBlt, 0L);
  }
}

void MGAInitGC(void)
{
  switch(vgaBitsPerPixel) {
    case 8:
      cfbTEOps.CopyArea = MGACopyArea;
      cfbNonTEOps.CopyArea = MGACopyArea;
      cfbTEOps1Rect.CopyArea = MGACopyArea;
      cfbNonTEOps1Rect.CopyArea = MGACopyArea;
      cfbTEOps.PolyFillRect = mgaPolyFillRect;
      cfbNonTEOps.PolyFillRect = mgaPolyFillRect;
      cfbTEOps1Rect.PolyFillRect = mgaPolyFillRect;
      cfbNonTEOps1Rect.PolyFillRect = mgaPolyFillRect;
      break;
    case 16:
      cfb16TEOps.CopyArea = MGACopyArea;
      cfb16NonTEOps.CopyArea = MGACopyArea;
      cfb16TEOps1Rect.CopyArea = MGACopyArea;
      cfb16NonTEOps1Rect.CopyArea = MGACopyArea;
      cfb16TEOps.PolyFillRect = mgaPolyFillRect;
      cfb16NonTEOps.PolyFillRect = mgaPolyFillRect;
      cfb16TEOps1Rect.PolyFillRect = mgaPolyFillRect;
      cfb16NonTEOps1Rect.PolyFillRect = mgaPolyFillRect;
      break;
    case 24:
    case 32:
      cfb32TEOps.CopyArea = MGACopyArea;
      cfb32NonTEOps.CopyArea = MGACopyArea;
      cfb32TEOps1Rect.CopyArea = MGACopyArea;
      cfb32NonTEOps1Rect.CopyArea = MGACopyArea;
      cfb32TEOps.PolyFillRect = mgaPolyFillRect;
      cfb32NonTEOps.PolyFillRect = mgaPolyFillRect;
      cfb32TEOps1Rect.PolyFillRect = mgaPolyFillRect;
      cfb32NonTEOps1Rect.PolyFillRect = mgaPolyFillRect;
      break;
  }
}
