/*****************************************************************************
Copyright 1993 by Apple Computer, Inc, Cupertino, California
			All Rights Reserved

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies.

APPLE MAKES NO WARRANTY OR REPRESENTATION, EITHER EXPRESS,
OR IMPLIED, WITH RESPECT TO THIS SOFTWARE, ITS QUALITY,
PERFORMANCE, MERCHANABILITY, OR FITNESS FOR A PARTICULAR
PURPOSE. AS A RESULT, THIS SOFTWARE IS PROVIDED "AS IS,"
AND YOU THE USER ARE ASSUMING THE ENTIRE RISK AS TO ITS
QUALITY AND PERFORMANCE. IN NO EVENT WILL APPLE BE LIABLE 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES RESULTING FROM ANY DEFECT IN THE SOFTWARE.

THE WARRANTY AND REMEDIES SET FORTH ABOVE ARE EXCLUSIVE
AND IN LIEU OF ALL OTHERS, ORAL OR WRITTEN, EXPRESS OR
IMPLIED.

/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or X Consortium
not be used in advertising or publicity pertaining to 
distribution  of  the software  without specific prior 
written permission. Sun and X Consortium make no 
representations about the suitability of this software for 
any purpose. It is provided "as is" without any express or 
implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

#include "X.h"
#include "Xproto.h"
#include "scrnintstr.h"
#include "colormapst.h"

#define PSZ 8
#include "../../cfb/cfb.h"
#include "../../include/gcstruct.h"
#include "mibstore.h"
#include "mga.h"
#define REGMACH64_H
#define REGRAGE128_H
#include "mach64.h"
#include "ct_driver.h"
#include "rage128.h"

extern Bool cfbCloseScreen();
extern miBSFuncRec cfbBSFuncRec;
extern Bool cfbCreateScreenResources();
extern int colormapSize;
extern int pmacPrivateIndex;

extern void tt128CopyWindow();
extern unsigned char *IMSMMIOBase;


Bool
pmac8ScreenInit(pScreen, pbits, xsize, ysize, dpix, dpiy, width)
    register ScreenPtr pScreen;
    pointer pbits;              /* pointer to screen bitmap */
    int xsize, ysize;           /* in pixels */
    int dpix, dpiy;             /* dots per inch */
    int width;                  /* pixel width of frame buffer */
{
    int i;
    pointer oldDevPrivate;
    VisualPtr   visuals;
    DepthPtr    depths;
    int         nvisuals;
    int         ndepths;
    int         rootdepth;
    VisualID    defaultVisual;
    Bool	cfbresult;

    cfbresult=cfbSetupScreen(pScreen, pbits, xsize, ysize, dpix, dpiy, width);
#ifdef STATIC_COLOR
    pmacColorProcRestore(pScreen);
#endif
    if(!cfbresult)
        return FALSE;

    rootdepth = 0;
    if (!cfbInitVisuals (&visuals, &depths, &nvisuals, &ndepths, &rootdepth, &defaultVisual, 1<<7, 8))
	return FALSE;

    if(mach64MemReg) {
      mach64InitGC();
      pScreen->CopyWindow = mach64CopyWindow;
      pScreen->PaintWindowBackground = mach64PaintWindow;
      pScreen->PaintWindowBorder = mach64PaintWindow;
      pScreen->CreateGC = mach64CreateGC;
    }
    else if(MGAMMIOBase) {
      MGAInitGC();
      pScreen->PaintWindowBackground = mgaPaintWindow;
      pScreen->PaintWindowBorder = mgaPaintWindow;
      pScreen->CopyWindow = mgaCopyWindow;
    }
    else if(IMSMMIOBase) {
      tt128InitGC();
      pScreen->CopyWindow = tt128CopyWindow;
    }
    else if(ctMMIOBase) {
      ctHiQVInitGC();
      pScreen->CopyWindow = ctcfbCopyWindow;
    }
    if(rage128MemReg) {
      rage128InitGC();
      pScreen->CopyWindow = rage128CopyWindow;
      pScreen->PaintWindowBackground = rage128PaintWindow;
      pScreen->PaintWindowBorder = rage128PaintWindow;
      pScreen->CreateGC = rage128CreateGC;
    }

    if (! miScreenInit(pScreen, pbits, xsize, ysize, dpix, dpiy, width,
		       rootdepth, ndepths, depths,
		       defaultVisual, nvisuals, visuals,
		       (miBSFuncPtr) 0))
	return FALSE;
    /* overwrite miCloseScreen with our own */
    pScreen->CloseScreen = cfbCloseScreen;
    /* init backing store here so we can overwrite CloseScreen without stepping
     * on the backing store wrapped version */
    miInitializeBackingStore (pScreen, &cfbBSFuncRec);
    colormapSize = 256;

    return TRUE;
}
