/*-
 * pmacColor.c --
 *	Functions for updating the colormap on the powermac screen.
 */

/*****************************************************************************
Copyright 1988-1993 by Apple Computer, Inc, Cupertino, California
			All Rights Reserved

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies.

APPLE MAKES NO WARRANTY OR REPRESENTATION, EITHER EXPRESS,
OR IMPLIED, WITH RESPECT TO THIS SOFTWARE, ITS QUALITY,
PERFORMANCE, MERCHANABILITY, OR FITNESS FOR A PARTICULAR
PURPOSE. AS A RESULT, THIS SOFTWARE IS PROVIDED "AS IS,"
AND YOU THE USER ARE ASSUMING THE ENTIRE RISK AS TO ITS
QUALITY AND PERFORMANCE. IN NO EVENT WILL APPLE BE LIABLE 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES RESULTING FROM ANY DEFECT IN THE SOFTWARE.

THE WARRANTY AND REMEDIES SET FORTH ABOVE ARE EXCLUSIVE
AND IN LIEU OF ALL OTHERS, ORAL OR WRITTEN, EXPRESS OR
IMPLIED.

*****************************************************************************/
/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,p and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the name of Sun not be used in
advertising or publicity pertaining to distribution  of  the
software  without specific prior written permission. Sun
makes no representations about the suitability of this
software for any purpose. It is provided "as is" without any
express or implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

#include    "powermac.h"

#include    "colormap.h"
#include    "colormapst.h"
#include    "resource.h"
#include    "mac_color.h"
#ifndef	XPMAC_CODE
#include <asm/vc_ioctl.h>
#endif	/* XPMAC_CODE */
#include <asm/io.h>

extern int TellLostMap(), TellGainedMap();
#ifdef STATIC_COLOR
extern void pmacColorProcRestore();
#endif

extern unsigned char *cmap_data;

int colormapSize = 256;
int noMacGamma = FALSE;
struct ColorSpec Map[256];
unsigned char savedHWColormap[256*3];
unsigned char *DacIndex = NULL;
unsigned char *DacData = NULL;

void
SlamHardware(pScreen, Map)
    ScreenPtr	pScreen;
    struct ColorSpec Map[];
{
    int	i;
    unsigned char *p;

    p = cmap_data;
    *(unsigned long *)p = colormapSize * 3;
    p += sizeof (unsigned long);
    i = 0;
    if (noMacGamma)
	for (; i < colormapSize; i++) {
	    *p++ = Map[i].color.red;
	    *p++ = Map[i].color.green;
	    *p++ = Map[i].color.blue;
	}
    else
	for (; i < colormapSize; i++) {
	    *p++ = mac_std_gamma_correction[Map[i].color.red >> 8];
	    *p++ = mac_std_gamma_correction[Map[i].color.green >> 8];
	    *p++ = mac_std_gamma_correction[Map[i].color.blue >> 8];
	}

    if(inXpmac) {
      if(DacIndex && DacData) {
	int shift = 0;

	if(pScreen->rootDepth == 15 || pScreen->rootDepth == 16)
	  shift = 3;
	p = cmap_data + sizeof(unsigned long);
	for(i = 0; i < colormapSize; i++) {
	  *DacIndex = i << shift; eieio();
	  *DacData = *p++; eieio();
	  *DacData = *p++; eieio();
	  *DacData = *p++; eieio();
	}
      } else
	ioctl(console_fd,VC_SETCMAP,cmap_data);
    }
}

static void
pmac24UpdateColormap(pScreen, cmap)
    ScreenPtr	pScreen;
    ColormapPtr	cmap;
{
    register int i;
    register Entry *pent;
    register VisualPtr pVisual = cmap->pVisual;

    for (i = 0; i < colormapSize; i++) {
	Map[i].value = i;
	pent = &cmap->red[i];
	Map[i].color.red = pent->co.local.red;
	pent = &cmap->green[i];
	Map[i].color.green = pent->co.local.green;
	pent = &cmap->blue[i];
	Map[i].color.blue = pent->co.local.blue;
    }
    SlamHardware(pScreen, Map);
}

void
pmacColorUpdateColormap(pScreen, cmap)
    ScreenPtr	pScreen;
    ColormapPtr	cmap;
{
    register int i;
    register Entry *pent;
    register VisualPtr pVisual = cmap->pVisual;

    if ((pVisual->class | DynamicClass) == DirectColor) {
	/* 888 - This is an 8 bpp assumption */
	for (i = 0; i < colormapSize; i++) {
	    Map[i].value = i;
	    pent = &cmap->red[(i & pVisual->redMask) >>
			      pVisual->offsetRed];
	    if (pent->fShared)
		Map[i].color.red = pent->co.shco.red->color;
	    else
		Map[i].color.red = pent->co.local.red;
	    pent = &cmap->green[(i & pVisual->greenMask) >>
				pVisual->offsetGreen];
	    if (pent->fShared)
		Map[i].color.green = pent->co.shco.green->color;
	    else
		Map[i].color.green = pent->co.local.green;
	    pent = &cmap->blue[(i & pVisual->blueMask) >>
			       pVisual->offsetBlue];
	    if (pent->fShared)
		Map[i].color.blue = pent->co.shco.blue->color;
	    else
		Map[i].color.blue = pent->co.local.blue;
	}
    } else {
	for (i = 0, pent = cmap->red;
	     i < pVisual->ColormapEntries;
	     i++, pent++) {
	    if (pent->fShared) {
		Map[i].value = i;
		Map[i].color.red = pent->co.shco.red->color;
		Map[i].color.green = pent->co.shco.green->color;
		Map[i].color.blue = pent->co.shco.blue->color;
	    }
	    else {
		Map[i].value = i;
		Map[i].color.red = pent->co.local.red;
		Map[i].color.green = pent->co.local.green;
		Map[i].color.blue = pent->co.local.blue;
	    }
	}
    }
    SlamHardware(pScreen, Map);
}

/*-
 *-----------------------------------------------------------------------
 * pmacColorSaveScreen --
 *	Preserve the color screen by turning on or off the video
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Video state is switched
 *
 *-----------------------------------------------------------------------
 */
static Bool
pmacColorSaveScreen (pScreen, on)
    ScreenPtr	  pScreen;
    int    	  on;
{
    if (on == SCREEN_SAVER_FORCER)
      SetTimeSinceLastInputEvent ();
    return FALSE;
}

/*-
 *-----------------------------------------------------------------------
 * pmacColorInstallColormap --
 *	Install given colormap.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Existing map is uninstalled.
 *	All clients requesting ColormapNotify are notified
 *
 *-----------------------------------------------------------------------
 */
void
pmacColorInstallColormap(cmap)
    ColormapPtr	cmap;
{
    register ColormapPtr installedMap = 
			 pmacFbs[cmap->pScreen->myNum].installedMap;

    if (cmap == installedMap)
	return;
    if (installedMap)
	WalkTree(installedMap->pScreen, TellLostMap,
		 (char *) &(installedMap->mid));
    WalkTree(cmap->pScreen, TellGainedMap, (char *) &(cmap->mid));
    pmacFbs[cmap->pScreen->myNum].installedMap = cmap;
    if (cmap->pVisual->nplanes > 8)
	pmac24UpdateColormap(cmap->pScreen, cmap);
    else
	pmacColorUpdateColormap(cmap->pScreen, cmap);
}

/*-
 *-----------------------------------------------------------------------
 * pmacColorUninstallColormap --
 *	Uninstall given colormap.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	default map is installed
 *	All clients requesting ColormapNotify are notified
 *
 *-----------------------------------------------------------------------
 */
static void
pmacColorUninstallColormap(cmap)
    ColormapPtr	cmap;
{
    if (cmap == pmacFbs[cmap->pScreen->myNum].installedMap) {
	Colormap defMapID = cmap->pScreen->defColormap;

	if (cmap->mid != defMapID) {
	    ColormapPtr defMap = (ColormapPtr) LookupIDByType(defMapID, RT_COLORMAP);

	    if (defMap)
		(*cmap->pScreen->InstallColormap)(defMap);
	    else
	        ErrorF("pmacColor: Can't find default colormap\n");
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * pmacColorListInstalledColormaps --
 *	Fills in the list with the IDs of the installed maps
 *
 * Results:
 *	Returns the number of IDs in the list
 *
 * Side Effects:
 *	None
 *
 *-----------------------------------------------------------------------
 */
static int
pmacColorListInstalledColormaps(pScreen, pCmapList)
    ScreenPtr	pScreen;
    Colormap	*pCmapList;
{
    *pCmapList = (pmacFbs[pScreen->myNum].installedMap)->mid;
    return (1);
}

/*-
 *-----------------------------------------------------------------------
 * pmacColorStoreColors --
 *	Sets the pixels in pdefs into the specified map.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	None
 *
 *-----------------------------------------------------------------------
 */
static void
pmacColorStoreColors(pmap, ndef, pdefs)
    ColormapPtr	pmap;
    int		ndef;
    xColorItem	*pdefs;
{
    if (pmap == pmacFbs[pmap->pScreen->myNum].installedMap)
	if (pmap->pVisual->nplanes > 8)
	    pmac24UpdateColormap(pmap->pScreen, pmap);
	else
	    pmacColorUpdateColormap(pmap->pScreen, pmap);
}

Bool
pmacCreateColormap(pMap)
    ColormapPtr pMap;
{
    if (pMap->pScreen->rootDepth == 24 && pMap->class == DirectColor) {
	int i;

	for (i = 0; i < 256 ; i++)
	    pMap->red[i].co.local.red = pMap->green[i].co.local.green = pMap->blue[i].co.local.blue = (i << 8) | i;
    }
    return(cfbInitializeColormap(pMap));
}

#ifdef STATIC_COLOR
void
pmacColorProcRestore(pScreen)
    register ScreenPtr	pScreen;
{
    pScreen->InstallColormap = pmacColorInstallColormap;
    pScreen->UninstallColormap = pmacColorUninstallColormap;
    pScreen->ListInstalledColormaps = pmacColorListInstalledColormaps;
    pScreen->StoreColors = pmacColorStoreColors;
}
#endif

/*-
 *-----------------------------------------------------------------------
 * pmacColorInit --
 *	Attempt to find and initialize a color framebuffer 
 *
 * Results:
 *	TRUE if everything went ok. FALSE if not.
 *
 * Side Effects:
 *	Most of the elements of the ScreenRec are filled in. Memory is
 *	allocated for the frame buffer and the buffer is mapped. The
 *	video is enabled for the frame buffer...
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
Bool
pmacColorInit (index, pScreen, argc, argv)
    int	    	  index;    	/* The index of pScreen in the ScreenInfo */
    ScreenPtr	  pScreen;  	/* The Screen to initialize */
    int	    	  argc;	    	/* The number of the Server's arguments. */
    char    	  **argv;   	/* The arguments themselves. Don't change! */
{
    Bool (*InitProc)();
    int depth, bitsPerPixel;
    extern Bool pmac8ScreenInit(),  pmac16ScreenInit();
    extern Bool pmac24ScreenInit(), pmac32ScreenInit();
    extern fbFd pmacFbs[];
    extern int defaultColorVisualClass;
    int i,j;

    depth = pmacFbs[index].defaultDepth;
    bitsPerPixel = depth;

    /* set up procs which aren't handled by cfb */
    /* save screen proc */
    pScreen->SaveScreen = pmacColorSaveScreen;
    /* colormap procs */
    pScreen->InstallColormap = pmacColorInstallColormap;
    pScreen->UninstallColormap = pmacColorUninstallColormap;
    pScreen->ListInstalledColormaps = pmacColorListInstalledColormaps;
    pScreen->StoreColors = pmacColorStoreColors;
    pScreen->CreateColormap = pmacCreateColormap;

    switch (depth) {
	case 8:
	    InitProc = pmac8ScreenInit;
	    defaultColorVisualClass = PseudoColor;
	    bitsPerPixel = 8;
	    break;
	case 15:
	    bitsPerPixel = 16;
	    /* fall through */
	case 16:
	    InitProc = pmac16ScreenInit;
	    defaultColorVisualClass = TrueColor;
	    break;
	case 24:
	    InitProc = pmac24ScreenInit;
	    bitsPerPixel = 32;
 	    defaultColorVisualClass = TrueColor;
	    break;
	case 32:
#ifdef SEPARATE_24
	    InitProc = pmac32ScreenInit;
#else
	    InitProc = pmac24ScreenInit;
#endif
 	    defaultColorVisualClass = TrueColor;
	    break;
	default:
	    FatalError("Encountered bogus depth: %d\n", pScreen->rootDepth);
	    break;
    }

    if (!(*InitProc) (pScreen, pmacFbs[index].fb,
			pmacFbs[index].info.v_right -
			pmacFbs[index].info.v_left,
			pmacFbs[index].info.v_bottom -
			pmacFbs[index].info.v_top,
			pmacFbs[index].info.v_hres /* >> 16 */,
			pmacFbs[index].info.v_vres /* >> 16 */,
			(pmacFbs[index].info.v_rowbytes*8)/bitsPerPixel))
    {
	return (FALSE);
    }


    /* make sure the color map gets installed */

    pmacFbs[index].installedMap = NULL;

    (void) pmacColorSaveScreen( pScreen, SCREEN_SAVER_OFF );

    j = pmacScreenInit(pScreen);
    i = cfbCreateDefColormap(pScreen);

    return(i&&j);
}
