/*****************************************************************************
Copyright 1988-1993 by Apple Computer, Inc., Cupertino, CA
			All Rights Reserved

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies.

APPLE MAKES NO WARRANTY OR REPRESENTATION, EITHER EXPRESS,
OR IMPLIED, WITH RESPECT TO THIS SOFTWARE, ITS QUALITY,
PERFORMANCE, MERCHANABILITY, OR FITNESS FOR A PARTICULAR
PURPOSE. AS A RESULT, THIS SOFTWARE IS PROVIDED "AS IS,"
AND YOU THE USER ARE ASSUMING THE ENTIRE RISK AS TO ITS
QUALITY AND PERFORMANCE. IN NO EVENT WILL APPLE BE LIABLE 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES RESULTING FROM ANY DEFECT IN THE SOFTWARE.

THE WARRANTY AND REMEDIES SET FORTH ABOVE ARE EXCLUSIVE
AND IN LIEU OF ALL OTHERS, ORAL OR WRITTEN, EXPRESS OR
IMPLIED.

*****************************************************************************/
/*-
 * pmacInitS.c --
 *	Initialization functions for screen/keyboard/mouse, etc.
 *
 * Copyright (c) 1987 by the Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 */

/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or X Consortium
not be used in advertising or publicity pertaining to 
distribution  of  the software  without specific prior 
written permission. Sun and X Consortium make no 
representations about the suitability of this software for 
any purpose. It is provided "as is" without any express or 
implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

#include    "powermac.h"
#include    <servermd.h>
#include    "dixstruct.h"
#include    "dix.h"
#include    "opaque.h"
#include    "mipointer.h"
#include    "osdep.h"
#include    "Xpoll.h"
#include    <asm/vc_ioctl.h>
#include    <linux/vt.h>

#include <stdio.h>
#ifdef XTESTEXT1
#include "keysym.h"
#endif /* XTESTEXT1 */

#include <asm/ioctls.h>

#define XPMAC_RELEASE "April 10, 2000"

extern int pmacMouseProc();
extern int pmacKbdProc();
extern int pmacKbdSetUp();

extern Bool pmacMonoInit();
extern Bool pmacColorInit();

extern Bool pmacSlotProbe();

extern void SetInputCheck();

void pmacWakeup(pointer, int, pointer);
void pmacBlock(pointer, OSTimePtr, pointer);
void pmacIntroMessage();


extern fd_set EnabledDevices;
extern int noVTswitch;

#ifdef __STDC__
static long GetDevBase (int);
#endif

DevicePtr pPointerDevice, pKeyboardDevice;

extern int inXpmac;

extern void pmacMouseEnqueueEvent();

PtrPrivRec 	sysMousePriv = {
    pmacMouseEnqueueEvent,	/* Function to process an event */
    (ScreenPtr)NULL,			/* Screen pointer is on */
    -1,                         /* fd starts at -1 */
    "/dev/usbmouse",            /* device to read from */
    (drvr_proc *)mdrvr_ps2                   /* driver to process ps2 events */ 
};

static int autoRepeatHandlersInstalled; /* FALSE each time InitOutput called */

/*
 * NUMSCREENS is the number of supported frame buffers (i.e. the number of
 * structures in pmacFbData which have an actual probeProc).
 */
#define NUMSCREENS 1

/* stores key codes for mouse buttons */
int	middleKeyTrigger = DEF_MIDDLE_KEY_TRIGGER;
int	rightKeyTrigger = DEF_RIGHT_KEY_TRIGGER;

/* stores the keyboard type */
char	pmacKbdType[16];

fbFd	pmacFbs[NUMSCREENS];  /* Space for descriptors of open frame buffers */
unsigned int fbSize;
int	vtno = -1, vtOrig;
void	*offscreenFb;
int	keepTty = FALSE;
int     noAccel = FALSE;
extern  int enforceMach64;
extern  int enforceTT128;
extern  int enforceChips65;

static PixmapFormatRec	formats[] = {
    1, 1, BITMAP_SCANLINE_PAD,	/* 1-bit deep */
/*    2, 2, BITMAP_SCANLINE_PAD,	/* 2-bit deep */
/*    4, 4, BITMAP_SCANLINE_PAD,	/* 4-bit deep */
    8, 8, BITMAP_SCANLINE_PAD,		/* 8-bit deep */
    15, 16, BITMAP_SCANLINE_PAD,	/* 16-bit deep */
    24, 32, BITMAP_SCANLINE_PAD,	/* 24-bit deep */
/*    32, 32, BITMAP_SCANLINE_PAD,	/* 32-bit deep */
};
#define NUMFORMATS	(sizeof formats)/(sizeof formats[0])

#define NUMSLOTS 1
#define NUMMODES 7		/* 1,2,4,8,16,24,32 */
static struct legalModeStruct {	/* only 1 and 8 bits are supported */
    int depth;
    int legal;
} legalModes[NUMMODES] = {{1,1},{2,0},{4,0},{8,1},{16,1},{24,1},{32,1}};

static int screenDepth[MAXSCREENS] = {0,0,0,0,0,0};

static struct videoModeStruct{
	int depth[NUMMODES];
        int mode[NUMMODES];   
	struct video_data info[NUMMODES];
} videoModes[NUMSLOTS] = {
	{{0,0,0,0,0,0,0}, {0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}};
#include <asm/page.h>
unsigned char *cmap_data;

static int req_video_mode;		/* screen resolution requested by user */
static int req_depth;			/* screen depth requested by user */
struct vc_mode mode;

#include <asm/io.h>
extern unsigned char savedHWColormap[];
extern unsigned char *DacIndex;
extern unsigned char *DacData;

/*-
 *-----------------------------------------------------------------------
 * SigIOHandler --
 *      Signal handler for SIGIO - mouse input is available.
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static void SigIOHandler(sig)
    int         sig;
{
    int olderrno = errno;
    (*sysMousePriv.driver)(sysMousePriv.fd);
    errno = olderrno;
}
 


static inline int restoreHWColormap(void)
{
  int i;

  if(DacIndex && DacData) {
    *DacIndex = 0; eieio();
    for(i = 0; i < 256*3; i++)
      *DacData = savedHWColormap[i];
  }
}

/*-
 *-----------------------------------------------------------------------
 * InitOutput --
 *	Initialize screenInfo for all actually accessible framebuffers.
 *
 * Results:
 *	screenInfo init proc field set
 *
 * Side Effects:
 *	None
 *
 *-----------------------------------------------------------------------
 */

void
InitOutput(pScreenInfo, argc, argv)
    ScreenInfo 	  *pScreenInfo;
    int     	  argc;
    char    	  **argv;
{
    int     	  i, index, ac = argc;
    char	  **av = argv;

    pScreenInfo->imageByteOrder = IMAGE_BYTE_ORDER;
    pScreenInfo->bitmapScanlineUnit = BITMAP_SCANLINE_UNIT;
    pScreenInfo->bitmapScanlinePad = BITMAP_SCANLINE_PAD;
    pScreenInfo->bitmapBitOrder = BITMAP_BIT_ORDER;

    pScreenInfo->numPixmapFormats = NUMFORMATS;
    for (i=0; i< NUMFORMATS; i++)
        pScreenInfo->formats[i] = formats[i];

    autoRepeatHandlersInstalled = FALSE;

    ParseDepths(argc, argv);

    if (pmacScreenProbe(pScreenInfo, argc, argv) == FALSE)
	FatalError("Can't find any displays\n");

    pScreenInfo->numScreens = 1;

    RegisterBlockAndWakeupHandlers(pmacBlock, pmacWakeup, (void *)0);
}

static Bool
IsLegalDepth(depth)
int depth;
{
    int i;

    for (i=0; i< NUMMODES; i++) {
	if ((legalModes[i].legal) && (legalModes[i].depth == depth))
	    return TRUE;
    }
    return FALSE;
}

ParseDepths(argc, argv)
    int		argc;
    char 	**argv;
{
#define SCREEN_LO 	0
#define SCREEN_HI 	(MAXSCREENS - 1)
#define EQSTRING(s1,s2)	(!(strcmp((s1), (s2))))

    int i;
    int screen = -1;
    int depth = -1;
    int supportedDepth = 0;

    i = 0;
    while (i < argc) {
	if (EQSTRING(argv[i], "-screen")) {
	    if (i <= argc - 4) {
	        screen = atoi(argv[i+1]);
                if ((screen < SCREEN_LO) || (screen > SCREEN_HI)) {
		    ErrorF("Invalid screen number %d; a legal number is %d < screen < %d\n", 
		        screen, SCREEN_LO, SCREEN_HI);
                }

		if (!EQSTRING(argv[i+2], "-depth")) {
		    goto usage;
		}

	        /* is this a depth supported by the server? */
		depth = atoi(argv[i+3]);
		if (IsLegalDepth(depth)) {
		    screenDepth[screen] = depth;
		    supportedDepth = 1;
		}
		if (!supportedDepth) {
		    ErrorF("Invalid screen depth specified for screen %d: %d. Using default.\n", screen, depth);
		}
	    }
	    else {
usage:
		ErrorF("Usage: -screen screennum -depth depthnum\n");
	    }
	    i += 4;
	}
	else {
            i++;
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * InitInput --
 *	Initialize all supported input devices...what else is there
 *	besides pointer and keyboard?
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Two DeviceRec's are allocated and registered as the system pointer
 *	and keyboard devices.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
void
InitInput(argc, argv)
    int     	  argc;
    char    	  **argv;
{
    DevicePtr p, k;
    static int  zero = 0;
    
    p = AddInputDevice(pmacMouseProc, TRUE);
    k = AddInputDevice(pmacKbdProc, TRUE);
#ifdef XTESTEXT1
    xtest_command_key = XK_Clear + MIN_KEYCODE; /* Had to set my own key to
						   ensure it was available
						   on our keyboards. mjb */
#endif

    pPointerDevice = p;
    pKeyboardDevice = k;

    RegisterPointerDevice(p);
    RegisterKeyboardDevice(k);
    miRegisterPointerDevice(screenInfo.screens[0], p);
    {
       struct sigaction act;
       sigemptyset(&act.sa_mask);
       sigaddset(&act.sa_mask, SIGIO);
       act.sa_flags = SA_RESTART;
       act.sa_handler = SigIOHandler;
       sigaction(SIGIO, &act, 0);
    }

#define SET_FLOW(fd) fcntl(fd, F_SETFL, FNDELAY | FASYNC)
#define WANT_SIGNALS(fd) fcntl(fd, F_SETOWN, getpid())

    if (sysMousePriv.fd >= 0) {
        if (SET_FLOW(sysMousePriv.fd) == -1 || 
                 WANT_SIGNALS(sysMousePriv.fd) == -1)
        {
            (void) close (sysMousePriv.fd);
            sysMousePriv.fd = -1;
            ErrorF("Async mouse I/O failed in InitInput");
        }
    }

    if (!mieqInit (k, p))
	FatalError ("could not initialize event queue");
    SetTimeSinceLastInputEvent ();
}

/*-
 *-----------------------------------------------------------------------
 * pmacScreenProbe --
 *	Attempt to find and initialize a framebuffer. 
 *	Grab a fresh virtual console.
 *
 * Results:
 *	TRUE if everything went ok. FALSE if not.
 *
 * Side Effects:
 *	Memory is allocated for the frame buffer and the buffer is mapped.
 *
 *-----------------------------------------------------------------------
 */
Bool
pmacScreenProbe(pScreenInfo, argc, argv)
    ScreenInfo	  *pScreenInfo;	/* The screenInfo struct */
    int	    	  argc;	    	/* The number of the Server's arguments. */
    char    	  **argv;   	/* The arguments themselves. Don't change! */
{
    unsigned int	i;
    unsigned char	*addr;
    struct vt_consize	vc_size;
    struct vt_mode	vt;
    extern void pmacVTRequest(int);

    if (serverGeneration == 1) {
	int fd;
	char vtname[32];
	struct vt_stat vts;

	/* check that we're setuid root */
	if (geteuid() != 0)
	    FatalError("Xpmac: server must be setuid-root\n");

	if ((fd = open("/dev/tty0", O_WRONLY)) < 0)
	    FatalError("failed to open /dev/tty0: %s\n", strerror(errno));
	
	if (vtno == -1) /* find free console? */
	    if ((ioctl(fd, VT_OPENQRY, &vtno) < 0) || (vtno == -1))
		FatalError("Cannot find a free VT\n");

	/* find original vt */
	if (ioctl(fd, VT_GETSTATE, &vts) == 0)
	    vtOrig = vts.v_active;

	close(fd);

	if (keepTty || noVTswitch)
	    vtno = vtOrig;
	else
	    setpgrp();

	if (!keepTty) /* detach! */
	    if ((fd = open("/dev/tty",O_RDWR)) >= 0)
	    {
		ioctl(fd, TIOCNOTTY, 0);
		close(fd);
	    }

	/* open our vt */
	sprintf(vtname,"/dev/tty%d",vtno);
	if ((console_fd = open(vtname, O_RDWR|O_NDELAY, 0)) < 0)
	    FatalError("Cannot open %s (%s)\n", vtname, strerror(errno));

	/* setup console switching */
	if (!noVTswitch) {
	    extern int vtRequestPending;

	    /* grab the new console */
	    if (ioctl(console_fd, VT_ACTIVATE, vtno) != 0)
		ErrorF("VT_ACTIVATE failed\n");
	    if (ioctl(console_fd, VT_WAITACTIVE, vtno) != 0)
		ErrorF("VT_WAITACTIVE failed\n");

	    vtRequestPending = FALSE;
	    signal(SIGUSR1, pmacVTRequest);

	    pmacConsoleIoctl(VT_GETMODE, (unsigned long) &vt);
	    vt.mode = VT_PROCESS;
	    vt.relsig = SIGUSR1;
	    vt.acqsig = SIGUSR1;
	    pmacConsoleIoctl(VT_SETMODE, (unsigned long) &vt);

	}

	/* Change the console ownership to the real UID so xterm -C works */
	chown("/dev/tty0", getuid(), getgid());
	chown(vtname, getuid(), getgid());

	/* get current video mode */
	pmacConsoleIoctl(VC_GETMODE, (unsigned long) &mode);

	/* always reset screen. should fail gracefully under MkLinux */
	if (req_video_mode)
	    mode.mode = req_video_mode;
	if (req_depth)
	    mode.depth = req_depth;

	if (pmacIsMklinux())
	    mode.mode = 0;
	else {
	    if (ioctl(console_fd, VC_SETMODE, (unsigned long) &mode) < 0)
		mode.mode = 0;
	    else /* fetch new values for x, y resolution etc. */
	    pmacConsoleIoctl(VC_GETMODE, (unsigned long) &mode);
	}

	/* set the text console to the new mode too */
	memset(&vc_size, 0, sizeof(vc_size));
	vc_size.v_vlin = mode.height;
	vc_size.v_rows = mode.height / 16;
	vc_size.v_vcol = mode.width;
	vc_size.v_cols = mode.width / 8;
	if (ioctl(console_fd,VT_RESIZEX, (unsigned long) &vc_size) < 0)
	    ErrorF("Couldn't change console screen size: %s\n",
		    strerror(errno));
	/* clear and reset screen */
	ErrorF("\033[H\033[J");
	fflush(stdout);

	pmacIntroMessage();

	pmacFbs[0].slot = 0;
	pmacFbs[0].installedMap = NULL;

	/* set things up for default mode */
	pmacFbs[0].defaultDepth = mode.depth;
	pmacFbs[0].info.v_pixelsize = mode.depth;
	pmacFbs[0].info.v_rowbytes = mode.pitch;
	pmacFbs[0].info.v_right = mode.width;
	pmacFbs[0].info.v_left = pmacFbs[0].info.v_top = 0;
	pmacFbs[0].info.v_bottom = mode.height;

	pmacFbs[0].info.v_hres = pmacFbs[0].info.v_vres = 75;

	fbSize = mode.height * mode.pitch + 0x10;

	/* assume aligned */
	cmap_data = malloc(PAGE_SIZE);
	if (!cmap_data)
	    FatalError("Couldn't allocate colormap memory\n");
	if(pmacMapFrameBuffer(&addr, mode.fb_address, fbSize))
	    FatalError("Couldn't map frame buffer: %s\n", strerror(errno));
	if(!noAccel) {
	  MGABlitterInit(mode, &addr);
	  tt128BlitterInit(mode, &addr);
	  ctHiQVBlitterInit(mode, &addr);
	  Mach64BlitterInit(mode, &addr);
	  Rage128BlitterInit(mode, &addr);
	}
	pmacFbs[0].fb = (pointer) addr;

	if (!noVTswitch) {
	    extern char *savedFb;

	    offscreenFb = malloc(fbSize);
	    if (!offscreenFb) {
		ErrorF("Cannot allocate offscreen buffer, "
		    "switching disabled\n");
		noVTswitch = TRUE;
	    }
	    savedFb = addr;
	}
    } else {
	/* what should we do when we get here? */
    }

#if 0 /* we don't touch colormaps under linux-pmac anymore */
    if (mode.cmap_adr_address != 0) {
	if (pmacMapCmapRegs(&cmap_adr, &cmap_data, mode.cmap_adr_address,
			    mode.cmap_data_address))
	    FatalError("Could not map colormap registers\n");
    }
#ifndef	XPMAC_CODE
    else {
	cmap_adr = (unsigned char *) 0;
	cmap_data = cmap_fake;
    }
#endif	/* XPMAC_CODE */
#endif /* 0 */

    pmacBlackScreen(0);

    switch (pmacFbs[0].defaultDepth) {
#if 0 /* no mono support in linux-pmac nor mklinux */
    case 1:
	i = AddScreen(pmacMonoInit, argc, argv);
	break;
#endif
    case 8:
    case 16:
    case 24:
    case 32:
	i = AddScreen(pmacColorInit, argc, argv);
	break;

    default:
	FatalError("Encountered bogus depth: %d", pmacFbs[0].defaultDepth);
	break;
    }

    return TRUE;
}

/*-
 *-------------------------------------------------------------%---------
 * pmacScreenInit --
 *	Things which must be done for all types of frame buffers...
 *	Should be called last of all.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The graphics context for the screen is created. The CreateGC,
 *	CreateWindow and ChangeWindowAttributes vectors are changed in
 *	the screen structure.
 *
 *	Both a BlockHandler and a WakeupHandler are installed for the
 *	first screen.  Together, these handlers implement autorepeat
 *	keystrokes on the pmac.
 *
 *-----------------------------------------------------------------------
 */
Bool
pmacScreenInit (pScreen)
    ScreenPtr	  pScreen;
{
    extern void   pmacBlockHandler();
    extern void   pmacWakeupHandler();
    static ScreenPtr autoRepeatScreen;
    extern miPointerScreenFuncRec   pmacPointerCursorFuncs;

    /*
     *	Block/Unblock handlers
     */
    if (autoRepeatHandlersInstalled == FALSE) {
	autoRepeatScreen = pScreen;
	autoRepeatHandlersInstalled = TRUE;
    }

    if (pScreen == autoRepeatScreen) {
        pScreen->BlockHandler = pmacBlockHandler;
        pScreen->WakeupHandler = pmacWakeupHandler;
    }

    miDCInitialize (pScreen, &pmacPointerCursorFuncs);

    return TRUE; 
}

/*-
 *-----------------------------------------------------------------------
 * pmacBlackScreen --
 *    Fill a frame buffer with the black pixel.
 *
 * Results:
 *    None
 *
 * Side Effects:
 *
 *-----------------------------------------------------------------------
 */
int
pmacBlackScreen(index)
      int index;
{
    fbFd *pf;
    register unsigned char* fb;
    register int fbinc, line, lw;
    register unsigned int *fbt;
    short pad, shift, success;
    long pixFill;

    pf = &pmacFbs[index];
    fb = pf->fb; /* Assumed longword aligned! */

    switch (pf->info.v_pixelsize) {
#if 0
    case 1:
	pad = 31;
	shift = 5;
	pixFill = 0xffffffff;
	success = 1;
	break;
#endif
    case 8:
	pad = 3;
	shift = 2;
	pixFill = 0x01010101;
	success = 1;
	break;
    case 16:
	pad = 1;
	shift = 1;
	pixFill = 0x0421;
	success = 1;
	break;
    case 24:
    case 32:
	pad = 0;
	shift = 0;
	pixFill = 0x010101;
	success = 1;
	break;
    default:
	ErrorF("Bad depth in pmacBlackScreen.\n");
	break;
    }
    if (success) {
	fbinc = pf->info.v_rowbytes;
	for (line = pf->info.v_top; line < pf->info.v_bottom; line++) {
	    fbt = (unsigned int *)fb;
	    lw = ((pf->info.v_right - pf->info.v_left) + pad) >> shift;
	    do {
		*fbt++ = pixFill;
	    } while (--lw);
	    fb += fbinc;
	}
    }
}

/* Called by GiveUp(). */
void
ddxGiveUp()
{
    if (console_fd > 0) {
	int off = 0;

	/* call ioctl ourselves to ignore errors */
	if (!noVTswitch) {
	    struct vt_mode vt;

	    ioctl(console_fd, VT_GETMODE, (unsigned long) &vt);
	    vt.mode = VT_AUTO;
	    ioctl(console_fd, VT_SETMODE, (unsigned long) &vt);
	    ioctl(console_fd, VT_ACTIVATE, vtOrig );
	}

	if ( mode.mode )
	    ioctl(console_fd, VC_FBMAP, (unsigned long) &off);
	ioctl(console_fd, KDSETMODE, KD_TEXT);
	ioctl(console_fd, KDSKBMODE, K_XLATE);
	write(console_fd, "\033]R", 3);	/* reset palette */
	close(console_fd);
	console_fd = -1;
	restoreHWColormap();
    }
}

void
AbortDDX()
{
    signal(SIGIO,SIG_IGN);
    if (sysMousePriv.fd > 0) {
        close(sysMousePriv.fd);
        sysMousePriv.fd = -1;
    }

#if 0
    if (console_fd > 0) {
	pmacKbdSetUp(console_fd, FALSE); /* Must NOT FatalError() anywhere! */
	close(console_fd);
	console_fd = -1;
    }
#endif
    ddxGiveUp();
}

int
ddxProcessArgument (argc, argv, i)
    int       argc;
    char *argv[];
    int       i;
{
    extern int noOptionPower;
    extern int noOptionMouse;
    extern int noMonitorSleep;
    extern int noMacGamma;
    extern int fix4mbBoundary;


    if (strcmp(argv[i], "-mode") == 0) {
	if (++i < argc) {
	    req_video_mode = atoi(argv[i]);
	    if (req_video_mode <= 0 || req_video_mode > MAX_VIDEO_MODE)
		UseMsg();
	} else
	    UseMsg();
	return 2;
    }
    else if (strcmp(argv[i], "-depth") == 0) {
	if (++i < argc) {
	    req_depth = atoi(argv[i]);
	    if (req_depth != 8 && req_depth != 16
		&& req_depth != 24 && req_depth != 32)
		UseMsg();
	} else
	    UseMsg();
	return 2;
    }
    else if (strcmp(argv[i], "-nooptionpower") == 0) {
	noOptionPower = TRUE;
	return 1;
    }
    else if (strcmp(argv[i], "-nooptionmouse") == 0) {
	noOptionMouse = TRUE;
	return 1;
    }
    else if (strcmp(argv[i], "-nomonitorsleep") == 0) {
	noMonitorSleep = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-middlekey" ) == 0) {
	if(++i < argc) {
		middleKeyTrigger = atoi( argv[i] );
		if (middleKeyTrigger == 0)
		    UseMsg();
		middleKeyTrigger -= MIN_KEYCODE;
	} else
	    UseMsg();
	return 2;
    }
    else if ( strcmp ( argv[i], "-rightkey" ) == 0) {
	if(++i < argc) {
		rightKeyTrigger = atoi( argv[i] );
		if (rightKeyTrigger == 0)
		    UseMsg();
		rightKeyTrigger -= MIN_KEYCODE;
	} else
	    UseMsg();
	return 2;
    }
    else if ( strcmp ( argv[i], "-kbdtype" ) == 0) {
	if(++i < argc) {
	    strcpy(pmacKbdType, argv[i]);
	} else
	    UseMsg();
	return 2;
    }
    else if ( strcmp ( argv[i], "-nomacgamma") == 0) {
	noMacGamma = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-novtswitch") == 0) {
	noVTswitch = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-keeptty") == 0) {
	keepTty = TRUE;
	return 1;
    }
    else if ((argv[i][0] == 'v') && (argv[i][1] == 't')) {
	if (sscanf(argv[i], "vt%2d", &vtno) == 0) {
	    vtno = -1;
	    return 0;
	}
	return 1;
    }
    else if ( strcmp ( argv[i], "-fix4mb") == 0) {
	fix4mbBoundary = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-mach64") == 0) {
	enforceMach64 = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-tt128") == 0) {
	enforceTT128 = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-ct6555") == 0) {
	enforceChips65 = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-novideodriver") == 0) {
	enforceMach64 = TRUE;
	return 1;
    }
    else if ( strcmp ( argv[i], "-noaccel") == 0) {
	noAccel = TRUE;
	return 1;
    }
    else return 0;
}

void
ddxUseMsg()
{
    ErrorF("Xpmac options:\n");
    ErrorF("-mode #                set screen to video mode {1-20}\n");
    ErrorF("-depth #               run screen at depth {8,16,24}\n");
    ErrorF("-nooptionpower         disable using <option>-<power> to kill X\n");
    ErrorF("-nooptionmouse         *don't* require option key\n");
    ErrorF("                       for simulated mouse buttons\n");
    ErrorF("-nomonitorsleep        disable using <power> to enable monitor sleep\n");
    ErrorF("-middlekey #           decimal keycode for middle mouse button\n");
    ErrorF("-rightkey #            decimal keycode for right mouse button\n");
    ErrorF("-kbdtype string        specify the keyboard type\n");
    ErrorF("-nomacgamma            turn off Macintosh gamma correction\n");
    ErrorF("-novtswitch            turn off virtual terminal switching\n");
    ErrorF("-keeptty               don't detach controlling tty\n");
    ErrorF("-fix4mb                process 4MB boundary for Matrox MGA2x64\n");
    ErrorF("-tt128                 enforce TwinTurbo128 acceleration without video driver\n");
    ErrorF("-ct6555                enforce C&T 6555x acceleration without video driver\n");
    ErrorF("-mach64                enforce Mach64 acceleration without video driver\n");
    ErrorF("-novideodriver         same as -mach64\n");
    ErrorF("-noaccel               disable HW acceleration\n");
    ErrorF("vtXX                   use the specified VT number\n");
}

void
MessageF(s)
char *s;
{
    ErrorF(s);
}


/*
 * The block and wakeup routines are copied
 * from the XFree86 code
 */

void
pmacBlock(blockData, pTimeout, pReadmask)
     pointer blockData;
     OSTimePtr pTimeout;
     pointer  pReadmask;
{
}


/* ARGSUSED */
void
pmacWakeup(blockData, err, pReadmask)
     pointer blockData;
     int err;
     pointer pReadmask;
{
  fd_set devicesWithInput;

  if ((int)err >= 0) {
    XFD_ANDSET((fd_set *)&devicesWithInput, (fd_set *)pReadmask,
	       (fd_set *)&EnabledDevices);
    if (XFD_ANYSET((fd_set *)&devicesWithInput))
      {
	    pmacEnqueueEvents ();
      }
  }

  /* don't forget about vt switching if still pending */
  {
    extern int vtRequestPending;
    extern void pmacVTSwitch();
  
    if (vtRequestPending) pmacVTSwitch();
  }
}

#ifdef DPMSExtension
/*
 * DPMSSet --
 *      Device dependent DPMS mode setting hook.  This is called whenever
 *      the DPMS mode is to be changed.
 */
void
DPMSSet(CARD16 level)
{
    return; /* We don't do anything cause we really don't support it.. */
}
 
/*
 * DPMSSupported --
 *      Return TRUE if any screen supports DPMS.
 *
 * We're not currently going to support DPMS in the Xpmac server..
 */
Bool
DPMSSupported(void)
{
    return FALSE;
}
#endif /* DPMSExtension */

void
OsVendorInit()
{
    static int inited;
    if (!inited) {
        sysMousePriv.fd = open (sysMousePriv.device, O_RDONLY);
        if (sysMousePriv.fd < 0)
            ErrorF("Cannot open usb mouse device, error %d\n", errno);
        inited = 1;
    }
}

void
OsVendorFatalError()
{
}

void pmacIntroMessage()
{
    ErrorF("Xpmac / X Window System (X11R6.3)\n"
	    "Release Date: " XPMAC_RELEASE
	    "\n\nRunning in mode=%d, depth=%d; req mode=%d depth=%d\n"
	    "Using VT number %d\n",
	    mode.mode, mode.depth, req_video_mode, req_depth, vtno);
}

/************************************************************************/
/*      Mouse drivers                                                   */
/************************************************************************/

void
mdrvr_ps2( int fd )     /* 0b00yx0321, dx, -dy, [dz] ; x/y = sign of dx/dy
, 1=pressed */
{
        DevicePtr   pPointer=LookupPointerDevice();
        signed char buf[3];
        signed char me[3];
        static signed char lst_bval=0;
        int ret = 0;

        if (inXpmac) {

           buf[0] = 0; buf[1] = 0; buf[2] = 0;
           ret = read(fd, buf, 3);
           if (ret == 3 ) {

#if 0
ErrorF(" Raw buf[0] is %x, buf[1] is %x, buf[2] is %x\n",buf[0],buf[1],buf[2]);
ErrorF(" Last Button Value was %x\n",lst_bval);
#endif

                buf[0] = buf[0] & 0x07;
                if (buf[0] & 0x01) buf[0] = 1;
                if (buf[0] & 0x02) buf[0] = 2;
                if (buf[0] & 0x04) buf[0] = 4;
                me[0] = 0;
                me[1] = (-buf[2]) & 0x7f; 
                me[2] = buf[1];
                if ((buf[0] != 0) && (lst_bval == 0)) {
                     me[0] = buf[0];
                }
                if ((lst_bval != 0) && (buf[0] == 0)) {
                     me[0] = lst_bval;
                     me[1] = me[1] | 0x80;
                }
                lst_bval = buf[0];

#if 0
ErrorF(" final me[0] is %x, me[1] is %x, me[2] is %x\n",me[0],me[1],me[2]);
#endif

                (*sysMousePriv.EnqueueEvent)(pPointer,&me[0],USB_MOUSE);

           } else {
	     ErrorF("read from mouse only returned %d\n",ret);
	   }
	}
}

