/*-
 * pmacIo.c --
 *	Functions to handle input from the keyboard and mouse.
 *
 * Copyright (c) 1987 by the Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 */

/****************************************************************
Copyright 1988-1993 by Apple Computer, Inc, Cupertino, California
			All Rights Reserved

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies.

APPLE MAKES NO WARRANTY OR REPRESENTATION, EITHER EXPRESS,
OR IMPLIED, WITH RESPECT TO THIS SOFTWARE, ITS QUALITY,
PERFORMANCE, MERCHANABILITY, OR FITNESS FOR A PARTICULAR
PURPOSE. AS A RESULT, THIS SOFTWARE IS PROVIDED "AS IS,"
AND YOU THE USER ARE ASSUMING THE ENTIRE RISK AS TO ITS
QUALITY AND PERFORMANCE. IN NO EVENT WILL APPLE BE LIABLE 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES RESULTING FROM ANY DEFECT IN THE SOFTWARE.

THE WARRANTY AND REMEDIES SET FORTH ABOVE ARE EXCLUSIVE
AND IN LIEU OF ALL OTHERS, ORAL OR WRITTEN, EXPRESS OR
IMPLIED.

****************************************************************/

/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or X Consortium
not be used in advertising or publicity pertaining to 
distribution  of  the software  without specific prior 
written permission. Sun and X Consortium make no 
representations about the suitability of this software for 
any purpose. It is provided "as is" without any express or 
implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/


#include <linux/vt.h>
#include <asm/vc_ioctl.h>
#include <signal.h>
#include "powermac.h"
#include "opaque.h"
#include "colormapst.h"
#include "pixmapstr.h"
#include <asm/io.h>

#define	INPBUFSIZE		3*64
#define IS_POWER_KEY(x)		(KEY_DETAIL(x) == 0x7f)
#define IS_COMMAND_KEY(x)	(KEY_DETAIL(x) == 0x37)
#define IS_OPTION_KEY(x)	(KEY_DETAIL(x) == 0x3a)
#define IS_RIGHT_OPTION_KEY(x)	(KEY_DETAIL(x) == 0x7c)
#define IS_CAPS_LOCK_KEY(x)	(KEY_DETAIL(x) == 0x39)
#define IS_NUM_LOCK_KEY(x)	(KEY_DETAIL(x) == 0x47)
#define IS_SCROLL_LOCK_KEY(x)	(KEY_DETAIL(x) == 0x6b)
#define IS_LEFT_CTRL_KEY(X)	(KEY_DETAIL(X) == 0x36)
#define IS_RIGHT_CTRL_KEY(X)	(KEY_DETAIL(X) == 0x7d)

#define KEY_F1 0x7a
#define KEY_F2 0x78
#define KEY_F3 0x63
#define KEY_F4 0x76
#define KEY_F5 0x60
#define KEY_F6 0x61
#define KEY_F7 0x62
#define KEY_F8 0x64
#define KEY_F9 0x65
#define KEY_F10 0x6d
#define KEY_F11 0x67
#define KEY_F12 0x6f

extern int	   screenIsSaved;
extern void	   SaveScreens();
extern void	   pmacToggleMonitorSleep();
extern void	   pmacToggleLed( int ledmask );
extern int	   middleKeyTrigger, rightKeyTrigger;
extern PtrPrivRec  sysMousePriv;

void pmacVTSwitch();

int		lastEventTime = 0;
struct timeval	lastEventTimeTv;
int		noOptionMouse = FALSE;
int		noOptionPower = FALSE;
int		noMonitorSleep = FALSE;
int		noVTswitch = FALSE;
int		inXpmac = TRUE;
int		vtRequestPending;
char		*savedFb;
int		pmacPrivateIndex;
PixmapPtr	ppix;
struct		ColorSpec *oldMap;

extern unsigned char savedHWColormap[];
extern unsigned char *DacIndex;
extern unsigned char *DacData;

static inline int restoreHWColormap(void)
{
  int i;

  if(DacIndex && DacData) {
    *DacIndex = 0; eieio();
    for(i = 0; i < 256*3; i++)
      *DacData = savedHWColormap[i];
  }
}

/*-
 *-----------------------------------------------------------------------
 * TimeSinceLastInputEvent --
 *	Function used for screensaver purposes by the os module.
 *
 * Results:
 *	The time in milliseconds since there last was any
 *	input.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
int
TimeSinceLastInputEvent()
{
    struct timeval	now;

    gettimeofday (&now, (struct timezone *)0);

    if (lastEventTime == 0) {
	lastEventTimeTv = now;
	lastEventTime = TVTOMILLI(now);
    }
    return TVTOMILLI(now) - lastEventTime;
}

/*-
 *-----------------------------------------------------------------------
 * ProcessInputEvents --
 *	Retrieve all waiting input events and pass them to DIX in their
 *	correct chronological order. Only reads from the system pointer
 *	and keyboard.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Events are passed to the DIX layer.
 *
 *-----------------------------------------------------------------------
 */
void
ProcessInputEvents ()
{
    mieqProcessInputEvents ();
    miPointerUpdate ();
}

/*
 *-----------------------------------------------------------------------
 * pmacEnqueueEvents
 *	When a SIGIO is received, read device hard events and
 *	enqueue them using the mi event queue
 */
pmacEnqueueEvents ()
{
    DevicePtr			pPointer;
    DevicePtr			pKeyboard;
    register PtrPrivPtr		ptrPriv;
    register KbPrivPtr		kbdPriv;
    unsigned char		pmacevents[INPBUFSIZE];
    register unsigned char	*me, *meL;
    static int			optionKeyUp = 1, rightOptionKeyUp = 1;
    static int			commandKeyUp = 1, lcontrolKeyUp = 1, rcontrolKeyUp = 1;
    int				n, i;
    char			leds;

    enum {
        NoneYet, Pntr, Kbd
    }                       lastType = NoneYet; /* Type of last event */

    pPointer = LookupPointerDevice();
    pKeyboard = LookupKeyboardDevice();
    if (!pPointer->on || !pKeyboard->on)
	return;
    ptrPriv = (PtrPrivPtr)pPointer->devicePrivate;
    kbdPriv = (KbPrivPtr)pKeyboard->devicePrivate;

    while ((n = read (console_fd, pmacevents,
		      INPBUFSIZE*sizeof pmacevents[0])) > 0)
    {
	meL = pmacevents + (n/(sizeof pmacevents[0]));

	for (me = pmacevents; me < meL; me++)
	{
	    gettimeofday (&lastEventTimeTv, (struct timezone *)0);
	    lastEventTime = TVTOMILLI(lastEventTimeTv);

	    if (IS_OPTION_KEY(*me))
	    	optionKeyUp = KEY_UP(*me);
	    if (IS_RIGHT_OPTION_KEY(*me))
		rightOptionKeyUp = KEY_UP(*me);
	    if (IS_COMMAND_KEY(*me))
		commandKeyUp = KEY_UP(*me);
	    if (IS_LEFT_CTRL_KEY(*me))
		lcontrolKeyUp = KEY_UP(*me);
	    if (IS_RIGHT_CTRL_KEY(*me))
		rcontrolKeyUp = KEY_UP(*me);
            /*
             * Kill the X server with the <option>-<power> combo
	     * Shutdown the monitor with <power>
             */

            if (IS_POWER_KEY(*me)) {
		if (!(optionKeyUp && rightOptionKeyUp) && !noOptionPower)
		   GiveUp(0);
		if (!(KEY_UP(*me)) && !noMonitorSleep &&
		   (optionKeyUp && rightOptionKeyUp))
		   pmacToggleMonitorSleep();
	    }

	    /*
             * Modify the keyboard leds
             */

            if (IS_CAPS_LOCK_KEY(*me)) {
		pmacToggleLed( LED_CAP );
            }
            if (IS_NUM_LOCK_KEY(*me) && !KEY_UP(*me)) {
		pmacToggleLed( LED_NUM );
            }
            if (IS_SCROLL_LOCK_KEY(*me) && !KEY_UP(*me)) {
		pmacToggleLed( LED_SCR );
            }

	    /*
	     * Patch up the event if the option key is {up,down} so as
	     * to allow the simulation to mouse keys and still allow
	     * all keys to be accessable.
	     *
	     * This is done by checking to see if the correct option
	     * behavior is being exhibited (set by the -nooptionmouse flag)
	     * and if so, change the key code being passed in to the
	     * key code for the respective mouse key.
	     */

	    if (IS_TRIGGER_KEY(*me)) {
	    	int keyUp = KEY_UP(*me);
		int optionDown = !(optionKeyUp && rightOptionKeyUp);
		static int middleDown, rightDown;

	    	if (IS_RIGHT_TRIGGER_KEY(*me)) {
		    if (optionDown != noOptionMouse || keyUp && rightDown) {
			*me = keyUp | RIGHT_KEY_CODE;
			rightDown = !keyUp;
		    }
	    	} else {
		    if (optionDown != noOptionMouse || keyUp && middleDown) {
			*me = keyUp | MIDDLE_KEY_CODE;
			middleDown = !keyUp;
		    }
		}
	    }

	    /*
	     * Handle virtual console switching
	     */
	    if (vtRequestPending) {
  	        unsigned char c;
                pmacVTSwitch();
  	        rcontrolKeyUp = lcontrolKeyUp = 1;
	        // simulate key-ups
	        c = 0xb6;
	        (* kbdPriv->EnqueueEvent) (pKeyboard,&c);
	        c = 0xfd;
	        (* kbdPriv->EnqueueEvent) (pKeyboard,&c);
	        lastType = Kbd;
	        continue;
            }

	    if (KEY_DETAIL(*me) == MOUSE_ESCAPE) { 
		/* 1/10/96 - MEB, make sure all three bytes are there */
		while (me+3 > meL) {
		    if ((n = read(console_fd, meL, me+3 - meL)) <= 0)
			break;
		    meL += n;
		}

		me[0] = (optionKeyUp? 0: 1) + (rightOptionKeyUp? 0: 2);
    	    	(* ptrPriv->EnqueueEvent) (pPointer,me,ADB_MOUSE);
			
	    	me += 2;
	    	lastType = Pntr;
	    }

	    else if (IS_MOUSE_KEY(*me))
            {
    	    	(* ptrPriv->EnqueueEvent) (pPointer,me,ADB_MOUSE);
	    	lastType = Pntr;
	    }

	    else {
		int vt;
		vt = 0;

		switch (KEY_DETAIL(*me)) {
		    case KEY_F1: vt = 1; break;
		    case KEY_F2: vt = 2; break;
		    case KEY_F3: vt = 3; break;
		    case KEY_F4: vt = 4; break;
		    case KEY_F5: vt = 5; break;
		    case KEY_F6: vt = 6; break;
		    case KEY_F7: vt = 7; break;
		    case KEY_F8: vt = 8; break;
		    case KEY_F9: vt = 9; break;
		    case KEY_F10: vt = 10; break;
		    case KEY_F11: vt = 11; break;
		    case KEY_F12: vt = 12; break;
		    default: break;
		}

		if ( ! ((lcontrolKeyUp && rcontrolKeyUp) || commandKeyUp) ) {
		    if (vt && !noVTswitch) {
			i = ioctl(console_fd, VT_ACTIVATE, vt);
		    } else {
			if (*me==0x33)
			    GiveUp(0);
			else {
			    (* kbdPriv->EnqueueEvent) (pKeyboard,me);
			    lastType = Kbd;
			}
		    }
		} else {
		    (* kbdPriv->EnqueueEvent) (pKeyboard,me);
		    lastType = Kbd;
		}
            }
	}
    }
}


void
pmacVTSwitch() {
    extern unsigned int fbSize;
    extern void * offscreenFb;
    ColormapPtr cmap;
    ScreenPtr pScreen;
    PixmapPtr psix;
    cmap = pmacFbs[0].installedMap;
    pScreen = cmap->pScreen;

    switch (pmacFbs[0].defaultDepth) {
	 case 8:
	         psix = (PixmapPtr) pScreen->devPrivate;
		 break;
	 case 16:
		 psix = (PixmapPtr) pScreen->devPrivates[pmacPrivateIndex].ptr;
		 break;
	 case 32:
		 psix = (PixmapPtr) pScreen->devPrivates[pmacPrivateIndex].ptr;
		 break;
    }

    if (inXpmac) {
	 inXpmac = FALSE;
	  // need to save framebuffer
	  bcopy(savedFb,offscreenFb,fbSize);
	  psix->devPrivate.ptr = offscreenFb;
	  ioctl(console_fd, VT_RELDISP, 1);
	  vtRequestPending = FALSE;
	  restoreHWColormap();
    } else {
	  extern struct ColorSpec Map[];
	  extern struct vc_mode mode;

	  // restore Xpmac
	  if (ioctl(console_fd, VT_RELDISP, 2)<0)
		    FatalError("Error restoring display\n");
	  inXpmac = TRUE;

	  if (mode.mode)
		    ioctl(console_fd, VC_SETMODE, (unsigned long) &mode);

	  MGABlitterInit(mode, 0);
	  tt128BlitterInit(mode, 0);
	  ctHiQVBlitterInit(mode, 0);
	  Mach64BlitterInit(mode, 0);
	  Rage128BlitterInit(mode, 0);

	  psix->devPrivate.ptr = savedFb;
	  bcopy(offscreenFb,savedFb,fbSize);
	  SlamHardware(pScreen,Map);
	  vtRequestPending = FALSE;
     }
} 




/*-
 *-----------------------------------------------------------------------
 * SetTimeSinceLastInputEvent --
 *	Set the lastEventTime to now.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	lastEventTime is altered.
 *
 *-----------------------------------------------------------------------
 */
void
SetTimeSinceLastInputEvent()
{
    gettimeofday (&lastEventTimeTv, (struct timezone *)0);
    lastEventTime = TVTOMILLI(lastEventTimeTv);
}

CARD32
GetTimeInMillis()
{
    struct timeval  tp;

    gettimeofday(&tp, 0);
    return(tp.tv_sec * 1000) + (tp.tv_usec / 1000);
}

/* signal handler for virtual terminal switching */
void pmacVTRequest(int sig)
{
    signal(sig, (void(*)())pmacVTRequest);
    vtRequestPending = TRUE;
}
