/* $XConsortium: pmacKbd.c,v 1.24 94/04/17 20:29:29 dpw Exp $ */
/*****************************************************************************
Copyright 1988-1993 by Apple Computer, Inc, Cupertino, California
			All Rights Reserved

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies.

APPLE MAKES NO WARRANTY OR REPRESENTATION, EITHER EXPRESS,
OR IMPLIED, WITH RESPECT TO THIS SOFTWARE, ITS QUALITY,
PERFORMANCE, MERCHANABILITY, OR FITNESS FOR A PARTICULAR
PURPOSE. AS A RESULT, THIS SOFTWARE IS PROVIDED "AS IS,"
AND YOU THE USER ARE ASSUMING THE ENTIRE RISK AS TO ITS
QUALITY AND PERFORMANCE. IN NO EVENT WILL APPLE BE LIABLE 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES RESULTING FROM ANY DEFECT IN THE SOFTWARE.

THE WARRANTY AND REMEDIES SET FORTH ABOVE ARE EXCLUSIVE
AND IN LIEU OF ALL OTHERS, ORAL OR WRITTEN, EXPRESS OR
IMPLIED.

*****************************************************************************/
/*-
 * pmacKbd.c --
 *	Functions for retrieving data from a keyboard.
 *
 * Copyright (c) 1987 by the Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */

/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or X Consortium
not be used in advertising or publicity pertaining to 
distribution  of  the software  without specific prior 
written permission. Sun and X Consortium make no 
representations about the suitability of this software for 
any purpose. It is provided "as is" without any express or 
implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/


#define NEED_EVENTS
#include "powermac.h"
#include <stdio.h>
#include "Xproto.h"
#include "inputstr.h"
#include <linux/kd.h>

#ifdef XKB
#include <X11/extensions/XKB.h>
#include <X11/extensions/XKBstr.h>
#include <X11/extensions/XKBsrv.h>
#endif

extern CARD8 *pmacModMap[];
extern KeySymsRec pmacKeySyms[];
extern struct timeval lastEventTimeTv;

extern char pmacKbdType[];

static void 	  pmacBell();
static void 	  pmacKbdCtrl();
void 	  	  pmacKbdEnqueueEvent();
static int	  autoRepeatKeyDown = 0;
static int  	  autoRepeatDebug = 0;
static int	  autoRepeatReady;
static int	  autoRepeatFirst;
static struct timeval autoRepeatLastKeyDownTv;
static struct timeval autoRepeatDeltaTv;
static KeybdCtrl  sysKbCtrl;
static int bell_pitch, bell_duration;


static KbPrivRec  	sysKbPriv = {
    -1,				/* Type of keyboard */
    pmacKbdEnqueueEvent,	/* Function to process an event */
    0,				/* offset for device keycodes */
    &sysKbCtrl,                 /* Initial full duration = .20 sec. */
};

#ifdef XTESTEXT1
#define XTestSERVER_SIDE
#include "xtestext1.h"
/*
 * defined in xtestext1di.c
 */
extern int	on_steal_input;
extern Bool	XTestStealKeyData();
extern short	xtest_mousex, xtest_mousey;
#endif /* XTESTEXT1 */


/*-
 *-----------------------------------------------------------------------
 * (Copied from the SUN port)
 * pmacInitKbdNames --
 *	Handle the XKB initialization
 *
 * Results:
 *	None.
 *
 * Comments: 
 *     This function needs considerable work, in conjunctions with
 *     the need to add geometry descriptions of Macintosh Keyboards.
 *     It would also be nice to have #defines for all the keyboard
 *     layouts so that we don't have to have these hard-coded
 *     numbers.
 *
 *-----------------------------------------------------------------------
 */
#ifdef XKB
#if NeedFunctionPrototypes
static void pmacInitKbdNames (
    XkbComponentNamesRec* names,
    KbPrivPtr pKbd)
#else
static void pmacInitKbdNames (names, pKbd)
    XkbComponentNamesRec* names;
    KbPrivPtr pKbd;
#endif
{
#ifndef XKBBUFSIZE
#define XKBBUFSIZE 64
#endif
    static char keycodesbuf[XKBBUFSIZE];
    static char geometrybuf[XKBBUFSIZE];
    static char  symbolsbuf[XKBBUFSIZE];

    names->keymap = NULL;
    names->compat = "compat/complete";
    names->types  = "types/complete";
    names->keycodes = keycodesbuf;
    names->geometry = geometrybuf;
    names->symbols = symbolsbuf;
    (void) strcpy (keycodesbuf, "keycodes/");
    (void) strcpy (geometrybuf, "geometry/");
    (void) strcpy (symbolsbuf, "symbols/macintosh/");

    (void) strcat (names->keycodes, "macintosh");
    (void) strcat (names->geometry, "macintosh");

    if (strlen(pmacKbdType) > 0) {
        (void) strcat (names->symbols, pmacKbdType);
    } else {
        (void) strcat (names->symbols, "us(extended)");
    }
}
#endif /* XKB */

/*-
 *-----------------------------------------------------------------------
 * pmacKbdProc --
 *	Handle the initialization, etc. of a keyboard.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *
 * Note:
 *	When using linux-pmac, all input comes off a single fd, stored in the
 *	global pmacConsoleFd.  Therefore, only one device should be enabled and
 *	disabled, even though the application still sees both mouse and
 *	keyboard.  We have arbitrarily chosen to enable and disable pmacConsoleFd
 *	in the keyboard routine pmacKbdProc rather than in pmacMouseProc.
 *
 *-----------------------------------------------------------------------
 */

int
pmacKbdProc (pKeyboard, what)
    DevicePtr	  pKeyboard;	/* Keyboard to manipulate */
    int	    	  what;	    	/* What to do to it */
{
    switch (what) {
	case DEVICE_INIT:
	    pKeyboard->on = FALSE;
	    if (pKeyboard != LookupKeyboardDevice()) {
		ErrorF ("Cannot open non-system keyboard. \r\n");
		return (!Success);
	    }
#if 0 /* /dev/console could point to a serial port.. bad */
	    if (pmacConsoleFd <= 0) {
		if ((pmacConsoleFd = open("/dev/console", O_RDWR|O_NDELAY)) < 0)
			FatalError("Could not open /dev/console. \r\n");
            }
#endif
	    /* pmacConsoleFd has been replaced with the global console_fd */
	    /* pmacConsoleFd = console_fd; */

	    /*
	     * Perform final initialization of the system private keyboard
	     * structure and fill in various slots in the device record
	     * itself which couldn't be filled in before.
	     */
	    pKeyboard->devicePrivate = (pointer)&sysKbPriv;
	    sysKbCtrl = defaultKeyboardControl;
	    sysKbPriv.ctrl = &sysKbCtrl;

	    /*
	     * ensure that the keycodes on the wire are >= MIN_KEYCODE
	     */

	    sysKbPriv.type = KBTYPE_MACII;  

	    if (pmacKeySyms[sysKbPriv.type].minKeyCode < MIN_KEYCODE) {
		int offset = MIN_KEYCODE -pmacKeySyms[sysKbPriv.type].minKeyCode;

		pmacKeySyms[sysKbPriv.type].minKeyCode += offset;
		pmacKeySyms[sysKbPriv.type].maxKeyCode += offset;
		sysKbPriv.offset = offset;
	    }
#ifdef	XKB
	    if (noXkbExtension) {
#endif
	    	InitKeyboardDeviceStruct(
			    pKeyboard,
			    &(pmacKeySyms[sysKbPriv.type]),
			    (pmacModMap[sysKbPriv.type]),
			    pmacBell,
			    pmacKbdCtrl);
#ifdef	XKB
	} else {
		XkbComponentNamesRec names;
		pmacInitKbdNames(&names, &sysKbPriv);
		XkbInitKeyboardDeviceStruct((DeviceIntPtr) pKeyboard, &names,
			&pmacKeySyms[sysKbPriv.type],
			pmacModMap[sysKbPriv.type],
			pmacBell, pmacKbdCtrl);
	}
#endif

	    break;

	case DEVICE_ON:
	    pmacKbdSetUp(console_fd, TRUE);

	    AddEnabledDevice(console_fd);

	    pKeyboard->on = TRUE;
	    break;

	case DEVICE_CLOSE:
	    pmacKbdSetUp(console_fd, FALSE);

	    RemoveEnabledDevice(console_fd);

	    pKeyboard->on = FALSE;

#if 0
	    close(pmacConsoleFd);
	    pmacConsoleFd = -1;
#endif
	    break;

	case DEVICE_OFF:
	    pmacKbdSetUp(console_fd, FALSE);

	    RemoveEnabledDevice(console_fd);

	    pKeyboard->on = FALSE;
	    break;

    }
    return (Success);
}

#include <asm/termios.h>

int
pmacKbdSetUp(fd, openClose)
    int		fd;
    Bool	openClose;
{
    struct termios tio;
    static struct termios save_tio;
    int iarg;
    char leds;

    if (openClose) {
	if (ioctl(fd, KDSETMODE, KD_GRAPHICS) < 0)
		FatalError("failed to set graphics mode on console\n");
	if (ioctl(fd, KDSKBMODE, K_RAW) < 0) 
		FatalError("failed to set raw keyboard mode on console\n");

	if (tcgetattr(fd, &save_tio) < 0) {
		FatalError("Failed to ioctl TCGETS. %d", fd);
		return (!Success);
	}

	tio.c_iflag = (IGNPAR | IGNBRK) & (~PARMRK) & (~ISTRIP);
	tio.c_oflag = tio.c_lflag = 0;
	tio.c_cflag = CREAD | CS8 | B9600;
	tio.c_cc[VTIME]=0;
	tio.c_cc[VMIN]=1;

	if (tcsetattr(fd, TCSANOW, &tio) < 0) {
		FatalError("Failed to ioctl TCSETA. %i\r\n", errno);
		return (!Success);
	}

    } else {
	ioctl(fd, KDSETMODE, KD_TEXT);
	ioctl(fd, KDSKBMODE, K_XLATE);

	if (tcsetattr(fd, TCSANOW, &save_tio) < 0) {
		MessageF("Failed to ioctl TCSETA. %i\r\n", errno);
	}
	tcsetattr(fd, TCSAFLUSH, &tio);
    }
}

/*-
 *-----------------------------------------------------------------------
 * pmacBell --
 *	Ring the terminal/keyboard bell
 *
 * Results:
 *    Ring the keyboard bell for an amount of time proportional to
 *    "loudness."
 *
 * Side Effects:
 *	None, really...
 *
 *-----------------------------------------------------------------------
 */
static void pmacBell(int percent, DeviceIntPtr pKeyboard, pointer ctrl,
                      int unused)
{
  if (percent && bell_pitch)
    ioctl(console_fd, KDMKTONE, ((1193190 / bell_pitch) & 0xffff) |
          (((unsigned long) bell_duration * percent / 50) << 16));
}

/*-
 *-----------------------------------------------------------------------
 * pmacKbdCtrl --
 *	Alter some of the keyboard control parameters
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Some...
 *
 *-----------------------------------------------------------------------
 */
static void
pmacKbdCtrl (pKeyboard, ctrl)
    DevicePtr	  pKeyboard;	    /* Keyboard to alter */
    KeybdCtrl     *ctrl;
{
    KbPrivPtr   pPriv = (KbPrivPtr) pKeyboard->devicePrivate;
    *pPriv->ctrl = *ctrl;

    bell_duration = ctrl->bell_duration;
    bell_pitch    = ctrl->bell_pitch;
}


/*-
 *-----------------------------------------------------------------------
 * pmacKbdEnqueueEvent
 *	Process pmacevent destined for the keyboard.
 * 	
 * Results:
 *
 * Side Effects:
 *
 *-----------------------------------------------------------------------
 */

static xEvent	autoRepeatEvent;

void
pmacKbdEnqueueEvent(pKeyboard,me)
    DeviceRec *pKeyboard;
    register unsigned char *me;
{   
    xEvent		xE;
#ifndef	XPMAC_CODE
    xEvent		xE2;
#endif	/* XPMAC_CODE */
    PtrPrivPtr	  	ptrPriv;
    KbPrivPtr           pPriv;
    int                 delta;
    BYTE		key;
    CARD16              keyModifiers;

    ptrPriv = (PtrPrivPtr) LookupPointerDevice()->devicePrivate;

#ifndef	XPMAC_CODE
    if (KEY_DETAIL(*me) == 0x39 && KEY_UP(*me)) {
	/*
	 * Treat a KeyRelease as a KeyPress for CapsLock:
	 * we generate a fake KeyRelease after each KeyPress below...
	 */
	*me &= ~0x80;
    }
#endif	/* XPMAC_CODE */
    key = KEY_DETAIL(*me) + sysKbPriv.offset;
#ifdef	KBD
    if (noXkbExtension) {
#endif
    if ((keyModifiers = ((DeviceIntPtr)pKeyboard)->key->modifierMap[key]) == 0) {
        /*
         * Kill AutoRepeater on any real Kbd event.
         */
        autoRepeatKeyDown = 0;
	if (autoRepeatDebug)
	    ErrorF("pmacKbdEnqueueEvent: autoRepeat off\r\n");
    }
#ifdef KBD
    }
#endif

    xE.u.keyButtonPointer.time = lastEventTime;
    xE.u.u.type = (KEY_UP(*me) ? KeyRelease : KeyPress);
    xE.u.u.detail = key;

#ifdef KBD
    if (noXkbExtension) {
#endif
    if ((xE.u.u.type == KeyPress) && (keyModifiers == 0)) {
	  /* initialize new AutoRepeater event & mark AutoRepeater on */
	if (autoRepeatDebug)
	    ErrorF("pmacKbdEnqueueEvent: KEY_DOWN\r\n");
        autoRepeatEvent = xE;
        autoRepeatFirst = TRUE;
        autoRepeatKeyDown++;
        autoRepeatLastKeyDownTv = lastEventTimeTv;
    }
#ifdef KBD
   }
#endif

#ifdef XTESTEXT1
    if (!on_steal_input ||
	XTestStealKeyData(xE.u.u.detail, xE.u.u.type, XE_KEYBOARD,
			  xtest_mousex, xtest_mousey))
#endif /* XTESTEXT1 */
#ifndef	XPMAC_CODE
    if (KEY_DETAIL(*me) == 0x39 && !KEY_UP(*me)) {
	xE2 = xE;
    }
#endif	/* XPMAC_CODE */
    mieqEnqueue (&xE);
#ifndef	XPMAC_CODE
    if (KEY_DETAIL(*me) == 0x39 && !KEY_UP(*me)) {
	/* Generate a fake KeyRelease */
	xE2.u.u.type = KeyRelease;
	mieqEnqueue(&xE2);
    }
#endif	/* XPMAC_CODE */
}

static
pmacEnqueueAutoRepeat ()
{
    int	oldmask;
    int	delta;

    if (sysKbPriv.ctrl->autoRepeat != AutoRepeatModeOn) {
	autoRepeatKeyDown = 0;
	return;
    }
    /*
     * Generate auto repeat event.	XXX one for now.
     * Update time & pointer location of saved KeyPress event.
     */
    if (autoRepeatDebug)
	ErrorF("pmacKbdProcessEvent: autoRepeatKeyDown = %d\r\n",
		autoRepeatKeyDown);

    delta = TVTOMILLI(autoRepeatDeltaTv);
    autoRepeatFirst = FALSE;

    /*
     * Fake a key up event and a key down event
     * for the last key pressed.
     */
    autoRepeatEvent.u.keyButtonPointer.time += delta;
    autoRepeatEvent.u.u.type = KeyRelease;
    oldmask = sigblock (sigmask(SIGIO));
#ifdef XTESTEXT1
    if (!on_steal_input ||
	XTestStealKeyData(autoRepeatEvent.u.u.detail, autoRepeatEvent.u.u.type,
			  XE_KEYBOARD, xtest_mousex, xtest_mousey))
#endif /* XTESTEXT1 */
    mieqEnqueue (&autoRepeatEvent);
    autoRepeatEvent.u.u.type = KeyPress;
#ifdef XTESTEXT1
    if (!on_steal_input ||
	XTestStealKeyData(autoRepeatEvent.u.u.detail, autoRepeatEvent.u.u.type,
			  XE_KEYBOARD, xtest_mousex, xtest_mousey))
#endif /* XTESTEXT1 */
    mieqEnqueue (&autoRepeatEvent);
    sigsetmask (oldmask);
    /* Update time of last key down */
    tvplus(autoRepeatLastKeyDownTv, autoRepeatLastKeyDownTv,
	   autoRepeatDeltaTv);
}

Bool
LegalModifier(key, pDev)
    unsigned int key;
    DevicePtr	pDev;
{
    return (TRUE);
}

static KeybdCtrl *pKbdCtrl = (KeybdCtrl *) 0;

#include <sys/time.h>
/*ARGSUSED*/
void
pmacBlockHandler(nscreen, pbdata, pptv, pReadmask)
    int nscreen;
    pointer pbdata;
    struct timeval **pptv;
    pointer pReadmask;
{
    static struct timeval artv = { 0, 0 };    /* autorepeat timeval */

    if (!autoRepeatKeyDown)
        return;

    if (pKbdCtrl == (KeybdCtrl *) 0)
      pKbdCtrl = ((KbPrivPtr) LookupKeyboardDevice()->devicePrivate)->ctrl;

    if (pKbdCtrl->autoRepeat != AutoRepeatModeOn)
      return;

    if (autoRepeatFirst == TRUE)
        artv.tv_usec = 1000 * AUTOREPEAT_INITIATE;
    else
        artv.tv_usec = 1000 * AUTOREPEAT_DELAY;
    *pptv = &artv;

    if (autoRepeatDebug)
        ErrorF("pmacBlockHandler(%d,%d): \r\n", artv.tv_sec, artv.tv_usec);

}

/*ARGSUSED*/
void
pmacWakeupHandler(nscreen, pbdata, err, pReadmask)
    int nscreen;
    pointer pbdata;
    unsigned long err;
    pointer pReadmask;
{
    long now;

    if (autoRepeatDebug)
        ErrorF("pmacWakeupHandler(ar=%d, err=%d):\r\n", autoRepeatKeyDown, err);

    if (pKbdCtrl == (KeybdCtrl *) 0)
      pKbdCtrl = ((KbPrivPtr) LookupKeyboardDevice()->devicePrivate)->ctrl;
    
    if (pKbdCtrl->autoRepeat != AutoRepeatModeOn)
      return;

    if (autoRepeatKeyDown) {
	struct timeval tv;

	gettimeofday (&tv, (struct timezone *)0);
	tvminus(autoRepeatDeltaTv, tv, autoRepeatLastKeyDownTv);
	if (autoRepeatDeltaTv.tv_sec > 0 ||
			(!autoRepeatFirst && autoRepeatDeltaTv.tv_usec >
				(AUTOREPEAT_DELAY * 1000)) ||
			(autoRepeatDeltaTv.tv_usec >
				(AUTOREPEAT_INITIATE * 1000)))
                autoRepeatReady++;
    }

    if (autoRepeatReady)
    {
        pmacEnqueueAutoRepeat();
	autoRepeatReady = 0;
    }

}

#ifdef XTESTEXT1
void
XTestGenerateEvent(dev_type, keycode, keystate, mousex, mousey)
    int dev_type;
    int keycode;
    int keystate;
    int mousex;
    int mousey;
{
    xEvent tevent;

    tevent.u.u.type = (dev_type == XE_POINTER) ?
	(keystate == XTestKEY_UP) ? ButtonRelease : ButtonPress :
	(keystate == XTestKEY_UP) ? KeyRelease : KeyPress;
    tevent.u.u.detail = keycode;
    tevent.u.keyButtonPointer.rootX = mousex;
    tevent.u.keyButtonPointer.rootY = mousey;
    SetTimeSinceLastInputEvent();
    tevent.u.keyButtonPointer.time = lastEventTime;
    mieqEnqueue(&tevent);
    ProcessInputEvents();
}
#endif
