/*****************************************************************************
Copyright 1988-1993 by Apple Computer, Inc, Cupertino, California
			All Rights Reserved

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies.

APPLE MAKES NO WARRANTY OR REPRESENTATION, EITHER EXPRESS,
OR IMPLIED, WITH RESPECT TO THIS SOFTWARE, ITS QUALITY,
PERFORMANCE, MERCHANABILITY, OR FITNESS FOR A PARTICULAR
PURPOSE. AS A RESULT, THIS SOFTWARE IS PROVIDED "AS IS,"
AND YOU THE USER ARE ASSUMING THE ENTIRE RISK AS TO ITS
QUALITY AND PERFORMANCE. IN NO EVENT WILL APPLE BE LIABLE 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES RESULTING FROM ANY DEFECT IN THE SOFTWARE.

THE WARRANTY AND REMEDIES SET FORTH ABOVE ARE EXCLUSIVE
AND IN LIEU OF ALL OTHERS, ORAL OR WRITTEN, EXPRESS OR
IMPLIED.

*****************************************************************************/
/*-
 * powermac.h --
 *	Internal declarations for the linux-powermac ddx interface
 *
 * Copyright (c) 1987 by the Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * $XConsortium: powermac.h,v 1.19 94/01/28 22:30:02 rws Exp $
 */
#ifndef _POWERMAC_H_
#define _POWERMAC_H_

#include    <errno.h>
#include    <sys/param.h>
#include    <sys/types.h>
#include    <sys/time.h>
#include    <sys/file.h>
#include    <sys/signal.h>
#include    <linux/kd.h>

struct RGBColor {
	unsigned short red;
	unsigned short green;
	unsigned short blue;
};

typedef struct RGBColor RGBColor;

struct ColorSpec {
	short value;
	RGBColor color;
};

typedef struct ColorSpec ColorSpec;

struct video_data {
	long	v_baseoffset;
	long	v_rowbytes;
	long	v_hres;
	long	v_vres;
	long	v_right;
	long	v_left;
	long	v_top;
	long	v_bottom;
	long	v_pixelsize;
};

#include    "X.h"
#include    "Xproto.h"
#include    "Xos.h"
#include    "scrnintstr.h"
#include    "screenint.h"
#ifdef NEED_EVENTS
#include    "inputstr.h"
#endif /* NEED_EVENTS */
#include    "input.h"
#include    "cursorstr.h"
#include    "cursor.h"
#include    "pixmapstr.h"
#include    "pixmap.h"
#include    "windowstr.h"
#include    "gc.h"
#include    "gcstruct.h"
#include    "regionstr.h"
#include    "colormap.h"
#include    "miscstruct.h"
#include    "dix.h"
#include    "mfb.h"
#include    "mi.h"

/*
 * MAXEVENTS is the maximum number of events the mouse and keyboard functions
 * will read on a given call to their GetEvents vectors.
 */
#define MAXEVENTS 	32

/*
 * Data private to any powermac keyboard.
 *	ProcessEvent processes a single event and gives it to DIX
 *	DoneEvents is called when done handling a string of keyboard
 *	    events or done handling all events.
 *	devPrivate is private to the specific keyboard.
 */
#define KEY_DETAIL(e) 	((e) & 0x7f)
#define KEY_UP(e)     	((e) & 0x80)

#define MS_LEFT 	1
#define MS_MIDDLE 	2
#define MS_RIGHT 	3

#define MOUSE_ESCAPE 	0x7e			/* from <sys/video.h> */

#define DEF_MIDDLE_KEY_TRIGGER	0x13	/* default keycode (2) for middle */
#define DEF_RIGHT_KEY_TRIGGER	0x14	/* default keycode (3) for right */
#define MIDDLE_KEY_CODE		0x3f	/* actual keycode for middle */
#define RIGHT_KEY_CODE		0x40	/* actual keycode for right */

#define IS_MIDDLE_TRIGGER_KEY(c)	\
		(KEY_DETAIL(c) == middleKeyTrigger)
#define IS_RIGHT_TRIGGER_KEY(c)		\
		(KEY_DETAIL(c) == rightKeyTrigger)
#define IS_TRIGGER_KEY(c)		\
		(IS_MIDDLE_TRIGGER_KEY(c) || IS_RIGHT_TRIGGER_KEY(c))

#define IS_MIDDLE_KEY(c)		\
		(KEY_DETAIL(c) == MIDDLE_KEY_CODE)
#define IS_RIGHT_KEY(c)			\
		(KEY_DETAIL(c) == RIGHT_KEY_CODE)
#define IS_MOUSE_KEY(c)			\
		(IS_MIDDLE_KEY(c) || IS_RIGHT_KEY(c))

#define KBTYPE_MACII 	0

typedef struct kbPrivate {
    int	    	  type;           	/* Type of keyboard */
    void    	  (*EnqueueEvent)();	/* Function to process an event */
    int		  offset;		/* to be added to device keycodes */
    KeybdCtrl     *ctrl;                /* Current control structure (for
                                         * keyclick, bell duration, auto-
                                         * repeat, etc.) */
} KbPrivRec, *KbPrivPtr;

#define	MIN_KEYCODE	8	/* necessary to avoid the mouse buttons */

#ifdef XTESTEXT1
extern KeyCode xtest_command_key;
#define XE_POINTER 0
#define XE_KEYBOARD 1
#endif

/*
 * Data private to any powermac pointer device.
 *	ProcessEvent and DoneEvents have uses similar to the
 *	    keyboard fields of the same name.
 *	pScreen is the screen the pointer is on (only valid if it is the
 *	    main pointer device).
 *	x and y are absolute coordinates on that screen (they may be negative)
 */

#define ADB_MOUSE 1
#define USB_MOUSE 2

typedef void (drvr_proc) ( int fd );
drvr_proc mdrvr_ps2;

typedef struct ptrPrivate {
    void    	  (*EnqueueEvent)();	/* Function to process an event */
    ScreenPtr	  pScreen;  	    	/* Screen pointer is on */
    int           fd;                   /* fd device opened on */
    char *        device;               /* /dev/usbmouse */
    drvr_proc *   driver;               /* mdrvr_ps2 */
} PtrPrivRec, *PtrPrivPtr;

/*
 * Frame-buffer-private info.
 *	fb  	  	pointer to the mapped image of the frame buffer. Used
 *	    	  	by the driving routines for the specific frame buffer
 *	    	  	type.
 *	slot
 *	defaultDepth
 *	installedMap
 *	info	  	description of the frame buffer -- type, height, depth,
 *	    	  	width, etc.
 *	fbPriv	  	Data private to the frame buffer type.
 */

typedef struct {
    pointer 	  	fb; 	    /* Frame buffer itself */
    int			slot;
    int			defaultDepth;
    ColormapPtr		installedMap;
    struct video_data	info;	    /* Frame buffer characteristics */
    pointer 	  	fbPriv;	    /* Frame-buffer-dependent data */
} fbFd;

extern fbFd 	  pmacFbs[];

/*
 * Data describing each type of frame buffer. The probeProc is called to
 * see if such a device exists and to do what needs doing if it does. devName
 * is the expected name of the device in the file system. 
 */
typedef enum {
	neverProbed, probedAndSucceeded, probedAndFailed
} pmacProbeStatus;

typedef struct _pmacFbDataRec {
    Bool    (*probeProc)();	/* probe procedure for this fb */
    char    *devName;		/* device filename */
    pmacProbeStatus probeStatus;	/* TRUE if fb has been probed successfully */
} pmacFbDataRec;

extern pmacFbDataRec pmacFbData[];

#define CTB_SET_ENTRIES(fbArray) \
    (fbArray.info.vpPixelType == 16)? cscDirectSetEntries: cscSetEntries

/*
 * Initialization
 */
extern Bool 	  pmacScreenInit();

extern int  	  pmacCheckInput;    /* Non-zero if input is available */

extern int  	  lastEventTime;    /* Time (in ms.) of last event */
extern void 	  SetTimeSinceLastInputEvent();

#define AUTOREPEAT_INITIATE     (400)           /* milliseconds */
#define AUTOREPEAT_DELAY        (50)           /* milliseconds */
/*
 * We signal autorepeat events with the unique id AUTOREPEAT_EVENTID.
 */
#define AUTOREPEAT_EVENTID      (0x7d)          /* AutoRepeat id */

/*-
 * TVTOMILLI(tv)
 *	Given a struct timeval, convert its time into milliseconds...
 */
#define TVTOMILLI(tv)	(((tv).tv_usec/1000)+((tv).tv_sec*1000))

#define tvminus(tv, tv1, tv2) /* tv = tv1 - tv2 */ \
              if ((tv1).tv_usec < (tv2).tv_usec) { \
                      (tv1).tv_usec += 1000000; \
                      (tv1).tv_sec -= 1; \
              } \
              (tv).tv_usec = (tv1).tv_usec - (tv2).tv_usec; \
              (tv).tv_sec = (tv1).tv_sec - (tv2).tv_sec;

#define tvplus(tv, tv1, tv2)  /* tv = tv1 + tv2 */ \
              (tv).tv_sec = (tv1).tv_sec + (tv2).tv_sec; \
              (tv).tv_usec = (tv1).tv_usec + (tv2).tv_usec; \
              if ((tv).tv_usec > 1000000) { \
                      (tv).tv_usec -= 1000000; \
                      (tv).tv_sec += 1; \
              }

#define LookupPointerDevice()	pPointerDevice
#define LookupKeyboardDevice()	pKeyboardDevice
extern DevicePtr pPointerDevice, pKeyboardDevice;

extern int	console_fd;
extern int	inXpmac;

#define MAX_VIDEO_MODE	20

#if !defined(VC_FBMAP)
# define VC_FBMAP 0x766d
#endif

#endif /* _POWERMAC_H_ */

