/* $XFree86: xc/programs/Xserver/hw/xfree86/accel/mach64/mach64im.c,v 3.5 1996/12/23 06:39:18 dawes Exp $ */
/*
 * Copyright 1992,1993,1994,1995,1996 by Kevin E. Martin, Chapel Hill, North Carolina.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Kevin E. Martin not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Kevin E. Martin makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * KEVIN E. MARTIN, RICKARD E. FAITH, AND TIAGO GONS DISCLAIM ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Modified for the Mach-8 by Rickard E. Faith (faith@cs.unc.edu)
 * Modified for the Mach32 by Kevin E. Martin (martin@cs.unc.edu)
 * Modified for the Mach64 by Kevin E. Martin (martin@cs.unc.edu)
 */
/* $XConsortium: mach64im.c /main/4 1996/05/13 16:36:12 kaleb $ */
/* Modified for the Rage128 by Ryuichi Oikawa (roikawa@rr.iij4u.or.jp) */

#include "X.h"
#include "misc.h"
/*#include "xf86.h"*/
/*#include "xf86_HWlib.h"*/
#include "rage128.h"

#if defined(RAGE128_BIG_ENDIAN)
#define HOST_ENDIAN_CNTL  0
#else
#define HOST_ENDIAN_CNTL  DST_HOST_BIG_ENDIAN_EN
#endif

static inline void memcpy_double(char *dst, char *src, int n)
{
  while((unsigned long)dst & 0x07) {
    *dst++ = *src++;
    --n;
    if (n < 1) return;
  }
  n -= 8;
  while(n >= 0) {
    *(double *)dst = *(double *)src;
    n -= 8; src += 8; dst += 8;
  }
  n += 8;
  while (n > 0) {
    *dst++ = *src++;
    --n;
  }
}

/* Someone who knows the way to initiate burst writing */
/* other than 8 byte transfer, please improve me!!     */
static inline void WriteToHost(int *src, int count)
{
  double *host = (double *)(rage128MemReg + HOST_DATA0);
  int i;

  if((unsigned long)src & 0x07) {
    WaitQueue(1);
    *(int *)host = *src++;
    --count;
  }
  while (count & ~0xf) {
    WaitQueue(16);
    for(i = 0; i < 8; i++) {
      *host = *(double *)src;
      ++((double *)src);
      count -= 2;
    }
  }
  if(count < 1)
    return;
  WaitQueue(count);
  while(count >= 2) {
    *host = *(double *)src;
    ++((double *)src);
    count -= 2;
  }
  if(count)
    *(int *)host = *src;
}

static void rage128ImageReadVT();
static void rage128ImageWriteVT();

void	(*rage128ImageReadFunc)();
void	(*rage128ImageWriteFunc)();

static void rage128ImageRead();
static void rage128ImageWrite();
static void rage128ImageWriteHW();

static unsigned long PMask;
static int BytesPerPixel;
static int screenStride;


void
rage128ImageInit()
{
    int i;

    PMask = (1UL << rage128BitDepth) - 1;
    BytesPerPixel = vgaBitsPerPixel / 8;
    screenStride = rage128VirtX * BytesPerPixel;
    if (vgaBitsPerPixel < 16) {
       rage128ImageReadFunc = rage128ImageReadVT;
       rage128ImageWriteFunc = rage128ImageWriteVT;
    } else {
       rage128ImageReadFunc = rage128ImageRead;
       rage128ImageWriteFunc = rage128ImageWrite;
    }
}


static void
rage128ImageWriteHW(x, y, w, h, psrc, pwidth, px, py, alu, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    int			alu;
    unsigned long	planemask;
{
    int wordsPerLine, misalign = 0;

    WaitQueue(7);
    regw(DP_GUI_MASTER_CNTL, GMC_SRC_PITCH_OFFSET_DEFAULT |
                             GMC_DST_PITCH_OFFSET_DEFAULT |
                             GMC_SRC_CLIP_DEFAULT         |
                             GMC_DST_CLIP_LEAVE           |
                             GMC_BRUSH_SOLIDCOLOR         |
                             (rage128DataType << 8)       |
                             GMC_SRC_DSTCOLOR             |
                             GMC_BYTE_ORDER_MSB_TO_LSB    |
                             GMC_DP_CONVERSION_TEMP_6500  |
                             alu                          |
                             GMC_DP_SRC_HOST              |
                             GMC_3D_FCN_EN_CLR            |
                             GMC_DST_CLR_CMP_FCN_CLEAR    |
                             GMC_AUX_CLIP_CLEAR           |
                             GMC_WRITE_MASK_LEAVE);
    regw(DP_CNTL, DST_X_LEFT_TO_RIGHT |
	          DST_Y_TOP_TO_BOTTOM | HOST_ENDIAN_CNTL);
    regw(DP_WRITE_MASK, planemask);
    regw(SC_TOP_LEFT, (y << 16) | x);
    regw(SC_BOTTOM_RIGHT, ((y+h-1) << 16) | (x+w-1));

    w = (w + 3) / 4; /* round up to int boundry */
    psrc += pwidth * py + px * BytesPerPixel;
    wordsPerLine = w * BytesPerPixel;
    misalign = (wordsPerLine * h) & 0x01;
    regw(DST_Y_X, (y << 16) | x);
    regw(DST_HEIGHT_WIDTH, (h << 16) | (w * 4));

    while (h--) {
      WriteToHost((int *)psrc, wordsPerLine);
      psrc += pwidth;
    }
    if(misalign) {
      *(int *)(rage128MemReg + HOST_DATA0) = 0;
      ErrorF("Misaligned data in rage128im.c\n");
    }
    *(int *)(rage128MemReg + HOST_DATA_LAST) = 0;

    WaitIdleEmpty();
}

static void
rage128ImageWrite(x, y, w, h, psrc, pwidth, px, py, alu, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    int			alu;
    unsigned long	planemask;
{
    pointer curvm;
    int byteCount, j;

    if ((w <= 0) || (h <= 0) || alu == ROP3_D)
	return;

    if ((alu != ROP3_S) || ((planemask & PMask) != PMask)) {
	rage128ImageWriteHW(x, y, w, h, psrc, pwidth, px, py, alu, planemask);
	return;
    }

    psrc += pwidth * py + px * BytesPerPixel;
    curvm = (pointer)((unsigned char *)rage128VideoMem +
	    (x + y * rage128VirtX) * BytesPerPixel);
    byteCount = w * BytesPerPixel;
    WaitIdleEmpty();

    while(h--) {
      if (!(((unsigned long)curvm ^ (unsigned long)psrc) & 0x07))
	memcpy_double((char *)curvm, (char *)psrc, byteCount);
      else
	memcpy((void *)curvm, psrc, byteCount);
      curvm = (pointer)((unsigned char *)curvm + screenStride); 
      psrc += pwidth;
    }
}


static void
rage128ImageWriteVT(x, y, w, h, psrc, pwidth, px, py, alu, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    int			alu;
    unsigned long	planemask;
{
    if ((w <= 0) || (h <= 0) || alu == ROP3_D)
	return;
    rage128ImageWriteHW(x, y, w, h, psrc, pwidth, px, py, alu, planemask);
}



static void
rage128ImageRead(x, y, w, h, psrc, pwidth, px, py, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    unsigned long	planemask;
{
    int j;
    pointer curvm;

    if ((w <= 0) || (h <= 0))
	return;

    if ((planemask & PMask) != PMask) {
#ifdef NOT_DEFINED
        rage128ImageReadHW(x, y, w, h, psrc, pwidth, px, py, planemask);
	return;
#else
	ErrorF ("rage128ImageRead: unsupported planemask\n");
#endif
    }
    psrc += pwidth * py + px * BytesPerPixel;
    curvm = (pointer)((unsigned char *)rage128VideoMem +
	    (x + y * rage128VirtX) * BytesPerPixel);
    WaitIdleEmpty();

    while(h--) {
      if (!(((unsigned long)curvm ^ (unsigned long)psrc) & 7))
	memcpy_double((char *)psrc, (char *)curvm, w * BytesPerPixel);
      else
        memcpy(psrc, (void *)curvm, w * BytesPerPixel);
      curvm = (pointer)((unsigned char *)curvm + screenStride); 
      psrc += pwidth;
    }
}


static void
rage128ImageReadVT(x, y, w, h, psrc, pwidth, px, py, planemask)
    int			x;
    int			y;
    int			w;
    int			h;
    char		*psrc;
    int			pwidth;
    int			px;
    int			py;
    unsigned long	planemask;
{
    int j;
    pointer curvm;

    if ((w <= 0) || (h <= 0))
	return;

    if ((planemask & PMask) != PMask)
	ErrorF ("rage128ImageRead: unsupported planemask\n");
    psrc += pwidth * py + px * BytesPerPixel;
    curvm = (pointer)((unsigned char *)rage128VideoMem +
	    (x + y * rage128VirtX) * BytesPerPixel);
    WaitIdleEmpty();

    while(h--) {
      switch(BytesPerPixel) {
        case 2:
	  for(j = 0; j < w; j++)
	    ((short *)psrc)[j] = ((short *)curvm)[j];
	  break;
        case 4:
	  for(j = 0; j < w; j++)
	    ((long *)psrc)[j] = ((long *)curvm)[j];
	  break;
        default:
	  for(j = 0; j < w; j++)
	    ((char *)psrc)[j] = ((char *)curvm)[j];
	    break;
	}
	curvm = (pointer)((unsigned char *)curvm + screenStride); 
	psrc += pwidth;
    }
}


void
rage128ImageStippleFunc(x, y, w, h, psrc, pwidth, px, py, fgPixel, bgPixel, 
		       alu, planemask, opaque)
    int                 x;
    int                 y;
    int                 w;
    int                 h;
    char                *psrc;
    int                 pwidth;
    int                 px;
    int                 py;
    Pixel               fgPixel;
    Pixel               bgPixel;
    int                 alu;
    unsigned long       planemask;
    int 		opaque;
{
    int *pline;
    int wordsPerLine, monotype, misalign = 0;

    if (alu == ROP3_D || w <= 0 || h <= 0)
	return;

    if(opaque)  /* Is this correct? Please check. */
      monotype = GMC_SRC_MONO;
    else
      monotype = GMC_SRC_MONO_LBKGD;
    WaitQueue(9);
    regw(DP_GUI_MASTER_CNTL, GMC_SRC_PITCH_OFFSET_DEFAULT |
                             GMC_DST_PITCH_OFFSET_DEFAULT |
                             GMC_SRC_CLIP_DEFAULT         |
                             GMC_DST_CLIP_LEAVE           |
                             GMC_BRUSH_SOLIDCOLOR         |
                             (rage128DataType << 8)       |
                             monotype                     |
                             GMC_BYTE_ORDER_MSB_TO_LSB    |
                             GMC_DP_CONVERSION_TEMP_6500  |
                             alu                          |
                             GMC_DP_SRC_HOST              |
                             GMC_3D_FCN_EN_CLR            |
                             GMC_DST_CLR_CMP_FCN_CLEAR    |
                             GMC_AUX_CLIP_CLEAR           |
                             GMC_WRITE_MASK_LEAVE);
    regw(DP_CNTL, DST_X_LEFT_TO_RIGHT |
	          DST_Y_TOP_TO_BOTTOM | HOST_ENDIAN_CNTL);
    regw(DP_WRITE_MASK, planemask);
    regw(SC_TOP_LEFT, (y << 16) | x);
    regw(SC_BOTTOM_RIGHT, ((y+h-1) << 16) | (x+w-1));
    regw(DP_SRC_FRGD_CLR, fgPixel);
    regw(DP_SRC_BKGD_CLR, bgPixel);

    w = (w + (px % 32) + 31) >> 5; /* round up to int boundry and take into */
				   /* account the pixels skipped over */
    wordsPerLine = pwidth >> 2;
    misalign = (wordsPerLine * h) & 0x01;
    pline = (int*)psrc + (wordsPerLine * py) + (px >> 5);
    regw(DST_Y_X,  (y << 16) | (x - (px % 32)));
    regw(DST_HEIGHT_WIDTH, (h << 16) | (w * 32));

    while (h--) {
      WriteToHost((int *)pline, w);
      pline += wordsPerLine;
    }
    if(misalign)
      *(int *)(rage128MemReg + HOST_DATA0) = 0;
    *(int *)(rage128MemReg + HOST_DATA_LAST) = 0;

    WaitIdleEmpty();
}
