#include "rage128.h"
#include <asm/vc_ioctl.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <linux/mman.h>
#include <asm/io.h>
#include "videoprobe.h"

int rage128alu[16] = {
  ROP3_ZERO,         /* GXclear        */
  ROP3_DSa,          /* Gxand          */
  ROP3_SDna,         /* GXandReverse   */
  ROP3_S,            /* GXcopy         */
  ROP3_DSna,         /* GXandInverted  */
  ROP3_D,            /* GXnoop         */
  ROP3_DSx,          /* GXxor          */
  ROP3_DSo,          /* GXor           */
  ROP3_DSon,         /* GXnor          */
  ROP3_DSxn,         /* GXequiv        */
  ROP3_Dn,           /* GXinvert       */
  ROP3_SDno,         /* GXorReverse    */
  ROP3_Sn,           /* GXcopyInverted */
  ROP3_DSno,         /* GXorInverted   */
  ROP3_DSan,         /* GXnand         */
  ROP3_ONE           /* GXset          */
};

int rage128alup[16] = {
  ROP3_ZERO,         /* GXclear        */
  ROP3_DPa,          /* Gxand          */
  ROP3_PDna,         /* GXandReverse   */
  ROP3_P,            /* GXcopy         */
  ROP3_DPna,         /* GXandInverted  */
  ROP3_D,            /* GXnoop         */
  ROP3_DPx,          /* GXxor          */
  ROP3_DPo,          /* GXor           */
  ROP3_DPon,         /* GXnor          */
  ROP3_PDxn,         /* GXequiv        */
  ROP3_Dn,           /* GXinvert       */
  ROP3_PDno,         /* GXorReverse    */
  ROP3_Pn,           /* GXcopyInverted */
  ROP3_DPno,         /* GXorInverted   */
  ROP3_DPan,         /* GXnand         */
  ROP3_ONE           /* GXset          */
};

int rage128MaxX, rage128MaxY;
int rage128VirtX, rage128VirtY;
int rage128DataType;
int rage128BitDepth;
pointer rage128MemReg = (pointer)0;
pointer rage128VideoMem = (pointer)0;
unsigned int rage128MemorySize;
unsigned int rage128ApertureSize;
unsigned long rage128ApertureAddr;
unsigned long rage128RegisterAddr;
unsigned int rage128LinearOffset;

static pointer rage128MemReg0 = (pointer)0;
static pointer rage128MemReg1 = (pointer)0;
static unsigned long rage128MemReg0_phys;
static unsigned long rage128MemReg1_phys;
static unsigned long rage128Aperture0_phys;
static unsigned long rage128Aperture1_phys;

extern int vgaBitsPerPixel;
extern int vgaBytesPerPixel;
extern unsigned char *DacIndex;
extern unsigned char *DacData;
extern unsigned char savedHWColormap[];

#if !defined(PCI_DEVICE_ID_ATI_128RE)
#define PCI_DEVICE_ID_ATI_128RE 0x5245
#endif
#if !defined(PCI_DEVICE_ID_ATI_128RF)
#define PCI_DEVICE_ID_ATI_128RF 0x5246
#endif
#if !defined(PCI_DEVICE_ID_ATI_128RK)
#define PCI_DEVICE_ID_ATI_128RK 0x524B
#endif
#if !defined(PCI_DEVICE_ID_ATI_128RL)
#define PCI_DEVICE_ID_ATI_128RL 0x524C
#endif
#if !defined(PCI_DEVICE_ID_ATI_128PF)
#define PCI_DEVICE_ID_ATI_128PF 0x5046
#endif
#if !defined(PCI_DEVICE_ID_ATI_MOB_M)
#define PCI_DEVICE_ID_ATI_MOB_M 0x4C46
#endif

#define CFG_APER_0_LE     0x00000000
#define CFG_APER_0_BE16   0x00000001
#define CFG_APER_0_BE32   0x00000002
#define CFG_APER_1_LE     0x00000000
#define CFG_APER_1_BE16   0x00000004
#define CFG_APER_1_BE32   0x00000008
#define CFG_APER_REG_LE   0x00000000
#define CFG_APER_REG_BE32 0x00000010
#define CFG_APER_REG_LE32 0x10000000
#define LE_TO_BE          1
#define BE_TO_LE          2

static struct videoID {
  int device_id;
  char *device_name;
} rage128IDs[] = {
  { PCI_DEVICE_ID_ATI_128RE, "Rage128 RE     " },
  { PCI_DEVICE_ID_ATI_128RF, "Rage128 RF     " },
  { PCI_DEVICE_ID_ATI_128RK, "Rage128 RK     " },
  { PCI_DEVICE_ID_ATI_128RL, "Rage128 RL     " },
  { PCI_DEVICE_ID_ATI_128PF, "Rage128Pro PF  " },
  { PCI_DEVICE_ID_ATI_MOB_M, "RageMobility M3" },
  { 0x534B,                  "Rage128 GL     " },
  { 0x534C,                  "Rage128 GL     " },
  { 0x534D,                  "Rage128 GL     " },
  { 0x5345,                  "Rage128 VR     " },
  { 0x5346,                  "Rage128 VR     " },
  { 0x5347,                  "Rage128 VR     " },
  { 0x5041,                  "Rage128Pro GL  " },
  { 0x5042,                  "Rage128Pro GL  " },
  { 0x5043,                  "Rage128Pro GL  " },
  { 0x5044,                  "Rage128Pro GL  " },
  { 0x5045,                  "Rage128Pro GL  " },
  { 0x5047,                  "Rage128Pro VR  " },
  { 0x5048,                  "Rage128Pro VR  " },
  { 0x5049,                  "Rage128Pro VR  " },
  { 0x504A,                  "Rage128Pro VR  " },
  { 0x504B,                  "Rage128Pro VR  " },
  { 0x504C,                  "Rage128Pro VR  " },
  { 0x504D,                  "Rage128Pro VR  " },
  { 0x504E,                  "Rage128Pro VR  " },
  { 0x504F,                  "Rage128Pro VR  " },
  { 0x5050,                  "Rage128Pro VR  " },
  { 0x5051,                  "Rage128Pro VR  " },
  { 0x5052,                  "Rage128Pro VR  " },
  { 0x5053,                  "Rage128Pro VR  " },
  { 0x5054,                  "Rage128Pro VR  " },
  { 0x5055,                  "Rage128Pro VR  " },
  { 0x5056,                  "Rage128Pro VR  " },
  { 0x5057,                  "Rage128Pro VR  " },
  { 0x5058,                  "Rage128Pro VR  " },
  { 0x4C45,                  "RageMobility M3" }
};


static int rage128Probe(unsigned long frame_buffer_phys)
{
  int i, fd, endian_change = 0;
  struct pci_device_tag *rage128;
  unsigned int cfg;

  for(i = 0; i < sizeof(rage128IDs)/sizeof(struct videoID); i++)
    if((rage128 = ProbeVideoByAddress(PCI_VENDOR_ID_ATI, rage128IDs[i].device_id,
				      0, frame_buffer_phys, 0xf8000000))) {
      if((fd = open("/dev/mem", O_RDWR)) < 0)
	return(FALSE);
      rage128MemReg0_phys = rage128->base_address[2] & 0xffffc000;
      rage128MemReg = rage128MemReg0 = (pointer)mmap(0, 0x2000, PROT_READ|PROT_WRITE,
				     MAP_SHARED, fd, rage128MemReg0_phys);
      if((long)rage128MemReg == -1) {
	close(fd);
	return(FALSE);
      }
#if defined(RAGE128_BIG_ENDIAN)
      cfg = in_le32(rage128MemReg0 + CONFIG_CNTL);
      if(!(cfg & CFG_APER_REG_LE32)) {
	endian_change = LE_TO_BE;
	out_le32(rage128MemReg0 + CONFIG_CNTL, cfg | CFG_APER_REG_BE32);
      }
#else
      cfg = in_be32(rage128MemReg0 + CONFIG_CNTL);
      if((cfg & CFG_APER_REG_BE32)) {
	endian_change = BE_TO_LE;
	out_be32(rage128MemReg0 + CONFIG_CNTL, cfg & ~CFG_APER_REG_BE32);
      }
#endif
      rage128MemReg1_phys = regr(CONFIG_REG_1_BASE) & 0xffffe000;
      rage128MemReg1 = (pointer)mmap(0, 0x2000, PROT_READ|PROT_WRITE,
				     MAP_SHARED, fd, rage128MemReg1_phys);
      close(fd);
      if((long)rage128MemReg1 == -1) {
	ErrorF("Rage128: Failed to map Regbase1. Restoring endian mode.\n");
#if defined(RAGE128_BIG_ENDIAN)
	out_le32(rage128MemReg0 + CONFIG_CNTL, cfg);
#else
	out_be32(rage128MemReg0 + CONFIG_CNTL, cfg);
#endif
	return(FALSE);
      }
      rage128MemorySize = regr(CONFIG_MEMSIZE) & 0x03ffffff;
      rage128Aperture0_phys = regr(CONFIG_APER_0_BASE) & 0xf8000000;
      rage128Aperture1_phys = regr(CONFIG_APER_1_BASE) & 0xfe000000;
      rage128ApertureSize = regr(CONFIG_APER_SIZE) & 0xfe000000;

      ErrorF("ATI %s(0x%04x) detected.\n", rage128IDs[i].device_name, rage128IDs[i].device_id);
      ErrorF("Rage128 PCI: Membase:0x%08x, IObase:0x%04x, Regbase0:0x%08x\n",
	     rage128->base_address[0] & ~0x0f,
	     rage128->base_address[1] & ~0x0f,
	     rage128->base_address[2] & ~0x0f);
      ErrorF("Rage128: Aperture0:0x%08x, Aperture1:0x%08x, Aperture size:0x%08x, Memory size:%d MB\n",
	     rage128Aperture0_phys, rage128Aperture1_phys,
	     rage128ApertureSize, rage128MemorySize >> 20);
      ErrorF("Rage128: Framebuffer:0x%08x, Regbase0:0x%08x(0x%08x), Regbase1:0x%08x(0x%08x)\n",
	     frame_buffer_phys, rage128MemReg0_phys, rage128MemReg0,
	     rage128MemReg1_phys, rage128MemReg1);
#if defined(RAGE128_BIG_ENDIAN)
      ErrorF("Rage128: Configured as big endian mode.%s\n", endian_change?"(from little endian)":"");
#else
      ErrorF("Rage128: Configured as little endian mode.%s\n", endian_change?"(from big endian)":"");
#endif
      rage128MemReg = rage128MemReg0; /* or rage128MemReg1 */
      return(TRUE);
    }
  return(FALSE);
}

void Rage128BlitterInit(struct vc_mode mode, unsigned char **framebuffer)
{
  int bpp = mode.depth, width = mode.width, height = mode.height;
  int BytesPerPixel, pitch = mode.pitch;

  if(framebuffer) {
    if(!rage128Probe(mode.fb_address)) {
	rage128MemReg = 0;
	return;
    }
    ErrorF("Starting acceleration...\n");
    rage128VideoMem = *framebuffer;
#if 0
    {
      int i;
      unsigned char *rindex;

      DacIndex = rage128MemReg + PALETTE_INDEX;
      DacData = rage128MemReg + PALETTE_DATA;
      rindex = DacIndex + 2;
      *rindex = 0; eieio();
      for(i = 0; i < 256; i++) {
	savedHWColormap[3*i] = DacData[0];
	savedHWColormap[3*i+1] = DacData[1];
	savedHWColormap[3*i+2] = DacData[2];
      }
    }
#endif
  }
  if(!rage128MemReg)
    return;

  switch (bpp) {
    case 16:
      rage128BitDepth = 15;
      vgaBitsPerPixel = 16;
      vgaBytesPerPixel = 2;
      rage128VirtX = pitch >> 1;
      rage128DataType = DST_15BPP; /* RGB565 is not supported */
      break;
    case 24:
    case 32:
      rage128BitDepth = 24;
      vgaBitsPerPixel = 32;
      vgaBytesPerPixel = 4;
      rage128VirtX = pitch >> 2;
      rage128DataType = DST_32BPP; /* packed 24bpp is not supported */
      break;
    default:
      rage128BitDepth = 8;
      vgaBitsPerPixel = 8;
      vgaBytesPerPixel = 1;
      rage128VirtX = pitch;
      rage128DataType = DST_8BPP;
      break;
    }
    rage128MaxX = width - 1;
    rage128VirtY = height;
    rage128MaxY = height - 1;
    rage128LinearOffset = regr(CRTC_OFFSET) & 0x00ffffff;

    ResetEngine();
    WaitIdleEmpty();
    regw(DEFAULT_OFFSET, rage128LinearOffset);
    regw(DEFAULT_PITCH, rage128VirtX >> 3);
    regw(DEFAULT_SC_BOTTOM_RIGHT, (rage128MaxY << 16) | rage128MaxX);
    regw(DP_GUI_MASTER_CNTL, GMC_SRC_PITCH_OFFSET_DEFAULT |
                             GMC_DST_PITCH_OFFSET_DEFAULT |
                             GMC_SRC_CLIP_DEFAULT         |
                             GMC_DST_CLIP_DEFAULT         |
                             GMC_BRUSH_SOLIDCOLOR         |
                             (rage128DataType << 8)       |
                             GMC_SRC_DSTCOLOR             |
                             GMC_BYTE_ORDER_MSB_TO_LSB    |
                             GMC_DP_CONVERSION_TEMP_6500  |
                             ROP3_PATCOPY                 |
                             GMC_DP_SRC_RECT              |
                             GMC_3D_FCN_EN_CLR            |
                             GMC_DST_CLR_CMP_FCN_CLEAR    |
                             GMC_AUX_CLIP_CLEAR           |
                             GMC_WRITE_MASK_SET);
    regw(DST_BRES_ERR, 0);
    regw(DST_BRES_INC, 0);
    regw(DST_BRES_DEC, 0);
    regw(DP_BRUSH_FRGD_CLR, 0xFFFFFFFF);
    regw(DP_BRUSH_BKGD_CLR, 0x00000000);
    regw(DP_SRC_FRGD_CLR, 0xFFFFFFFF);
    regw(DP_SRC_BKGD_CLR, 0x00000000);
    regw(DP_WRITE_MASK, 0xFFFFFFFF);
    WaitIdleEmpty();

    rage128ImageInit();
}
