#ifndef REGRAGE128_H
#define REGRAGE128_H

/******************************************************************************
 *               NON-GUI Block Memory Mapped Registers                        *
 ******************************************************************************/
#define PALETTE_INDEX                              0x00b0
#define PALETTE_DATA                               0x00b4
#define CONFIG_CNTL                                0x00e0
#define GEN_RESET_CNTL                             0x00f0
#define CONFIG_MEMSIZE                             0x00f8
#define CONFIG_APER_0_BASE                         0x0100
#define CONFIG_APER_1_BASE                         0x0104
#define CONFIG_APER_SIZE                           0x0108
#define CONFIG_REG_1_BASE                          0x010c
#define CONFIG_MEMSIZE_EMBEDDED                    0x0114
#define CRTC_OFFSET                                0x0224
#define CRTC_OFFSET_CNTL                           0x0228


/******************************************************************************
 *                  GUI Block Memory Mapped Registers                         *
 *                     These registers are FIFOed.                            *
 ******************************************************************************/

#define DST_PITCH                                  0x1408
#define DST_HEIGHT                                 0x1410
#define DST_X                                      0x141c
#define DST_Y                                      0x1420
#define SRC_PITCH_OFFSET                           0x1428
#define DST_PITCH_OFFSET                           0x142c
#define SRC_Y_X                                    0x1434
#define DST_Y_X                                    0x1438
#define DST_HEIGHT_WIDTH                           0x143c
#define DP_GUI_MASTER_CNTL                         0x146c
#define DP_BRUSH_BKGD_CLR                          0x1478
#define DP_BRUSH_FRGD_CLR                          0x147c
#define DST_HEIGHT_WIDTH_8                         0x158c
#define DST_X_Y                                    0x1594
#define DST_HEIGHT_Y                               0x15a0
#define DST_X_SUB                                  0x15a4
#define DST_Y_SUB                                  0x15a8
#define SRC_PITCH                                  0x15b0
#define DST_HEIGHT_WIDTH_BW                        0x15b4
#define CLR_CMP_CNTL                               0x15c0
#define DP_SRC_FRGD_CLR                            0x15d8
#define DP_SRC_BKGD_CLR                            0x15dc
#define DST_BRES_ERR                               0x1628
#define DST_BRES_INC                               0x162c
#define DST_BRES_DEC                               0x1630
#define DST_BRES_LNTH                              0x1634
#define DST_BRES_LNTH_SUB                          0x1638
#define SC_TOP                                     0x1648
#define SC_BOTTOM                                  0x164c
#define SRC_SC_BOTTOM                              0x165c
#define AUX1_SC_TOP                                0x166c
#define AUX1_SC_BOTTOM                             0x1670
#define AUX2_SC_TOP                                0x167c
#define AUX2_SC_BOTTOM                             0x1680
#define AUX3_SC_TOP                                0x168c
#define AUX3_SC_BOTTOM                             0x1690
#define DP_CNTL                                    0x16c0
#define DP_MIX                                     0x16c8
#define DP_WRITE_MASK                              0x16cc
#define DP_CNTL_XDIR_YDIR_YMAJOR                   0x16d0
#define DEFAULT_OFFSET                             0x16e0
#define DEFAULT_PITCH                              0x16e4
#define DEFAULT_SC_BOTTOM_RIGHT                    0x16e8
#define SC_TOP_LEFT                                0x16ec
#define SC_BOTTOM_RIGHT                            0x16f0
#define SRC_SC_BOTTOM_RIGHT                        0x16f4
#define GUI_STAT                                   0x1740
#define HOST_DATA0                                 0x17c0
#define HOST_DATA_LAST                             0x17e0
#define SCALE_DST_X_Y                              0x19b0
#define SCALE_DST_HEIGHT_WIDTH                     0x19b4
#define DST_PITCH_OFFSET_C                         0x1c80
#define DP_GUI_MASTER_CNTL_C                       0x1c84
#define SC_TOP_LEFT_C                              0x1c88
#define SC_BOTTOM_RIGHT_C                          0x1c8c

/* CLR_CMP_CNTL bit constants */
#define COMPARE_SOURCE                             0x01000000

/* DP_CNTL bit constants */
#define DST_X_RIGHT_TO_LEFT                        0x00000000
#define DST_X_LEFT_TO_RIGHT                        0x00000001
#define DST_Y_BOTTOM_TO_TOP                        0x00000000
#define DST_Y_TOP_TO_BOTTOM                        0x00000002
#define DST_X_MAJOR                                0x00000000
#define DST_Y_MAJOR                                0x00000004
#define DST_X_TILE                                 0x00000008
#define DST_Y_TILE                                 0x00000010
#define DST_HOST_BIG_ENDIAN_EN                     0x00000200

/* DP_CNTL_YDIR_XDIR_YMAJOR bit constants (short version of DP_CNTL) */
#define DST_X_MAJOR_S                              0x00000000
#define DST_Y_MAJOR_S                              0x00000001
#define DST_Y_BOTTOM_TO_TOP_S                      0x00000000
#define DST_Y_TOP_TO_BOTTOM_S                      0x00008000
#define DST_X_RIGHT_TO_LEFT_S                      0x00000000
#define DST_X_LEFT_TO_RIGHT_S                      0x80000000

/* DP_DATATYPE bit constants */
#define DST_8BPP                                   0x00000002
#define DST_15BPP                                  0x00000003
#define DST_32BPP                                  0x00000006
#define DST_8BPP_RGB332                            0x00000007
#define DST_8BPP_Y8                                0x00000008
#define DST_8BPP_RGB8                              0x00000009
#define DST_32BPP_AYUV444                          0x0000000e
#define BRUSH_SOLIDCOLOR                           0x00000d00
#define BRUSH_NOTUSE                               0x00000f00
#define SRC_MONO                                   0x00000000
#define SRC_MONO_LBKGD                             0x00010000
#define SRC_DSTCOLOR                               0x00030000
#define HOST_BIG_ENDIAN_EN                         0x20000000
#define BYTE_ORDER_MSB_TO_LSB                      0x00000000
#define DP_CONVERSION_TEMP                         0x80000000

/* DP_GUI_MASTER_CNTL bit constants */
#define GMC_SRC_PITCH_OFFSET_DEFAULT               0x00000000
#define GMC_SRC_PITCH_OFFSET_LEAVE                 0x00000001
#define GMC_DST_PITCH_OFFSET_DEFAULT               0x00000000
#define GMC_DST_PITCH_OFFSET_LEAVE                 0x00000002
#define GMC_SRC_CLIP_DEFAULT                       0x00000000
#define GMC_DST_CLIP_DEFAULT                       0x00000000
#define GMC_DST_CLIP_LEAVE                         0x00000008
#define GMC_BRUSH_SOLIDCOLOR                       0x000000d0
#define GMC_BRUSH_NOTUSE                           0x000000f0
#define GMC_DST_8BPP                               0x00000200
#define GMC_DST_15BPP                              0x00000300
#define GMC_DST_32BPP                              0x00000600
#define GMC_DST_8BPP_RGB332                        0x00000700
#define GMC_DST_8BPP_Y8                            0x00000800
#define GMC_DST_8BPP_RGB8                          0x00000900
#define GMC_DST_32BPP_AYUV444                      0x00000e00
#define GMC_SRC_MONO                               0x00000000
#define GMC_SRC_MONO_LBKGD                         0x00001000
#define GMC_SRC_DSTCOLOR                           0x00003000
#define GMC_BYTE_ORDER_MSB_TO_LSB                  0x00000000
#define GMC_DP_CONVERSION_TEMP_9300                0x00008000
#define GMC_DP_CONVERSION_TEMP_6500                0x00000000
#define GMC_DP_SRC_RECT                            0x02000000
#define GMC_DP_SRC_HOST                            0x03000000
#define GMC_DP_SRC_HOST_BYTEALIGN                  0x04000000
#define GMC_3D_FCN_EN_CLR                          0x00000000
#define GMC_DST_CLR_CMP_FCN_CLEAR                  0x10000000
#define GMC_AUX_CLIP_CLEAR                         0x20000000
#define GMC_WRITE_MASK_LEAVE                       0x00000000
#define GMC_WRITE_MASK_SET                         0x40000000

/* DP_MIX bit constants *
#define DP_SRC_RECT                                0x00000200
#define DP_SRC_HOST                                0x00000300
#define DP_SRC_HOST_BYTEALIGN                      0x00000400

/******************************************************************************
 *                                                                            *
 *              DP_MIX and DP_GUI_MASTER_CNTL ROP3 RPN constants              *
 *                                                                            *
 *      P = pattern                                                           *
 *      S = source                                                            *
 *      D = destination                                                       *
 *      o = OR                                                                *
 *      a = AND                                                               *
 *      x = XOR                                                               *
 *      n = NOT                                                               *
 *                                                                            *
 ******************************************************************************/
#define ROP3_ZERO                                  0x00000000
#define ROP3_DPSoon                                0x00010000
#define ROP3_DPSona                                0x00020000
#define ROP3_PSon                                  0x00030000
#define ROP3_SDPona                                0x00040000
#define ROP3_DPon                                  0x00050000
#define ROP3_PDSxnon                               0x00060000
#define ROP3_PDSaon                                0x00070000
#define ROP3_SDPnaa                                0x00080000
#define ROP3_PDSxon                                0x00090000
#define ROP3_DPna                                  0x000a0000
#define ROP3_PSDnaon                               0x000b0000
#define ROP3_SPna                                  0x000c0000
#define ROP3_PDSnaon                               0x000d0000
#define ROP3_PDSonon                               0x000e0000
#define ROP3_Pn                                    0x000f0000
#define ROP3_PDSona                                0x00100000
#define ROP3_DSon                                  0x00110000
#define ROP3_SDPxnon                               0x00120000
#define ROP3_SDPaon                                0x00130000
#define ROP3_DPSxnon                               0x00140000
#define ROP3_DPSaon                                0x00150000
#define ROP3_PSDPSanaxx                            0x00160000
#define ROP3_SSPxDSxaxn                            0x00170000
#define ROP3_SPxPDxa                               0x00180000
#define ROP3_SDPSanaxn                             0x00190000
#define ROP3_PDSPaox                               0x001a0000
#define ROP3_SDPSxaxn                              0x001b0000
#define ROP3_PSDPaox                               0x001c0000
#define ROP3_DSPDxaxn                              0x001d0000
#define ROP3_PDSox                                 0x001e0000
#define ROP3_PDSoan                                0x001f0000
#define ROP3_DPSnaa                                0x00200000
#define ROP3_SDPxon                                0x00210000
#define ROP3_DSna                                  0x00220000
#define ROP3_SPDnaon                               0x00230000
#define ROP3_SPxDSxa                               0x00240000
#define ROP3_PDSPanaxn                             0x00250000
#define ROP3_SDPSaox                               0x00260000
#define ROP3_SDPSxnox                              0x00270000
#define ROP3_DPSxa                                 0x00280000
#define ROP3_PSDPSaoxxn                            0x00290000
#define ROP3_DPSana                                0x002a0000
#define ROP3_SSPxPDxaxn                            0x002b0000
#define ROP3_SPDSoax                               0x002c0000
#define ROP3_PSDnox                                0x002d0000
#define ROP3_PSDPxox                               0x002e0000
#define ROP3_PSDnoan                               0x002f0000
#define ROP3_PSna                                  0x00300000
#define ROP3_SDPnaon                               0x00310000
#define ROP3_SDPSoox                               0x00320000
#define ROP3_Sn                                    0x00330000
#define ROP3_SPDSaox                               0x00340000
#define ROP3_SPDSxnox                              0x00350000
#define ROP3_SDPox                                 0x00360000
#define ROP3_SDPoan                                0x00370000
#define ROP3_PSDPoax                               0x00380000
#define ROP3_SPDnox                                0x00390000
#define ROP3_SPDSxox                               0x003a0000
#define ROP3_SPDnoan                               0x003b0000
#define ROP3_PSx                                   0x003c0000
#define ROP3_SPDSonox                              0x003d0000
#define ROP3_SPDSnaox                              0x003e0000
#define ROP3_PSan                                  0x003f0000
#define ROP3_PSDnaa                                0x00400000
#define ROP3_DPSxon                                0x00410000
#define ROP3_SDxPDxa                               0x00420000
#define ROP3_SPDSanaxn                             0x00430000
#define ROP3_SDna                                  0x00440000
#define ROP3_DPSnaon                               0x00450000
#define ROP3_DSPDaox                               0x00460000
#define ROP3_PSDPxaxn                              0x00470000
#define ROP3_SDPxa                                 0x00480000
#define ROP3_PDSPDaoxxn                            0x00490000
#define ROP3_DPSDoax                               0x004a0000
#define ROP3_PDSnox                                0x004b0000
#define ROP3_SDPana                                0x004c0000
#define ROP3_SSPxDSxoxn                            0x004d0000
#define ROP3_PDSPxox                               0x004e0000
#define ROP3_PDSnoan                               0x004f0000
#define ROP3_PDna                                  0x00500000
#define ROP3_DSPnaon                               0x00510000
#define ROP3_DPSDaox                               0x00520000
#define ROP3_SPDSxaxn                              0x00530000
#define ROP3_DPSonon                               0x00540000
#define ROP3_Dn                                    0x00550000
#define ROP3_DPSox                                 0x00560000
#define ROP3_DPSoan                                0x00570000
#define ROP3_PDSPoax                               0x00580000
#define ROP3_DPSnox                                0x00590000
#define ROP3_DPx                                   0x005a0000
#define ROP3_DPSDonox                              0x005b0000
#define ROP3_DPSDxox                               0x005c0000
#define ROP3_DPSnoan                               0x005d0000
#define ROP3_DPSDnaox                              0x005e0000
#define ROP3_DPan                                  0x005f0000
#define ROP3_PDSxa                                 0x00600000
#define ROP3_DSPDSaoxxn                            0x00610000
#define ROP3_DSPDoax                               0x00620000
#define ROP3_SDPnox                                0x00630000
#define ROP3_SDPSoax                               0x00640000
#define ROP3_DSPnox                                0x00650000
#define ROP3_DSx                                   0x00660000
#define ROP3_SDPSonox                              0x00670000
#define ROP3_DSPDSonoxxn                           0x00680000
#define ROP3_PDSxxn                                0x00690000
#define ROP3_DPSax                                 0x006a0000
#define ROP3_PSDPSoaxxn                            0x006b0000
#define ROP3_SDPax                                 0x006c0000
#define ROP3_PDSPDoaxxn                            0x006d0000
#define ROP3_SDPSnoax                              0x006e0000
#define ROP3_PDSxnan                               0x006f0000
#define ROP3_PDSana                                0x00700000
#define ROP3_SSDxPDxaxn                            0x00710000
#define ROP3_SDPSxox                               0x00720000
#define ROP3_SDPnoan                               0x00730000
#define ROP3_DSPDxox                               0x00740000
#define ROP3_DSPnoan                               0x00750000
#define ROP3_SDPSnaox                              0x00760000
#define ROP3_DSan                                  0x00770000
#define ROP3_PDSax                                 0x00780000
#define ROP3_DSPDSoaxxn                            0x00790000
#define ROP3_DPSDnoax                              0x007a0000
#define ROP3_SDPxnan                               0x007b0000
#define ROP3_SPDSnoax                              0x007c0000
#define ROP3_DPSxnan                               0x007d0000
#define ROP3_SPxDSxo                               0x007e0000
#define ROP3_DPSaan                                0x007f0000
#define ROP3_DPSaa                                 0x00800000
#define ROP3_SPxDSxon                              0x00810000
#define ROP3_DPSxna                                0x00820000
#define ROP3_SPDSnoaxn                             0x00830000
#define ROP3_SDPxna                                0x00840000
#define ROP3_PDSPnoaxn                             0x00850000
#define ROP3_DSPDSoaxx                             0x00860000
#define ROP3_PDSaxn                                0x00870000
#define ROP3_DSa                                   0x00880000
#define ROP3_SDPSnaoxn                             0x00890000
#define ROP3_DSPnoa                                0x008a0000
#define ROP3_DSPDxoxn                              0x008b0000
#define ROP3_SDPnoa                                0x008c0000
#define ROP3_SDPSxoxn                              0x008d0000
#define ROP3_SSDxPDxax                             0x008e0000
#define ROP3_PDSanan                               0x008f0000
#define ROP3_PDSxna                                0x00900000
#define ROP3_SDPSnoaxn                             0x00910000
#define ROP3_DPSDPoaxx                             0x00920000
#define ROP3_SPDaxn                                0x00930000
#define ROP3_PSDPSoaxx                             0x00940000
#define ROP3_DPSaxn                                0x00950000
#define ROP3_DPSxx                                 0x00960000
#define ROP3_PSDPSonoxx                            0x00970000
#define ROP3_SDPSonoxn                             0x00980000
#define ROP3_DSxn                                  0x00990000
#define ROP3_DPSnax                                0x009a0000
#define ROP3_SDPSoaxn                              0x009b0000
#define ROP3_SPDnax                                0x009c0000
#define ROP3_DSPDoaxn                              0x009d0000
#define ROP3_DSPDSaoxx                             0x009e0000
#define ROP3_PDSxan                                0x009f0000
#define ROP3_DPa                                   0x00a00000
#define ROP3_PDSPnaoxn                             0x00a10000
#define ROP3_DPSnoa                                0x00a20000
#define ROP3_DPSDxoxn                              0x00a30000
#define ROP3_PDSPonoxn                             0x00a40000
#define ROP3_PDxn                                  0x00a50000
#define ROP3_DSPnax                                0x00a60000
#define ROP3_PDSPoaxn                              0x00a70000
#define ROP3_DPSoa                                 0x00a80000
#define ROP3_DPSoxn                                0x00a90000
#define ROP3_D                                     0x00aa0000
#define ROP3_DPSono                                0x00ab0000
#define ROP3_SPDSxax                               0x00ac0000
#define ROP3_DPSDaoxn                              0x00ad0000
#define ROP3_DSPnao                                0x00ae0000
#define ROP3_DPno                                  0x00af0000
#define ROP3_PDSnoa                                0x00b00000
#define ROP3_PDSPxoxn                              0x00b10000
#define ROP3_SSPxDSxox                             0x00b20000
#define ROP3_SDPanan                               0x00b30000
#define ROP3_PSDnax                                0x00b40000
#define ROP3_DPSDoaxn                              0x00b50000
#define ROP3_DPSDPaoxx                             0x00b60000
#define ROP3_SDPxan                                0x00b70000
#define ROP3_PSDPxax                               0x00b80000
#define ROP3_DSPDaoxn                              0x00b90000
#define ROP3_DPSnao                                0x00ba0000
#define ROP3_DSno                                  0x00bb0000
#define ROP3_SPDSanax                              0x00bc0000
#define ROP3_SDxPDxan                              0x00bd0000
#define ROP3_DPSxo                                 0x00be0000
#define ROP3_DPSano                                0x00bf0000
#define ROP3_PSa                                   0x00c00000
#define ROP3_SPDSnaoxn                             0x00c10000
#define ROP3_SPDSonoxn                             0x00c20000
#define ROP3_PSxn                                  0x00c30000
#define ROP3_SPDnoa                                0x00c40000
#define ROP3_SPDSxoxn                              0x00c50000
#define ROP3_SDPnax                                0x00c60000
#define ROP3_PSDPoaxn                              0x00c70000
#define ROP3_SDPoa                                 0x00c80000
#define ROP3_SPDoxn                                0x00c90000
#define ROP3_DPSDxax                               0x00ca0000
#define ROP3_SPDSaoxn                              0x00cb0000
#define ROP3_S                                     0x00cc0000
#define ROP3_SDPono                                0x00cd0000
#define ROP3_SDPnao                                0x00ce0000
#define ROP3_SPno                                  0x00cf0000
#define ROP3_PSDnoa                                0x00d00000
#define ROP3_PSDPxoxn                              0x00d10000
#define ROP3_PDSnax                                0x00d20000
#define ROP3_SPDSoaxn                              0x00d30000
#define ROP3_SSPxPDxax                             0x00d40000
#define ROP3_DPSanan                               0x00d50000
#define ROP3_PSDPSaoxx                             0x00d60000
#define ROP3_DPSxan                                0x00d70000
#define ROP3_PDSPxax                               0x00d80000
#define ROP3_SDPSaoxn                              0x00d90000
#define ROP3_DPSDanax                              0x00da0000
#define ROP3_SPxDSxan                              0x00db0000
#define ROP3_SPDnao                                0x00dc0000
#define ROP3_SDno                                  0x00dd0000
#define ROP3_SDPxo                                 0x00de0000
#define ROP3_SDPano                                0x00df0000
#define ROP3_PDSoa                                 0x00e00000
#define ROP3_PDSoxn                                0x00e10000
#define ROP3_DSPDxax                               0x00e20000
#define ROP3_PSDPaoxn                              0x00e30000
#define ROP3_SDPSxax                               0x00e40000
#define ROP3_PDSPaoxn                              0x00e50000
#define ROP3_SDPSanax                              0x00e60000
#define ROP3_SPxPDxan                              0x00e70000
#define ROP3_SSPxDSxax                             0x00e80000
#define ROP3_DSPDSanaxxn                           0x00e90000
#define ROP3_DPSao                                 0x00ea0000
#define ROP3_DPSxno                                0x00eb0000
#define ROP3_SDPao                                 0x00ec0000
#define ROP3_SDPxno                                0x00ed0000
#define ROP3_DSo                                   0x00ee0000
#define ROP3_SDPnoo                                0x00ef0000
#define ROP3_P                                     0x00f00000
#define ROP3_PDSono                                0x00f10000
#define ROP3_PDSnao                                0x00f20000
#define ROP3_PSno                                  0x00f30000
#define ROP3_PSDnao                                0x00f40000
#define ROP3_PDno                                  0x00f50000
#define ROP3_PDSxo                                 0x00f60000
#define ROP3_PDSano                                0x00f70000
#define ROP3_PDSao                                 0x00f80000
#define ROP3_PDSxno                                0x00f90000
#define ROP3_DPo                                   0x00fa0000
#define ROP3_DPSnoo                                0x00fb0000
#define ROP3_PSo                                   0x00fc0000
#define ROP3_PSDnoo                                0x00fd0000
#define ROP3_DPSoo                                 0x00fe0000
#define ROP3_ONE                                   0x00ff0000

/******************************************************************************
 *              DP_MIX, DP_GUI_MASTER_CNTL ROP3 named constants               *
 *                                                                            *
 *                  Pattern (P):     1 1 1 1 0 0 0 0                          *
 *                  Source  (S):     1 1 0 0 1 1 0 0                          *
 *                  Destination (D): 1 0 1 0 1 0 1 0                          *
 *                                                                            *
 ******************************************************************************/
#define ROP3_SRCERASE                              0x00440000   // S & ~D
#define ROP3_DSTINVERT                             0x00550000   // ~D
#define ROP3_PATINVERT                             0x005a0000   // P ^ D
#define ROP3_SRCINVERT                             0x00660000   // S ^ D
#define ROP3_SRCAND                                0x00880000   // S & D
#define ROP3_DSTCOPY                               0x00aa0000   // D
#define ROP3_SRCCOPY                               0x00cc0000   // S
#define ROP3_SRCPAINT                              0x00ee0000   // S | D
#define ROP3_PATCOPY                               0x00f00000   // P
#define ROP3_PATPAINT                              0x00fb0000   // P | ~S | D

/* Other Miscellaneous defines */
#define GUI_ACTIVE                                 0x80000000
#define SOFT_RESET_GUI                             0x00000001
#define INC_X           0x0020
#define INC_Y           0x0080

#if defined (__powerpc__)
extern pointer rage128MemReg;

#if defined(RAGE128_BIG_ENDIAN)
static inline void regw(volatile unsigned int regindex, unsigned int regdata)
{
    *(unsigned int *)(rage128MemReg + regindex) = regdata;
}

static inline unsigned int regr(volatile unsigned int regindex)
{
  return(*(unsigned int *)(rage128MemReg + regindex));
}

#else

static inline void regw(unsigned long regindex, unsigned long regdata)
{
  register unsigned long base_addr = (unsigned long)rage128MemReg;
  asm volatile ("stwbrx %1,%2,%3; eieio"
          : "=m" (*(volatile unsigned *)(base_addr+regindex))
          : "r" (regdata), "b" (regindex), "r" (base_addr));
}

static inline unsigned long regr(unsigned long regindex)
{
  register unsigned long base_addr = (unsigned long)rage128MemReg, val;
  asm volatile ("lwbrx %0,%1,%2; eieio"
           : "=r"(val)
           : "b"(regindex), "r"(base_addr),
             "m" (*(volatile unsigned *)(base_addr+regindex)));
  return(val);
}
#endif
#endif /* __powerpc__ */

#define ResetEngine() { regw(GEN_RESET_CNTL, regr(GEN_RESET_CNTL) | SOFT_RESET_GUI); \
                        regw(GEN_RESET_CNTL, regr(GEN_RESET_CNTL) & ~SOFT_RESET_GUI); \
                        regr(GEN_RESET_CNTL); \
                      }

#define WaitQueue(q)  { int i = 100000000; \
                        while(((regr(GUI_STAT) & 0x00000fff) < (q)) && i) i--; \
                        if(i <= 0) ResetEngine() \
                      }

#define WaitIdleEmpty() { int i = 100000000; \
                          WaitQueue(64); \
			  while((regr(GUI_STAT) & GUI_ACTIVE) && i) i--; \
                          if(i <= 0) ResetEngine() \
                        }

#define RAGE128_BIT_BLT(_srcx, _srcy, _dstx, _dsty, _w, _h, _dir) \
{ \
    WaitQueue(4); \
    regw(SRC_Y_X, (((_srcy) << 16) | (_srcx))); \
    regw(DP_CNTL, (_dir)); \
    regw(DST_Y_X, (((_dsty) << 16) | (_dstx))); \
    regw(DST_HEIGHT_WIDTH, (((_h) << 16) | (_w))); \
}

#ifndef NULL
#define NULL    0
#endif

#endif /* REGRAGE128_H */
